/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** configuration.h: configuration parse
****************************************************************************/

#ifndef TMODEL_CONFIG_H_
#define TMODEL_CONFIG_H_

/*******************************************************/
/*           Thermal model input parse                 */
/*******************************************************/

#include <string>
#include <exception>
#include <vector>

/*******************************************************/
class steady_coeff_type {

	public:

	double err_thre_;
	double grad_thre_;
};

class configuration {
	friend std::ostream & operator<<(std::ostream &, const configuration & cf);
	private:

        //The lookup table file containing results of the BTE simulation
        std::string LUT_file_;
        //The device file containing its geometry, material properties and
        //solver options
        std::string device_file_;
        //The chip package file containing the chip layers, dimensions and
        //material properties
        std::string chip_file_;
        //The floorplan file containing the list of chip elements
        std::string floorplan_file_;
        //The layout file containing locations of devices of interest and their 
        //power consumption values
        std::string layout_file_;

        //Time-domain or frequency-domain solver
        char dyn_solver_type_;

		//input power profile
		int moments_;
        bool nhar_adap_; //true if the adaptive nhar algorithm is to be used
		steady_coeff_type steady_coeff_;
	    std::string pfile_;
		static std::string help_;
		static configuration * ap_;
	public:
		class init_error: public std::exception {
			public:
				init_error(const std::string & err) : what_(err) {}
				virtual const char * what() const throw() {return what_.c_str();}
				virtual ~init_error() throw() {};

			private:
				std::string what_;
		};

		class parse_error: public std::exception {
			public:
				parse_error(const std::string & err) : what_(err) {}
				virtual const char * what() const throw() { return what_.c_str();}
				virtual ~parse_error() throw() {};
			
			private:
				std::string what_;
		};

		configuration(int argc, char * const argv []);

		static const configuration  & ap() {return *ap_;}
		static configuration  & wap() {return *ap_;}

        void usage(char * const argv, const char * opt_str);

        void help();

		std::string LUT_file() {return LUT_file_;}
		std::string LUT_file() const {return LUT_file_;}

		std::string device_file() {return device_file_;}
		std::string device_file() const {return device_file_;}

		std::string chip_file() {return chip_file_;}
		std::string chip_file() const {return chip_file_;}

		std::string floorplan_file() {return floorplan_file_;}
		std::string floorplan_file() const {return floorplan_file_;}

		std::string layout_file() {return layout_file_;}
		std::string layout_file() const {return layout_file_;}

		char dyn_solver_type() {return dyn_solver_type_;}
		char dyn_solver_type() const {return dyn_solver_type_;}

		int moments() {return moments_;}
		int moments() const {return moments_;}

        bool nhar_adap() {return nhar_adap_;}
        bool nhar_adap() const {return nhar_adap_;}

		double std_err_thre() {return steady_coeff_.err_thre_;}
		double std_err_thre() const {return steady_coeff_.err_thre_;}

		double std_grad_thre() {return steady_coeff_.grad_thre_;}
		double std_grad_thre() const {return steady_coeff_.grad_thre_;}

		std::string powerfile() {return pfile_;}
		std::string powerfile() const {return pfile_;}

		~configuration() { ap_ = 0;}
};


		
#endif
