/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** device_structure.cc: contains functions for solving the entire structure
****************************************************************************/

#include "header.h"
#include "util.h"
#include "configuration.h"


//*==========================================================================*/
//* Generate the LUT containing the BTE thermal resistance and the Fourier
//* thermal resistance
void Device_structure::generateLUT(string filename) {

    //Get the BTE thermal resistance
    //Find the peak temperature and divide it by the power value

    double power = qin;
    cout << "The power value = " << power << endl;

    ofstream LUT;
    LUT.open(filename.c_str());

    //Get the peak temperature
    double peak_temp = get_peak_temp();

    cout << "Peak temperature = " << peak_temp << endl;

    LUT << (peak_temp - 273.15) / power;

    //Now resolve the device using Fourier, and find the thermal resistance
    Device_structure device;

    device.input(configuration::ap().device_file().c_str());

    device.substrate.all_fourier_ = 1;

    device.solve();

    peak_temp = device.get_peak_temp();

    LUT << " " << (peak_temp - 273.15) / power;

}
//*==========================================================================*/
//* Find the peak temperature in the device
double Device_structure::get_peak_temp() {

    double peak_temp = 0;

    int cells_x = substrate.f_cells_x;
    int cells_y = substrate.f_cells_y;
    int cells_z = substrate.f_cells_z;

    for(int i = 1 - substrate.tbound_x_0; 
                    i < (cells_x - 1 + substrate.tbound_x_end); i++) {
        for(int j = 1 - substrate.tbound_y_0;
	                    j < (cells_y - 1 + substrate.tbound_y_end); j++) {
            for(int k = 1 - substrate.tbound_z_0;
					        k < (cells_z - 1 + substrate.tbound_z_end); k++) {

				int siize = bte_system.size();
				//for all the bte regions
				for(int v = 0; v < siize; v++) {
					//if its inside the BTE region
					if((i >= (bte_coords[v].x-1) &&
                        j >= (bte_coords[v].y-1) &&
						k >= (bte_coords[v].z-1)) &&
					   (i <= (bte_coords[v].x+bte_coords[v].cells_x-2) &&
						j <= (bte_coords[v].y+bte_coords[v].cells_y-2)&&
						k <= (bte_coords[v].z+bte_coords[v].cells_z-2))){
						
					    if(bte_system[v].cell[i-bte_coords[v].x+1]
						   [j-bte_coords[v].y+1][k-bte_coords[v].z+1].T_
                            > peak_temp) {

					        peak_temp = bte_system[v].cell[i-bte_coords[v].x+1]
						                              [j-bte_coords[v].y+1]
                                                      [k-bte_coords[v].z+1].T_;
						}

					}
				}
			    if(substrate.f_cell[i][j][k].T_ > peak_temp) {
			        peak_temp = substrate.f_cell[i][j][k].T_;
			    }
            }
        }
    }

    return peak_temp;

}
   
//*==========================================================================*/
//* get coordinate information for all materials in entire structure
void Device_structure::get_material_coords() {

    int num_pieces; //the number of pieces a device has
    int num_devices = device.size();    //the total number of devices

    int i,j;

    int x,y,z;
    int cells_x,cells_y,cells_z;

    //loop over all the devices
    for(i=0; i<num_devices; i++)
    {
        //get the number of pieces in this device
        num_pieces = device[i].size();
        //loop over all the pieces in this device
        for(j=0;j<num_pieces; j++)
        {
            x=get_index(device[i][j].x_coor,'x',0);
            y=get_index(device[i][j].y_coor,'y',0);
            z=get_index(device[i][j].z_coor,'z',0);
            
            cells_x=get_index(device[i][j].x_coor+
                        device[i][j].length_x,'x',1)-x+1;
            cells_y=get_index(device[i][j].y_coor+
                        device[i][j].length_y,'y',1)-y+1;
            cells_z=get_index(device[i][j].z_coor+
                        device[i][j].length_z,'z',1)-z+1;
                                       
        
            substrate.mat_coords.push_back(Material_coords(x,y,z,
                            cells_x,cells_y,cells_z,
                            materials[device[i][j].material_index]));
        }
    }
}

/*==========================================================================*/
//* define the BTE regions and create a BTE_structre for each 
void Device_structure::set_solvers() {

    int num_pieces; //the number of pieces a device has
    int num_devices = device.size();    //the total number of devices

    //the coordiantes of the box surrounding the BTE regions
    double x_min, x_max, y_min, y_max, z_min, z_max;  
    //the coordiantes after extending 10*vg*t_rel in all directi(ns
    double x_min_ext, x_max_ext, y_min_ext, y_max_ext, z_min_ext, z_max_ext; 
    double del_x, del_y, del_z;

    int i,j;

    int x_min_index, x_max_index, y_min_index, y_max_index;
    int z_min_index, z_max_index;

    double limit,limit_max;
    
    del_x = substrate.del_x;
    del_y = substrate.del_y;
    del_z = substrate.del_z;

    x_min = x_max = y_min = y_max = z_min = z_max = 0;

    //set the solvers for all structures

    //loop over all the devices
    for(i=0; i<num_devices; i++)
    {
        //get the number of pieces in this device
        num_pieces = device[i].size();
        //loop over all the pieces in this device
        for(j=0; j<num_pieces; j++)
        {
            //check if they need to be solved using BTE or not
            device[i][j].set_solver(materials[device[i][j].material_index]);
        }

        //combine all BTE regions into a single one
        
        //first initialize x_min, x_max, y_min, .... with the first BTE piece
        int no_bte = 1;
        for(j=0;j<num_pieces; j++)
        {
            if(device[i][j].all_bte_==1)
            {
                x_min = device[i][j].x_coor;
                x_max = device[i][j].x_coor + device[i][j].length_x;
                y_min = device[i][j].y_coor;
                y_max = device[i][j].y_coor + device[i][j].length_y;
                z_min = device[i][j].z_coor;
                z_max = device[i][j].z_coor + device[i][j].length_z;
                no_bte = 0;
                substrate.all_fourier_ = 0;
                break;
            }
        }

        //check if there aren't any BTE pieces
        if(no_bte==1)
        //if yes
        {
            //solve all the system using Fourier
            substrate.all_fourier_=1;
            break;
        }
            
        //second get x_min, x_max, y_min, .... of all the BTE pieces starting
        //from the second BTE piece
        for(int k=j+1;k<num_pieces; k++)
        {
            if(device[i][k].all_bte_==1)
            {
                if(device[i][k].x_coor<x_min)
                {
                    x_min = device[i][k].x_coor;
                }
                if(device[i][k].x_coor + device[i][k].length_x>x_max)
                {
                    x_max = device[i][k].x_coor + device[i][k].length_x;
                }
                if(device[i][k].y_coor<y_min)
                {
                    y_min = device[i][k].y_coor;
                }
                if(device[i][k].y_coor + device[i][k].length_y>y_max)
                {
                    y_max = device[i][k].y_coor + device[i][k].length_y;
                }
                if(device[i][k].z_coor<z_min)
                {
                    z_min = device[i][k].z_coor;
                }
                if(device[i][k].z_coor + device[i][k].length_z>z_max)
                {
                    z_max = device[i][k].z_coor + device[i][k].length_z;
                }
            }
        }

        //extend the BTE regions 10*vg*t_rel in all directions where vg,t_rel
        //are the maximum of all the materials in the slice next to the BTE
        //region boundaries

        //extend in -ve x direction

        limit_max = 0;

        //first get max(vg*t_rel) for the materials just west of x_min

        //loop over all pieces
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just west of x_min in the volume defined
            //by x:(x_min-del_x)->(x_min), 
            //   y:(y_min)->(y_max)
            //   z:(z_min)->(z_max)
            if(device[i][j].isInside(x_min-del_x,x_min,y_min,
                                     y_max,z_min,z_max))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }

        x_min_ext = x_min - bte_region_expand_factor * limit_max;
                
        //do the same for all other directions

        limit_max = 0;

        //get max(vg*t_rel) for the materials just east of x_max
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just east of x_max in the volume defined
            //by x:(x_max)->(x_max+del_x), 
            //   y:(y_min)->(y_max)
            //   z:(z_min)->(z_max)
            if(device[i][j].isInside(x_max,x_max+del_x,y_min,
                                    y_max,z_min,z_max))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }

        x_max_ext = x_max + bte_region_expand_factor * limit_max;

        limit_max = 0;

        //get max(vg*t_rel) for the materials just south of y_min
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just south of y_min in the volume
            //defined by y:(y_min-del_y)->(y_min),
            //           x:(x_min)->(x_max)
            //           z:(z_min)->(z_max)
            if(device[i][j].isInside(x_min,x_max,y_min-del_y,
                                     y_min,z_min,z_max))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }
        
        y_min_ext = y_min - bte_region_expand_factor * limit_max;

        limit_max = 0;

        //get max(vg*t_rel) for the materials just north of y_max
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just north of y_max in the volume
            //defined by y:(y_max)->(y_max+del_x)
            //           x:(x_min)->(x_max)
            //           z:(z_min)->(z_max)
            if(device[i][j].isInside(x_min,x_max,y_max,
                                     y_max+del_x,z_min,z_max))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }

        y_max_ext = y_max + bte_region_expand_factor * limit_max;

        limit_max = 0;

        //get max(vg*t_rel) for the materials just bottom of z_min
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just bottom of z_min in the volume 
            //defined by z:(z_min-del_z)->(z_min)
            //   x:(x_min)->(x_max)
            //   y:(y_min)->(y_max)
            if(device[i][j].isInside(x_min,x_max,y_min,
                                     y_max,z_min-del_z,z_min))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }
        
        z_min_ext = z_min - bte_region_expand_factor * limit_max;

        limit_max = 0;

        //get max(vg*t_rel) for the materials just top of z_max
        for(j=0;j<num_pieces;j++)
        {
            //check if this piece lies just top of z_max in the volume
            //defined by z:(z_max)->(z_max+del_z)
            //           x:(x_min)->(x_max)
            //           y:(y_min)->(y_max)
            if(device[i][j].isInside(x_min,x_max,y_min,
                                     y_max,z_max,z_max+del_z))
            {
                //if it lies check if its vg*t_rel is greater than
                //max(vg*t_rel)
                limit = materials[device[i][j].material_index].vg_*
                        materials[device[i][j].material_index].t_rel_;
                if(limit>limit_max)
                {
                    limit_max = limit;
                }
            }
        }
        
        z_max_ext = z_max + bte_region_expand_factor * limit_max;

        //flags to indicate if this BTE structure has a neighbouring Fourier
        //one
        //default is that each BTE structure has a neighbouring Foruier region
        int hasNeigh[6]={1,1,1,1,1,1};
        
        //for specular boundaries, =1 if specular, =0 else
        int spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end;

        //for diffuse boundaries, =1 if diffuse, =0 else
        int diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end;

        //for constant temperature boundaries, =1 if temp bound, =0 else
        int tbound_x_0, tbound_x_end, tbound_y_0, tbound_y_end, tbound_z_0,
        tbound_z_end;
        double tbx_0_val, tbx_end_val, tby_0_val, tby_end_val, tbz_0_val,
        tbz_end_val;

        //the BTE structure doesn't have any of the boundary types, except if
        //it touches the substrate boundary, so initially initialize all of
        //them to 0
        spec_x_0=spec_x_end=spec_y_0=spec_y_end=spec_z_0=spec_z_end=0;
        diff_x_0=diff_x_end=diff_y_0=diff_y_end=diff_z_0=diff_z_end=0;
        tbound_x_0=tbound_x_end=tbound_y_0=tbound_y_end=
                                                    tbound_z_0=tbound_z_end=0;
        tbx_0_val=tbx_end_val=tby_0_val=
                        tby_end_val=tbz_0_val=tbz_end_val=substrate.T_ini;

        //saturate the extended boundaries such that they don't exceed the
        //substrate boundaries

        //if exceeds bounadry
        if(x_min_ext<=0)
        {
            //saturate
            x_min_ext=0;
            //flag that it has no west Fourier neighbour
            hasNeigh[1]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_x_0 = substrate.spec_x_0;
            diff_x_0 = substrate.diff_x_0;
            tbound_x_0 = substrate.tbound_x_0;
            tbx_0_val = substrate.tbx_0_val;
        }
        //same for all other boundaries
        if(x_max_ext>=substrate.length_x)
        {
            x_max_ext=substrate.length_x;
            hasNeigh[0]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_x_end = substrate.spec_x_end;
            diff_x_end = substrate.diff_x_end;
            tbound_x_end = substrate.tbound_x_end;
            tbx_end_val = substrate.tbx_end_val;

        }
        if(y_min_ext<=0)
        {
            y_min_ext=0;
            hasNeigh[3]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_y_0 = substrate.spec_y_0;
            diff_y_0 = substrate.diff_y_0;
            tbound_y_0 = substrate.tbound_y_0;
            tby_0_val = substrate.tby_0_val;
        }
        if(y_max_ext>=substrate.length_y)
        {
            y_max_ext=substrate.length_y;
            hasNeigh[2]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_y_end = substrate.spec_y_end;
            diff_y_end = substrate.diff_y_end;
            tbound_y_end = substrate.tbound_y_end;
            tby_end_val = substrate.tby_end_val;
        }
        if(z_min_ext<=0)
        {
            z_min_ext=0;
            hasNeigh[5]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_z_0 = substrate.spec_z_0;
            diff_z_0 = substrate.diff_z_0;
            tbound_z_0 = substrate.tbound_z_0;
            tbz_0_val = substrate.tbz_0_val;
        }
        if(z_max_ext>=substrate.length_z)
        {
            z_max_ext=substrate.length_z;
            hasNeigh[4]=0;
            //set its boundary type to that of the corresponding substrate
            //boundary
            spec_z_end = substrate.spec_z_end;
            diff_z_end = substrate.diff_z_end;
            tbound_z_end = substrate.tbound_z_end;
            tbz_end_val = substrate.tbz_end_val;
        }
        //get the cell indices of the boundaries
        x_min_index = get_index(x_min_ext,'x',0);
        x_max_index = get_index(x_max_ext,'x',1);
        y_min_index = get_index(y_min_ext,'y',0);
        y_max_index = get_index(y_max_ext,'y',1);
        z_min_index = get_index(z_min_ext,'z',0);
        z_max_index = get_index(z_max_ext,'z',1);

        //check for all materials inside this BTE region and store in a vector
        //to pass to the BTE structure later
        
        vector<Material_coords> mat_coords_tmp;
        int mat_x_min,mat_x_max,mat_y_min,mat_y_max,mat_z_min,mat_z_max;

        for(j=0;j<num_pieces;j++)
        {
            if(device[i][j].isInside(x_min_ext,x_max_ext,y_min_ext,
                                     y_max_ext,z_min_ext,z_max_ext))
            {
                //get indices of material boundary
                mat_x_min=get_index(device[i][j].x_coor,'x',0);
                mat_x_max=get_index(device[i][j].x_coor+
                                    device[i][j].length_x,'x',1);
                mat_y_min=get_index(device[i][j].y_coor,'y',0);
                mat_y_max=get_index(device[i][j].y_coor+
                                    device[i][j].length_y,'y',1);
                mat_z_min=get_index(device[i][j].z_coor,'z',0);
                mat_z_max=get_index(device[i][j].z_coor+
                                    device[i][j].length_z,'z',1);
                
                //get material boundary that is inside the BTE region
                if(mat_x_min<x_min_index)
                {
                    mat_x_min=x_min_index;
                }
                
                if(mat_x_max>x_max_index)
                {
                    mat_x_max=x_max_index;
                }
                if(mat_y_min<y_min_index)
                {
                    mat_y_min=y_min_index;
                }
                if(mat_y_max>y_max_index)
                {
                    mat_y_max=y_max_index;
                }
                if(mat_z_min<z_min_index)
                {
                    mat_z_min=z_min_index;
                }
                if(mat_z_max>z_max_index)
                {
                    mat_z_max=z_max_index;
                }
                
                //add this material to mat_coords
                //the coordiantes are relative to the BTE structure
                mat_coords_tmp.push_back(
                        Material_coords(mat_x_min-x_min_index,
                                        mat_y_min-y_min_index,
                                        mat_z_min-z_min_index,
                                        mat_x_max-mat_x_min+2,
                                        mat_y_max-mat_y_min+2,
                                        mat_z_max-mat_z_min+2,
                                        materials[device[i][j].material_index]
                                        ));
            }
        }                           

        //get the parts of edges that correspond to this BTE structure
        vector<double> edge_x_tmp = 
            vector<double>(substrate.edge_x.begin()+x_min_index,
                           substrate.edge_x.begin()+x_max_index+1);
        vector<double> edge_y_tmp = 
            vector<double>(substrate.edge_y.begin()+y_min_index,
                           substrate.edge_y.begin()+y_max_index+1);
        vector<double> edge_z_tmp = 
            vector<double>(substrate.edge_z.begin()+z_min_index,
                           substrate.edge_z.begin()+z_max_index+1);

        //get the parts of centroids that correspond to this BTE structure
        vector<double> cen_x_tmp = 
            vector<double>(substrate.cen_x.begin()+x_min_index,
                           substrate.cen_x.begin()+x_max_index+1);
        vector<double> cen_y_tmp = 
            vector<double>(substrate.cen_y.begin()+y_min_index,
                           substrate.cen_y.begin()+y_max_index+1);
        vector<double> cen_z_tmp = 
            vector<double>(substrate.cen_z.begin()+z_min_index,
                           substrate.cen_z.begin()+z_max_index+1);
        
        //add the 2 cells at the boundaries to edges
        edge_x_tmp.insert(edge_x_tmp.begin(), edge_x_tmp[0]);
        edge_x_tmp.push_back(edge_x_tmp[edge_x_tmp.size()-1] + del_x);
        edge_y_tmp.insert(edge_y_tmp.begin(), edge_y_tmp[0]);
        edge_y_tmp.push_back(edge_y_tmp[edge_y_tmp.size()-1] + del_y);
        edge_z_tmp.insert(edge_z_tmp.begin(), edge_z_tmp[0]);
        edge_z_tmp.push_back(edge_z_tmp[edge_z_tmp.size()-1] + del_z);
          

        //add the 2 cells at the boundaries to centroids
        cen_x_tmp.insert(cen_x_tmp.begin(), cen_x_tmp[0] - 0.5*del_x);
        cen_x_tmp.push_back(cen_x_tmp[cen_x_tmp.size()-1] + 0.5*del_x);
        cen_y_tmp.insert(cen_y_tmp.begin(), cen_y_tmp[0] - 0.5*del_y);
        cen_y_tmp.push_back(cen_y_tmp[cen_y_tmp.size()-1] + 0.5*del_y);
        cen_z_tmp.insert(cen_z_tmp.begin(), cen_z_tmp[0] - 0.5*del_z);
        cen_z_tmp.push_back(cen_z_tmp[cen_z_tmp.size()-1] + 0.5*del_z);
         
        //initialize the BTE_structure and pass all the necessary information
        //information passed:
        //x_coor,y_coor,z_coor,
        //length_x,length_y,length_z,
        //cells_x,cells_y,cells_z,
        //mat_coords,
        //edge_x,edge_y,edge_z,
        //cen_x,cen_y,cen_z,
        //neighbour information,
        //spec_x_0,spec_x_end,.....
        //diff_x_0,diff_x_end,.....
        //tbound_x_0,tbound_x_end,.....
        //tbx_0_val,tbx_end_val,.....
        bte_system.push_back(BTE_structure(x_min_ext,y_min_ext,z_min_ext,
                  x_max_ext-x_min_ext,y_max_ext-y_min_ext,z_max_ext-z_min_ext,
                  x_max_index-x_min_index+3,
                  y_max_index-y_min_index+3,
                  z_max_index-z_min_index+3,
                  mat_coords_tmp,
                  edge_x_tmp,edge_y_tmp,edge_z_tmp,
                  cen_x_tmp,cen_y_tmp,cen_z_tmp,
                  hasNeigh,
                  spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end,
                  diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end,
                  tbound_x_0,tbound_x_end,tbound_y_0,tbound_y_end,
                  tbound_z_0,tbound_z_end,
                  tbx_0_val,tbx_end_val,tby_0_val,tby_end_val,
                  tbz_0_val,tbz_end_val));

        //add the information of this BTE structure to bte_coords
        bte_coords.push_back(Block_coords(x_min_index,y_min_index,z_min_index,
                                          x_max_index-x_min_index+3,
                                          y_max_index-y_min_index+3,
                                          z_max_index-z_min_index+3));
                
    }

}

//*==========================================================================*/
//* define the whole system as one BTE region containing all devices inside 
void Device_structure::set_all_bte() {

    int num_pieces; //the number of pieces a device has
    int num_devices = device.size();    //the total number of devices

    //the coordiantes of the box surrounding the BTE region
    double x_min, x_max, y_min, y_max, z_min, z_max;  

    int i,j;

    int x_min_index, x_max_index, y_min_index, y_max_index;
    int z_min_index, z_max_index;

    x_min = substrate.x_coor;
    x_max = substrate.x_coor + substrate.length_x;
    y_min = substrate.y_coor;
    y_max = substrate.y_coor + substrate.length_y;
    z_min = substrate.z_coor;
    z_max = substrate.z_coor + substrate.length_z;

    //set the solvers for all structures

    //flags to indicate if this BTE structure has a neighbouring Fourier
    //one
    int hasNeigh[6]={0,0,0,0,0,0};
        
    //for specular boundaries, =1 if specular, =0 else
    int spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end;

    //for diffuse boundaries, =1 if diffuse, =0 else
    int diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end;

    //for constant temperature boundaries, =1 if temp bound, =0 else
    int tbound_x_0, tbound_x_end, tbound_y_0, tbound_y_end, tbound_z_0,
    tbound_z_end;
    double tbx_0_val, tbx_end_val, tby_0_val, tby_end_val, tbz_0_val,
    tbz_end_val;

    //the BTE structure has boundary types same as the substrate
    spec_x_0 = substrate.spec_x_0;
    diff_x_0 = substrate.diff_x_0;
    tbound_x_0 = substrate.tbound_x_0;
    tbx_0_val = substrate.tbx_0_val;
    spec_x_end = substrate.spec_x_end;
    diff_x_end = substrate.diff_x_end;
    tbound_x_end = substrate.tbound_x_end;
    tbx_end_val = substrate.tbx_end_val;
    spec_y_0 = substrate.spec_y_0;
    diff_y_0 = substrate.diff_y_0;
    tbound_y_0 = substrate.tbound_y_0;
    tby_0_val = substrate.tby_0_val;
    spec_y_end = substrate.spec_y_end;
    diff_y_end = substrate.diff_y_end;
    tbound_y_end = substrate.tbound_y_end;
    tby_end_val = substrate.tby_end_val;
    spec_z_0 = substrate.spec_z_0;
    diff_z_0 = substrate.diff_z_0;
    tbound_z_0 = substrate.tbound_z_0;
    tbz_0_val = substrate.tbz_0_val;
    spec_z_end = substrate.spec_z_end;
    diff_z_end = substrate.diff_z_end;
    tbound_z_end = substrate.tbound_z_end;
    tbz_end_val = substrate.tbz_end_val;

    //get the cell indices of the boundaries
    x_min_index = 1;
    x_max_index = substrate.f_cells_x - 2;
    y_min_index = 1;
    y_max_index = substrate.f_cells_y - 2;
    z_min_index = 1; 
    z_max_index = substrate.f_cells_z - 2;

    vector<Material_coords> mat_coords_tmp;
    int mat_x_min,mat_x_max,mat_y_min,mat_y_max,mat_z_min,mat_z_max;

    for(i=0;i<num_devices;i++)
    {
        num_pieces = device[i].size();    
        for(j=0;j<num_pieces;j++)
        {
            //get indices of material boundary
            mat_x_min=get_index(device[i][j].x_coor,'x',0);
            mat_x_max=get_index(device[i][j].x_coor+
                                device[i][j].length_x,'x',1);
            mat_y_min=get_index(device[i][j].y_coor,'y',0);
            mat_y_max=get_index(device[i][j].y_coor+
                                device[i][j].length_y,'y',1);
            mat_z_min=get_index(device[i][j].z_coor,'z',0);
            mat_z_max=get_index(device[i][j].z_coor+
                                device[i][j].length_z,'z',1);
            
            //add this material to mat_coords
            mat_coords_tmp.push_back(
                    Material_coords(mat_x_min-x_min_index,
                                    mat_y_min-y_min_index,
                                    mat_z_min-z_min_index,
                                    mat_x_max-mat_x_min+1,
                                    mat_y_max-mat_y_min+1,
                                    mat_z_max-mat_z_min+1,
                                    materials[device[i][j].material_index]
                                    ));
        }
    }                           

    //initialize the BTE_structure and pass all the necessary information
    //information passed:
    //x_coor,y_coor,z_coor,
    //length_x,length_y,length_z,
    //cells_x,cells_y,cells_z,
    //mat_coords,
    //edge_x,edge_y,edge_z,
    //cen_x,cen_y,cen_z,
    //neighbour information,
    //spec_x_0,spec_x_end,.....
    //diff_x_0,diff_x_end,.....
    //tbound_x_0,tbound_x_end,.....
    //tbx_0_val,tbx_end_val,.....
    bte_system.push_back(BTE_structure(x_min,y_min,z_min,
              x_max-x_min,y_max-y_min,z_max-z_min,
              x_max_index-x_min_index+3,
              y_max_index-y_min_index+3,
              z_max_index-z_min_index+3,
              mat_coords_tmp,
              substrate.edge_x,substrate.edge_y,substrate.edge_z,
              substrate.cen_x,substrate.cen_y,substrate.cen_z,
              hasNeigh,
              spec_x_0,spec_x_end,spec_y_0,spec_y_end,spec_z_0,spec_z_end,
              diff_x_0,diff_x_end,diff_y_0,diff_y_end,diff_z_0,diff_z_end,
              tbound_x_0,tbound_x_end,tbound_y_0,tbound_y_end,
              tbound_z_0,tbound_z_end,
              tbx_0_val,tbx_end_val,tby_0_val,tby_end_val,
              tbz_0_val,tbz_end_val));
    //add the information of this BTE structure to bte_coords
    bte_coords.push_back(Block_coords(x_min_index,y_min_index,z_min_index,
                                      x_max_index-x_min_index+3,
                                      y_max_index-y_min_index+3,
                                      z_max_index-z_min_index+3));
}

//*==========================================================================*/
//get the index
int Device_structure::get_index(double coor, char direction, int max) {

    if(direction=='x')
    {
        int cells_x=substrate.f_cells_x;
        for(int i=1;i<cells_x-1;i++)
        {
            if(coor>=substrate.edge_x[i]&&coor<substrate.edge_x[i+1])
            {
                if(coor<substrate.cen_x[i])
                {
                    if(max==0)
                    {
                        return i;
                    }
                    else
                    {
                        return i-1;
                    }
                }
                else
                {
                    if(max==0)
                    {
                        return i+1;
                    }
                    else
                    {
                        return i;
                    }
                }
            } 
        }
        if(coor==substrate.edge_x[cells_x-1])
        {
            return (cells_x-2);
        }
    }
    else if(direction=='y')
    {
        int cells_y=substrate.f_cells_y;
        for(int i=1;i<cells_y-1;i++)
        {
            if(coor>=substrate.edge_y[i]&&coor<substrate.edge_y[i+1])
            {
                if(coor<substrate.cen_y[i])
                {
                    if(max==0)
                    {
                        return i;
                    }
                    else
                    {
                        return i-1;
                    }
                }
                else
                {
                    if(max==0)
                    {
                        return i+1;
                    }
                    else
                    {
                        return i;
                    }
                }
            } 
        }
        if(coor==substrate.edge_y[cells_y-1])
        {
            return (cells_y-2);
        }
    }
    else if(direction=='z')
    {
        int cells_z=substrate.f_cells_z;
        for(int i=1;i<cells_z-1;i++)
        {
            if(coor>=substrate.edge_z[i]&&coor<substrate.edge_z[i+1])
            {
                if(coor<substrate.cen_z[i])
                {
                    if(max==0)
                    {
                        return i;
                    }
                    else
                    {
                        return i-1;
                    }
                }
                else
                {
                    if(max==0)
                    {
                        return i+1;
                    }
                    else
                    {
                        return i;
                    }
                }
            } 
        }
        if(coor==substrate.edge_z[cells_z-1])
        {
            return (cells_z-2);
        }
    }
    return -1;
    
}

//*===========================================================================*/
void Device_structure::lock_bte_cells()
{
    int size = bte_coords.size();

    //go over all BTE structures
    for(int h=0;h<size;h++)
    {
        int x = bte_coords[h].x;
        int cells_x = bte_coords[h].cells_x;
        int y = bte_coords[h].y;
        int cells_y = bte_coords[h].cells_y;
        int z = bte_coords[h].z;
        int cells_z = bte_coords[h].cells_z;

        for(int i=x;i<(x+cells_x-2);i++)
        {
            for(int j=y;j<(y+cells_y-2);j++)
            {
                for(int k=z;k<(z+cells_z-2);k++)
                {
                    substrate.f_cell[i][j][k].locked = 1;
                }
            }
        }
    }
}

//*===========================================================================*/            
//* solve the entire structure by solving the BTE and Fourier sections of the
//* individual systems
void Device_structure::solve() {


    initialization();

	int size = bte_system.size();

    int total_bte_cells = 0;

    for(int bte = 0; bte < size; bte++)
    {
        total_bte_cells = total_bte_cells + (bte_system[bte].cells_x-2)*
                          (bte_system[bte].cells_y-2)*
                          (bte_system[bte].cells_z-2);
    }
    cout<<"Total number of BTE cells = "<<total_bte_cells<<endl;
    cout<<"Total number of Fourier Cells = "<<(substrate.f_cells_x*
                                                    substrate.f_cells_y*
                                                    substrate.f_cells_z)-
                                                    total_bte_cells<<endl;

	//flag for convergence, >0 if solution not converged
	int conv = 1;	

	//check to make sure the reflecting boudaries don't hit walls
	check_reflecting_boundaries();
	 
    lock_bte_cells();

    substrate.condition_coeffs_fourier();

    conv = 1;    
    int count = 0;
	int conver[size];
	int conver_sub = 1;

    double qout = 0;

	for(int g = 0; g < size; g++) {
		conver[g] = 1;
	}

	if(substrate.transient_ == 1 && substrate.all_bte_ == 1) {
		cout<<"Calling Transient BTE solver"<<endl;

		//open output file here to save time so it isn't opened and closed lots
		ofstream fout_tran;
	    fout_tran.open("tran_out");
	
		//since it is transient, resize all of the storage for the future
		//energy densities (f_angle_next_t). Resize to match size of f_angle
		resize_trans_f_angle();
	
		//loop over necessary time
	    //go over time steps until reach the end
		double del_t = substrate.step_time; //time step
		double stop_time = substrate.total_time;
		cout << "Step time is: " << del_t;
		cout << " Stop time is: "<< substrate.total_time << endl;
	    double time = 0; //current time
    	while(time < stop_time) {

			//set the Q's to the correct value (update) depending on time
			set_Q(time);
		
            //go over all structures
            for(int h = 0; h < size; h++) {
                bte_system[h].solve_transient(this,del_t);
            }

			//update time
			time = time + del_t;
		    
			//print the temperatures for this time, if necessary
			output_peak_trans(fout_tran,time);
        }
		fout_tran.close();
	}
	else if(substrate.transient_ == 1 && substrate.all_fourier_==1) {
		cout << "Calling Transient Fourier solver" << endl;

		//open output file here to save time so it isn't opened and closed lots
		ofstream fout_tran;
	    fout_tran.open("output_tran");
	
		//loop over necessary time
	    //go over time steps until reach the end
		double del_t = substrate.step_time; //time step
		double stop_time = substrate.total_time;
		cout << "Step time is: " << del_t;
		cout << " Stop time is: " << substrate.total_time << endl;
	    double time = 0; //current time
    	while(time < stop_time) {

			//set the Q's to the correct value (update) depending on time
			set_Q_fourier(time);
		
            //go over all structures
            substrate.solve_transient(this,del_t);

			//update time
			time = time + del_t;
		    
			//print the temperatures for this time, if necessary
			output_peak_trans_fourier(fout_tran,time);

        }
		fout_tran.close();
	}

    else if(substrate.all_bte_ == 0 && substrate.all_fourier_ == 0)
    {
		//initialize t_old
		for(unsigned int h = 0; h < bte_system.size(); h++) {
			for (int i = 0; i < bte_system[h].cells_x; i++) {
				for (int j = 0; j < bte_system[h].cells_y; j++) {
					for (int k = 0; k < bte_system[h].cells_z; k++) {
						bte_system[h].cell[i][j][k].T_old = bte_system[h].T_ini;
					}
				}
			}
		}

	    //solve the Fourier and BTE methods until they both converge

		while(conv!=0/*||count<300*/||qout<0.9999*qin){
    		count++;
//			cout<<"Global iteration #"<<count<<endl;

		    //reset the convergence flag to 0
		    conv = 0;

			//First solve BTE

		    //go over all structures
		    for(int h=0; h<size; h++) {


			    //conv > 0 if solution not converged
				//solve(int) returns 1 if solution not converged
  				conver[h] = bte_system[h].solve(iter_param,this);

				conv = conv +  conver[h];
				conver[h]=1;

	    	}
	
    		//Now solve the Fourier region

			conver_sub = substrate.solve_fourier(iter_param,this);

			conv = conv + conver_sub;

            update_bte_t_bound();

            update_bte_semireflect();

			double qout_bte = 0;
            for(int i = 1; i < substrate.f_cells_x-1; i++) {
                for(int j = 1; j < substrate.f_cells_y-1; j++) {
                    for(int k = 1; k < substrate.f_cells_z-1; k++) {
                        qout_bte = qout_bte + substrate.f_cell[i][j][k].Qprime;
                    }
                }
            }
            //cout<<"Heat going out of bte = "<<qout_bte<<endl;
            qout = substrate.calc_qout(bte_coords, bte_system);
            cout << "Heat going out = " << qout;
            cout <<" W out of " << qin << " W input heat" << endl;
            //cout<<"Converged "<<conv<<endl<<endl;

    	}
        
    }
    else if(substrate.all_bte_==1)
    {
        cout<<"Calling All BTE solver"<<endl;
		int count =0;
	    //solve the BTE method until it converges
	    while(conv!=0 ||count<1570||qout<0.9999*qin) {
			count++;
            cout<<"count = "<<count; //<<" "<<qout<<" "<<0.99*qin<<" "<<conv<<endl;
			//reset the convergence flag to 0
		    conv = 0;

		    //go over all structures
		    for(int h=0; h<size; h++) {

			    //conv > 0 if solution not converged
			    //solve(int) returns 1 if solution not converged
       			conv = conv + bte_system[h].solve(iter_param,this);
			
	    	}
            qout=substrate.calc_qout(bte_coords, bte_system);
            cout<<"Heat going out = "<<qout;
            cout<<" W out of "<<qin<<" W input heat"<<endl;
            //cout<<"Converged "<<conv<<endl<<endl;

        }
    }
    else
    {
	    cout<<"Calling the All Fourier solver"<<endl;
		int count =0;
	    //solve the Fourier method until it converges
	    while(conv!=0||qout<0.9999*qin) {
            count++;

		    //reset the convergence flag to 0
		    conv = 0;
    		//solve the substrate
    		conv = conv + substrate.solve_fourier(iter_param,this);
        
	        qout=substrate.calc_qout(bte_coords, bte_system);
            cout<<"Heat going out = "<<qout;
            cout<<" W out of "<<qin<<" W input heat"<<endl<<endl;
        }
    }

}
//*===========================================================================*/
//* class constructor
Device_structure::Device_structure() {
	//set the default number of iterations to 10
	iter_param=10;
	qin=0;

	//set the null material values
	null_material.name = "null";
	null_material.k_cond = 0;

	bte_region_expand_factor = 10;

}
