/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** dysol.h
****************************************************************************/

#ifndef _dysol_H
#define _dysol_H

#define ADAPH_FEULER

#ifdef YDEBUG
#define DBG_DO(x) (x)
#else
#define DBG_DO(x) do {} while (0)
#endif

#include <functional>
#include <vector>
#include <iostream>
#include "hoct_model.h"
#include "util.h"
#include "TimeIndex.h"

class dysol {

/*==========================================================================*/
public :

    //* Constructor of dysol class. */
	dysol();

    //* Global Runge-Kutta 4th order method */
    //* Using global interval to update the temperature,
    //* each time, based on the current temperature, obtainning the 
    //* ks for all the grids, and then update. */
	void rk4(RC_network_type & AGC, 
		std::vector<double> & temp, std::vector<double> & power,
		double interval);
    //* Calculate the temperature based on the current AGC after one step size 
    //* without updating the AGC, used to adjust the step size. */ 
	void forward_step(RC_network_type & AGC,
		std::vector<double> & temp, std::vector<double> & power,
		double & interval, std::vector<double> & c_temp);
    //* Determine the approximate step size each time before calling
    //* the rk4 function size. */
	void adj_ssize(RC_network_type & AGC,  
		std::vector<double> & temp, std::vector<double> & power,
		double & err_thres, double & in_inval, double & out_inval);
    //* Run-time solver using 4th order Runge-Kutta algorithm. */
	void rk4_glb_solver(RC_network_type & AGC, 
		std::vector<double> & temp, std::vector<double> & power,
		double interval, double err_thres, double & t_end, 
		double & final_intval);
 
    //* Semi third mothod for block-by-block update */
    //* Semi_th method, which update temperature of grids one by one:
    // * update the tepmerature and the slope of block[index] */
	void semi_th(const RC_network_type & AGC,
		std::vector<double> & temp, std::vector<std::vector<double> >& old_temp, 
                std::vector<double> & power, int index, double interval);
    //* Using different steps(0.75 * old_h two steps, 1.5 * old_h one 
    //* step to arrive the same time slot, and compare the diffrence from
    //* these two different method to determine the appropximate step size. */ 
	double adjsize_semi_th_implct(const RC_network_type & AGC,
		std::vector<double> & temp, std::vector<std::vector<double> >& old_temp, 
		double err_thres, int index, double c_tslot, double in_inval, double t_end);
    //* Run-time solver using forward Euler order method, temperatures are 
    //* updated block by block. */
	void semi_th_implct_solver(const RC_network_type & AGC,
		std::vector<double> & temp, std::vector<std::vector<double> >& old_temp, 
		std::vector<double> & power, double interval, double err_thres, 
                double t_end);
    //* Obtain the initial slope of all blocks based on the initial temperature
    //* use equation 17.
    //* All temperatures are current temperatures (temperature at time 0, slope
    //* at time 0. */
	void init_slope(const RC_network_type & AGC,
		std::vector<double> & temp, std::vector<double> & power);

	std::vector<double> alpha;
	std::vector<double> cur_slope;
	std::vector<std::vector<double> > cur_intval;
	std::vector<double> cur_tslot;
	std::vector<double> tao;
	std::vector<double> cap_;
	std::vector<double> cur_1_slope, cur_2_slope;

	std::vector<bool> quies_flag;
	double quies_sum; 
	static constexpr double DERIV_THRES = 1e2;
	static constexpr double TEMP_THRES = 0.015;
	static constexpr double FMAX_ = 1.0e10;

 private :
	
};


//* Output the peak temperature for a vector of double. */
void display_peakt(std::vector<double> temp);

#endif	
