// Copyright 2018 by Robert Dick.
// All rights reserved.

#include "MemChecker.h"
#include "RStd.h"

#include <cstring>
#include <limits>

using namespace std;
using namespace rstd;

namespace rstd {

/*###########################################################################*/
namespace {

const size_t ALIGN_BYTES = sizeof (double);

/* Pad to use at beginning and end of allocated memory chunks.  malloc()
returns aligned chunks, and ALIGN_BYTES preserves alignment, so this will be
aligned. */
const size_t PAD_SIZE = ALIGN_BYTES;

const unsigned char INIT_CONST = 0x5a;
const unsigned char CLEAR_CONST = 0xa5;

/*===========================================================================*/
	size_t actual_bytes(size_t bytes) {
// Make sure PAD_SIZE is an even multiple of ALIGN_BYTES.
		static_assert((PAD_SIZE / ALIGN_BYTES) * ALIGN_BYTES == PAD_SIZE,
			"Pad not aligned.");
		return bytes + 2 * PAD_SIZE;
	}

/*===========================================================================*/
	void * wall_new(size_t bytes) {
		size_t real_size = actual_bytes(bytes);
		char * mem = static_cast<char *>(malloc(real_size));

		if (! mem) {
			throw std::bad_alloc();
		}

		return mem + PAD_SIZE;
	}

/*===========================================================================*/
	void wall_delete(void * data) {
		if (! data) {
			return;
		}

		char * cdata = static_cast<char *>(data);
		free(cdata - PAD_SIZE);
	}
}

MemChecker::MemCheckStack MemChecker::mc_stack_;

/*===========================================================================*/
MemChecker * MemChecker::mc() {
	return (mc_stack_.empty()) ? 0 : mc_stack_.top();
}

/*===========================================================================*/
MemChecker::MemChecker(const char * name) :
	mem_map_(),
	min_alloc_(0),
	max_alloc_(0),
	min_byte_(0),
	max_byte_(0),
	ignore_min_(0),
	ignore_max_(0),
	max_print_(10),
	name_(name)
{
	mc_stack_.push(this);
}

/*===========================================================================*/
void MemChecker::max_print(size_t mprint) {
	max_print_ = mprint;
}

/*===========================================================================*/
void MemChecker::expect_alloc_cnt(size_t min_alloc_cnt,
size_t max_alloc_cnt) {
	min_alloc_ = min_alloc_cnt;
	max_alloc_ = max_alloc_cnt;
}

/*===========================================================================*/
void MemChecker::expect_alloc_cnt(size_t alloc_cnt) {
	expect_alloc_cnt(alloc_cnt, alloc_cnt);
}

/*===========================================================================*/
void MemChecker::expect_byte_cnt(size_t min_byte_cnt, size_t max_byte_cnt) {
	min_byte_ = min_byte_cnt;
	max_byte_ = max_byte_cnt;
}

/*===========================================================================*/
void MemChecker::expect_byte_cnt(size_t byte_cnt) {
	expect_byte_cnt(byte_cnt, byte_cnt);
}

/*===========================================================================*/
void MemChecker::expect_caller(size_t ignore_loc_min,
	size_t ignore_loc_max)
{
	ignore_min_ = reinterpret_cast<void *>(ignore_loc_min);
	ignore_max_ = reinterpret_cast<void *>(ignore_loc_max);
}

/*===========================================================================*/
void MemChecker::expect_caller(size_t ignore_loc) {
	expect_caller(ignore_loc, ignore_loc);
}

/*===========================================================================*/
bool
MemChecker::is_ignore(const void * addr) const {
	return addr >= ignore_min_ && addr <= ignore_max_;
}

/*===========================================================================*/
MemChecker::~MemChecker() {
		check(max_print_);
		RASSERT(mc_stack_.size());
		mc_stack_.pop();
}

/*===========================================================================*/
void *
MemChecker::checked_alloc(size_t bytes) {
	char * mem = static_cast<char *>(wall_new(bytes));
	RASSERT(mem);

	if (mem_map_.find(reinterpret_cast<size_t>(mem)) != mem_map_.end()) {
		std::cerr << "MemChecker " << name_ <<
			": Allocation of memory that was already allocated.\n";

		Rabort();
	}

#ifdef __GNUC__
	const void * caller = __builtin_return_address(1);
#else
	const void * caller = numeric_limits<size_t>::max();
#endif

	memset(mem - PAD_SIZE, INIT_CONST, actual_bytes(bytes));

	pair<size_t, AllocRecord>
		p(reinterpret_cast<size_t>(mem), AllocRecord(bytes, caller));

	pair<MemMap::iterator, bool> ret = mem_map_.insert(p);

	RASSERT(ret.second);

	return mem;
}

/*===========================================================================*/
void MemChecker::checked_dealloc(void * data) {
	if (! data) {
		return;
	}

	if (mem_map_.empty()) {
		cerr << "MemChecker " << name_ << ": Deletion although there were "
			"no outstanding allocations.\n";

		Rabort();
	}

	MemMap::iterator mem_iter = mem_map_.find(reinterpret_cast<size_t>(data));

	if (mem_iter == mem_map_.end()) {
		cerr << "MemChecker " << name_ <<
			": Deletion of memory that was not allocated.\n";

		Rabort();
	}

	AllocRecord record = mem_iter->second;

#if 0
// Zero-sized allocs allowed.
	if (! record.bytes) {
		cerr << "MemChecker " << name_ <<
			": Deletion of memory with a size of zero.\n";

		Rabort();
	}
#endif

	char * cdata = static_cast<char *>(data);

	MAPU(x, PAD_SIZE) {
		if (cdata[x - PAD_SIZE] != INIT_CONST) {
			cerr << "MemChecker " << name_ << ": Start pad damaged.\n";
			Rabort();
		}
	}

	MAPU(x, PAD_SIZE) {
		if (cdata[x + record.bytes] != INIT_CONST) {
			cerr << "MemChecker " << name_ << ": End pad damaged.\n";
			Rabort();
		}
	}

	memset(cdata - PAD_SIZE, CLEAR_CONST, actual_bytes(record.bytes));

	mem_map_.erase(mem_iter);

	wall_delete(data);
}

/*===========================================================================*/
size_t
MemChecker::out_allocs() const {
	size_t alloc_cnt = 0;

	for (MemMap::const_iterator i = mem_map_.begin(); i != mem_map_.end();
		++i)
	{
		if (! is_ignore(i->second.caller)) {
			++alloc_cnt;
		}
	}

	return alloc_cnt;
}

/*===========================================================================*/
size_t
MemChecker::out_bytes() const {
	size_t size = 0;

	for (MemMap::const_iterator i = mem_map_.begin(); i != mem_map_.end();
		++i)
	{
		if (! is_ignore(i->second.caller)) {
			size += i->second.bytes;
		}
	}

	return size;
}

/*===========================================================================*/
void MemChecker::check(size_t mprint) {
	size_t oa = out_allocs();
	size_t ob = out_bytes();

	if (oa < min_alloc_ || oa > max_alloc_ ||
		ob < min_byte_ || ob > max_byte_)
	{
		cerr << "MemChecker " << name_ << ": Expected min allocs: " <<
			min_alloc_ << ", max allocs: " << max_alloc_ << "\n";

		cerr << "MemChecker " << name_ << ": Expected min bytes: " <<
			min_byte_ << ", max bytes: " << max_byte_ << "\n";

		cerr << "MemChecker " << name_ << ": Outstanding number of allocs: " <<
			oa << "\n";

		cerr << "MemChecker " << name_ << ": Outstanding number of bytes: " <<
			ob << "\n";

		if (mprint) {
			cerr << "MemChecker " << name_ <<
				": Allocation information follows: " <<
				"(position, bytes, caller: 'l *0xYYYY' in gdb).\n";

			size_t print_cnt = 0;

			for (MemMap::const_iterator i = mem_map_.begin();
				i != mem_map_.end() && print_cnt < mprint; ++i)
			{
				if (! is_ignore(i->second.caller)) {
					cerr << "MemChecker " << name_ << ": (" << i->first << ", " <<
						i->second.bytes << ", " << i->second.caller << ")\n";
					++print_cnt;
				}
			}

			if (print_cnt < oa) {
				cerr << "Truncated after " << mprint << " leaks.\n";
			}

			if (ignore_min_ || ignore_max_) {
				cerr << "Ignored callers from " << ignore_min_ << " to " <<
					ignore_max_ << ".\n";
			}

			cerr << endl;
		}
	}
}

/*###########################################################################*/
void MemChecker_test() {
	{
		cout << "Making new MemChecker (local)\n";
		MemChecker mc("local1");

		cout << "Leaking 3 intentionally.\n";
		new int;
		new char[12];
		new double;
	}

#ifdef ROB_DEBUG
// These are too dangerous to try without the MemChecker on.

	{
		MemChecker mc("local2");
		long * x = new long;
		int iconst = 0xff & INIT_CONST;

		couthex << "Should be " << iconst << ": " << *x << "\n";

		*x = 0;
		delete x;

		int cconst = 0xff & CLEAR_CONST;
		couthex << "Should be " << cconst << ": " << *x << "\n";

		x = new long;
		cout << "Damaging end pad intentionally.  "
			"This will also cause an accidental leak and abort the program.\n";

		cout << flush;

		*(x + 1) = 0;
		delete x;
	}
#endif
}
}

/*###########################################################################*/
#ifdef ROB_DEBUG

// These need to be in the global namespace.

/*===========================================================================*/
void *
operator new(size_t bytes) _GLIBCXX_THROW (std::bad_alloc) {
	MemChecker * mc = MemChecker::mc();

	if (mc) {
		return mc->checked_alloc(bytes);
	} else {
		return wall_new(bytes);
	}
}

/*===========================================================================*/
void operator delete(void * data) throw () {
	MemChecker * mc = MemChecker::mc();

	if (mc) {
		mc->checked_dealloc(data);
	} else {
		wall_delete(data);
	}
}

/*===========================================================================*/
void *
operator new[](size_t bytes) _GLIBCXX_THROW (std::bad_alloc) {
	MemChecker * mc = MemChecker::mc();

	if (mc) {
		return mc->checked_alloc(bytes);
	} else {
		return wall_new(bytes);
	}
}

/*===========================================================================*/
void operator delete[](void * data) throw () {
	MemChecker * mc = MemChecker::mc();

	if (mc) {
		mc->checked_dealloc(data);
	} else {
		wall_delete(data);
	}
}

#endif

