/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** grids.cc: code for creating grids
****************************************************************************/

#include "grids.h"
#include "Geom.h"
#include <cassert>
#include <limits>
#include <utility>
#include <algorithm>
#include <iostream>

using namespace isac;
using namespace std;
using namespace rstd;

namespace isac {

/*===========================================================================*/
/* Creates element set forming a uniform 3-D grid.
 * Accepts divisions along each axis, dimension along each axis,
 * and default Element initialization value. */
vector<Element_type>
create_grid(int divx, int divy, int divz,
double size_x, double size_y, double size_z, double default_value) {
	vector<Element_type> elts;

	Element_type elt(size_x / divx, size_y / divy, size_z / divz, 
		0.0, 0.0, 0.0, default_value);

/* This loop sequence guarantees that the generated grid follow the usual 
matrix dequence */
/*
  00 01 02 ... 0N
  10
  20
  . 
  .
  .
  N0 N1 N2 ... NN

*/

    for (int j = (divy-1); j > -divy; j=(j-2)) {
		for (int i = (-divx+1); i < divx; i=(i+2)) {
			for (int k = 0; k < divz; ++k) {
				elt.center[0] = (0.5 * i) * elt.size[0];
				elt.center[1] = (0.5 * j) * elt.size[1];
				elt.center[2] = (0.5 + k) * (elt.size[2]);
				elts.push_back(elt);
			}
		}
	}
	
	return elts;
}


/*===========================================================================*/
/* Creates element set forming a uniform 2-D grid */
vector<Element_type>
create_grid(int divx, int divy, double size_x, double size_y,
double default_value) {
	vector<Element_type> elts;

	Element_type elt(size_x / divx, size_y / divy, 0.0, 0.0, default_value);
/* This loop sequence guarantees that the generated grid follow the usual 
matrix dequence */
/*
  00 01 02 ... 0N
  10
  20
  . 
  .
  .
  N0 N1 N2 ... NN

*/
	for (int j = (divy-1); j > -divy; j=(j-2)) {
		for (int i = (-divx+1); i < divx; i=(i+2)) {
			elt.center[0] = (0.5 * i) * elt.size[0];
			elt.center[1] = (0.5 * j) * elt.size[1];
			elts.push_back(elt);
		}
	}
	
	return elts;
}



} /* namespace isac */
