/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** hoct_model.cc: code for the hybrid orthogonal tree class
****************************************************************************/

#include <math.h>
#include <sys/time.h>
#include <assert.h>
#include <fstream>
#include "util.h"
#include "hoct_model.h"

using namespace std;


/*==========================================================================*/
/* Operator to output information in grid "pg" */
ostream& operator<<(ostream& os, const par_grid & pg)
{
    os<<pg.id_<<"::"<<pg.act_level_<<"  <"<<pg.coors_[X_]<<":"<<
        pg.coors_[Y_]<<":"<<pg.coors_[Z_]<<">  <"<<pg.sizes_[X_]<<":"<<
        pg.sizes_[Y_]<<":"<<pg.sizes_[Z_]<<
        ">  <";
        for(int i = 0 ; i < NEI_C_; i++) {
            for(int j = 0; j < static_cast<int>(pg.nei_[i].size()); j++) {
                  os << pg.nei_[i][j]->id()<< " id-type " 
                << pg.nei_[i][j] -> type()
                << ":";
            }os<<"|";
        }os<<">\n";
		os << "  children num is " << pg.int_chi_array().size()<< "\n";
        for(int i = 0; i < static_cast<int>
            (pg.int_chi_array_.size()); i++) {
            os<<(*pg.int_chi_array_[i]);
        }

    return os;
}


/*==========================================================================*/
/* Constructor for par_grid class */
par_grid::par_grid():
id_(INV_N_),
//level_(INV_N_),
self_level_(INV_N_),
act_level_(INV_N_),
type_(GRID_),
par_(NULL),
levelid_(),
nei_(),
int_chi_array_(),
power_nei_(),
adjp_coeff_(),
power_rela_(),
sizes_(),
coors_(),
power_(),
temp_(0),
cond_(0),
ccap_(0),
partition_flag_(),
gid_(),
chi_array_()
{
    for(int i = 0; i < DIM_C_; i++) {
        sizes_[i] = 0.0;
        coors_[i] = 0.0;
    }
    nei_.resize(NEI_C_);

}


/*==========================================================================*/
/* Destructor for par_grid class */
par_grid::~par_grid(){

}


/*==========================================================================*/
/* Check whether grids are neighbors or not. 
 * If they are neighbors, the returned value represents the direction 
 * (0--x direction, 1--y direction, 3--z direction)of their neighbor 
 * relationship; Otherwise, the returned value will be -1 */
int par_grid::nei_check(par_grid * a) {

    double dis[DIM_C_];
    double diff[DIM_C_];
    int adj_flag[DIM_C_];
    int counter = 0;
    int flag = -1;
    int dis_adj_dim = -1;

    /* If two are neighbors in one direction, then counter will increment by 1 
     * dis_adj_dim records that direction */
    for(int i = 0; i < DIM_C_; i++) {
        dis[i] = coors_[i] - a->coors(i);
        diff[i] = fabs(dis[i]) - (sizes_[i] + a->sizes(i))/2;
        if(fabs(diff[i]) < FMIN_) {
            counter ++;
            dis_adj_dim = i;
        }
    }

    if(counter != 1) {
        return  flag;
    }

    /* Check all directions, if one direction returns 1 for adj_check()
     * function call, counter is incremented by 1 and also check that
     * direction is not the one recorded by dis_adj_dim */ 
    for(int i = 0; i < DIM_C_; i++) {
        adj_flag[i] = adj_check(coors_[i], sizes_[i], 
                  a->coors(i), a->sizes(i));
        if(adj_flag[i] == 1) {
            counter ++;
            assert(dis_adj_dim != i);
        }
    }

    if(counter != 3) {
        return flag;
    }

    /* Return the direction of the neighbor relationship (X_NEG_, X_POS_,
     * Y_NEG_, Y_POS_, Z_NEG_ or Z_POS_)*/
    if(dis[dis_adj_dim] < FMIN_) {
        flag = dis_adj_dim * 2 + 1;
    }else {
        flag = dis_adj_dim * 2;
    }
    
    return flag;
}


/*===========================================================================*/
/* Verify whether the current grid and grid a are neighbors in Z dimension.
 * Check whether they have intersect area in x and y dimension. 
 * If they do not have overlap area in x/y dimension, then they are actually
 * not neighbors in Z direction.
 * Return true if they have overlap area, return false if not. */
int par_grid::xyintersect_check(par_grid * a) {
	
	int counter = 0;
	for (int i = 0; i < (DIM_C_ - 1); ++i){
        int adj_flag = adj_check(coors_[i], sizes_[i], 
            a->coors(i), a->sizes(i));
		if (adj_flag == 1){
			counter ++;
		}
	}

	if (counter == 2){
		return true;
	}else{
		return false;
	}
}


/*==========================================================================*/
/* Calculate the overlapping area in x/y dimension for two grids */
double par_grid::area_intersect(par_grid * grid_b){

	double iarea = 1.0;

    for (int i = 0; i < (DIM_C_ - 1); ++i){
		const double this_min = coors(i) - sizes(i)/2.0;
		const double this_max = coors(i) + sizes(i)/2.0;
		const double b_this_min = grid_b->coors(i) - grid_b->sizes(i)/2.0;
		const double b_this_max = grid_b->coors(i) + grid_b->sizes(i)/2.0;
		
		const double min_p = max(this_min, b_this_min);	
		const double max_p = min(this_max, b_this_max);

		if (max_p < min_p){
			return 0.0;
		}	

		iarea = iarea * (max_p - min_p);
	}

    return iarea;	
}

/*==========================================================================*/
/* Calculate the overlapping area in "dim" dimension for two grids */
double par_grid::area_intersect(par_grid * grid_b, vector<int> dim){

	double iarea = 1.0;

    for (int j = 0; j < (DIM_C_ - 1); ++j){
        int i = dim[j];
		const double this_min = coors(i) - sizes(i)/2.0;
		const double this_max = coors(i) + sizes(i)/2.0;
		const double b_this_min = grid_b->coors(i) - grid_b->sizes(i)/2.0;
		const double b_this_max = grid_b->coors(i) + grid_b->sizes(i)/2.0;
		
		const double min_p = max(this_min, b_this_min);	
		const double max_p = min(this_max, b_this_max);

		if (max_p < min_p){
			return 0.0;
		}	

		iarea = iarea * (max_p - min_p);
	}

    return iarea;	
}

/*==========================================================================*/
/* Calculate the overlapping area in x/y dimension between a grid and a
 * conductivity/capacity map block */
int par_grid::blk_intersect(const Map_block blk, Map_block & intrsct) {

    for (int i = 0; i < (DIM_C_ - 1); ++i) {
        const double this_min = coors(i) - sizes(i)/2.0;
        const double this_max = coors(i) + sizes(i)/2.0;
        const double blk_min = blk.center[i] - blk.size[i]/2.0;
        const double blk_max = blk.center[i] + blk.size[i]/2.0;

        const double min_p = max(this_min, blk_min);
        const double max_p = min(this_max, blk_max);

        if (max_p < min_p) {
            return 0;
        }
        
        intrsct.size[i] = max_p - min_p;
        intrsct.center[i] = (max_p + min_p)/2.0;
    }

    return 1;

}

/*===========================================================================*/
/* Constructor for grid_contour class */
grid_contour::grid_contour():
self_(NULL),
nei_()
{

}


/*===========================================================================*/
/* Constructor for grid_contour class */
grid_contour::grid_contour(par_grid * a):
self_(a),
nei_()
{
	//nei_.resize(NEI_C_);
} 


/*==========================================================================*/
/* Deconstructor for grid_contour class */
grid_contour::~grid_contour(){

    //delete(self_);
}


/*============================================================================*/
/* Constructor for RC_network_type class */
RC_network_type::RC_network_type():
size_(),
A_(),
G_(),
C_()
{ 

}


/*============================================================================*/
/* Constructor for par_logic_pos class */
par_logic_pos::par_logic_pos():
lpos_(),
divs_(),
sizes_(){

}


/*============================================================================*/
/* Do the refinement for "root" based on the "par"
 * "par" defines the children of X,Y,Z dimension of "root" when further 
 * refined. */
void gen_par_tree(par_grid * root, const vector<int> & par) {

    
    /* Find the total levels of children before the refinement is conducted. */
	int levels[DIM_C_];
	int total_level = 0;
	for(int i=0; i < DIM_C_; i++) {
		int a = static_cast<int>(log2(par[i]));
		levels[i] = a;
		total_level = total_level<a ? a : total_level;
	}

	root ->chi_array().resize(total_level);
	int curr_level = 0;
	int level_grids = 1;

	double sizes[DIM_C_];
	for(int i=0; i < DIM_C_; i++) {
		sizes[i] = root->sizes(i);
	}
   
	par_grid * curr_root; 

    /* Generate the children level by level */
	while(levels[X_] != 0 || levels[Y_] != 0 || levels[Z_] != 0) {

		vector<double> off_var;
		off_var.clear();
		off_var.resize(DIM_C_, 0);

		int p_flag[DIM_C_] = {0,0,0};
		int st_grids = 1;
	
		for(int i = 0; i < DIM_C_; i++) {
			if(levels[i]) {
				st_grids *= 2;	
				sizes[i] /= 2;
				levels[i] --;
				off_var[i] = -sizes[i]/2;
				p_flag[i] = 1;
			}
		}
	    
		vector<vector<double> > coor_offset;
		coor_offset.clear();
		coor_offset.resize(st_grids);
		for(int i = 0; i < st_grids; i++) {
			coor_offset[i].resize(DIM_C_);
		}

		int index_t = 0;
		for(int i=0; i <= p_flag[Z_]; i++) {
			for(int j=0; j <= p_flag[Y_]; j++) {
				for(int k=0; k <= p_flag[X_]; k++) {
					coor_offset[index_t][X_] = off_var[X_];
					coor_offset[index_t][Y_] = off_var[Y_];
					coor_offset[index_t][Z_] = off_var[Z_];
					index_t ++;
					off_var[X_] = -off_var[X_];
				}//end for
				off_var[Y_] = -off_var[Y_];
			}//end for
			off_var[Z_] = -off_var[Z_];
		}//end for
	
		int id_t = 0;	
		for(int i = 0; i < level_grids; i ++) {
			if(curr_level == 0) {
				curr_root = root;
			}
			else {
				curr_root = root->chi_array(curr_level-1, i);
			}
		  	curr_root->partition_flag_assign(p_flag);
		  	//curr_root->level_assign(total_level-curr_level);
		  	for(int j = 0; j < st_grids; j ++) {
				par_grid * a = new par_grid;

				a->id_assign(j);
				a->gid_assign(id_t);
				id_t ++;
				a->sizes_assign(sizes);
				/*	Assign the coordinates */
				a->coors_assign(coor_offset[j][X_] + curr_root->coors(X_),
					coor_offset[j][Y_] + curr_root->coors(Y_),
					coor_offset[j][Z_] + curr_root->coors(Z_));

				/*	Assign the conductivity and capacitivity */
                //Check if the conductivity in this element is homogoneous
                if(curr_root->cond_map().size() == 0) {
                    //If so, assign the conductivity the normal way
				    a->cond_assign(curr_root->cond());
                }
                else {
                    //Otherwise, calculate the equivalent conductivity of this
                    //element
                    vector<Map_block> & cond_map = curr_root->cond_map();
                    vector<Map_block> cur_cond_map;
                    //First, find all blocks of the conductivity map lying
                    //inside this element
                    int cond_map_size = cond_map.size();
                    for(int blk = 0; blk < cond_map_size; blk++) {
                        Map_block tmp;
                        if(a->blk_intersect(cond_map[blk], tmp)) {
                            //this block will be passed to the newly created
                            //grid
                            tmp.value = cond_map[blk].value;
                            cur_cond_map.push_back(tmp);
                        }
                    }
                    a->bulk_cond_assign(curr_root->bulk_cond());
                    a->cond_map_assign(cur_cond_map);
                    //Second, calculate the weighted average of the
                    //conductivities lying in this element
                    double elt_area = sizes[X_] * sizes[Y_];
                    double bulk_area = elt_area;
                    double eq_cond = 0.0;
                    cond_map_size = cur_cond_map.size();
                    for(int blk = 0; blk < cond_map_size; blk++) {
                        //Get this block's area
                        double blk_area = cur_cond_map[blk].size[X_] * 
                                          cur_cond_map[blk].size[Y_];

                        //Subtract this block's area from the total element's
                        //area
                        bulk_area -= blk_area;             
                        //Add the conductivity of this block weighted by its 
                        //area to the equivalent conductivity
                        eq_cond += cur_cond_map[blk].value * blk_area;
                    }
                    //Finally, add the bulk conductivity weighted by its area
                    eq_cond += a->bulk_cond() * bulk_area;
                    eq_cond /= elt_area;
                    //Assign the element the equivalent conductivity
                    a->cond_assign(eq_cond);
                }
                //Check if the heat capacity in this element is homogoneous
                if(curr_root->ccap_map().size() == 0) {
                    //If so, assign the heat capacity the normal way
				    a->ccap_assign(curr_root->ccap());
                }
                else {
                    //Otherwise, calculate the equivalent heat capacity of this
                    //element
                    vector<Map_block> & ccap_map = curr_root->ccap_map();
                    vector<Map_block> cur_ccap_map;
                    //First, find all blocks of the heat capacity map lying
                    //inside this element
                    int ccap_map_size = ccap_map.size();
                    for(int blk = 0; blk < ccap_map_size; blk++) {
                        Map_block tmp;
                        if(a->blk_intersect(ccap_map[blk], tmp)) {
                            //this block will be passed to the newly created
                            //grid
                            tmp.value = ccap_map[blk].value;
                            cur_ccap_map.push_back(tmp);
                        }
                    }
                    a->bulk_ccap_assign(curr_root->bulk_ccap());
                    a->ccap_map_assign(cur_ccap_map);
                    //Second, calculate the weighted average of the
                    //heat capacities lying in this element
                    double elt_area = sizes[X_] * sizes[Y_];
                    double bulk_area = elt_area;
                    double eq_ccap = 0.0;
                    ccap_map_size = cur_ccap_map.size();
                    for(int blk = 0; blk < ccap_map_size; blk++) {
                        //Get this block's area
                        double blk_area = cur_ccap_map[blk].size[X_] * 
                                          cur_ccap_map[blk].size[Y_];

                        //Subtract this block's area from the total element's
                        //area
                        bulk_area -= blk_area;             
                        //Add the heat capacity of this block weighted by its 
                        //area to the equivalent heat capacity
                        eq_ccap += cur_ccap_map[blk].value * blk_area;
                    }
                    //Finally, add the bulk heat capacity weighted by its area
                    eq_ccap += a->bulk_ccap() * bulk_area;
                    eq_ccap /= elt_area;
                    //Assign the element the equivalent heat capacity
                    a->ccap_assign(eq_ccap);
                }
				//a->cap_assign(a->ccap()*a->sizes(X_)*a->sizes(Y_)*a->sizes(Z_));

				curr_root->int_chi_array_assign(a);
				root->chi_array_assign(a, curr_level);
				a->par_assign(curr_root);
				a->act_level_assign(curr_root->act_level()+1);

				/* Assign power_nei information */
				vector<pair<int, par_grid *> > & cur_nei = 
					curr_root->power_nei(); 
				for (int jj = 0; jj < static_cast<int>(cur_nei.size()); ++jj){
					int dir_flag = a -> nei_check(cur_nei[jj].second);
					if (dir_flag>=0){
						a->power_nei().push_back(cur_nei[jj]);
					}
				}
				/* Assign temperature information */
				a->temp_assign(curr_root->temp());
			}//end for
		}
	
		level_grids *= st_grids;
		curr_level ++;
	}
}


/*===========================================================================*/
/* Find the upper_level grid (in level "upper_level") for the "cur_grid".
 * Each grid are in levels (par()->(self_level()+1)~ grid->self_level()),
 * so if par->self_level() equal to upper_level, par() is the 
 * upper_level grid, otherwise, cur_grid is the upper_level grid */
par_grid * upper_level_grid(par_grid * cur_grid, int upper_level) {

	par_grid * par = cur_grid->par();

    if (cur_grid->act_level() == (START_LEVEL_ - 1)){
        return cur_grid;
    }else{
        if (par->self_level() == upper_level ) {
		    return par;
	    }else {
		    return cur_grid;
	    }
    }
}


/*===========================================================================*/
/* Generate a contour structure (for hetergeneous partition, contour is 
 * necessary) for "root" based on the tree structure */
int hoct_tree_contour(par_grid * root, vector<vector<grid_contour> > & 
    hoct_contour, vector<int> & lid, int depth) {

	int size = root -> int_chi_array().size();
	if (size == 0) {
		hoct_contour[depth-1].push_back(grid_contour(root));
		root->levelid_assign(depth-1, lid[depth-1]);
	    //cout<<"here:"<<lid[depth-1]<<":"<<hoct_contour[depth-1][lid[depth-1]]
        //.self_->lid(depth-1)<<endl;
		root->self_level_assign(depth-1);
		lid[depth-1] ++;
		return (depth - 1);
	}

    int chi_level = MAX_L_;
	for(int i = 0; i < size; i++) {
		par_grid * curr_grid = root->int_chi_array(i);
		int cle = hoct_tree_contour(curr_grid, hoct_contour, lid, depth);
		chi_level = chi_level > cle? cle: chi_level;
	}
	for(int i = 0; i < size; i++) {
		par_grid * curr_grid = root->int_chi_array(i);
		for(int j = chi_level; j < curr_grid->self_level(); j++) {
			hoct_contour[j].push_back(grid_contour(curr_grid));
			curr_grid->levelid_assign(j, lid[j]);
			lid[j] ++;
		}
	}

	chi_level--;
	hoct_contour[chi_level].push_back(grid_contour(root));
	root->self_level_assign(chi_level);
	root->levelid_assign(chi_level, lid[chi_level]);
	lid[chi_level] ++;
	
	return chi_level;
}

/*===========================================================================*/
/* Generate the first 4 levels of the hoct_contour, which won't change when 
 * further partition is conducted */
void hoct_cont_init(par_grid * root, vector<vector<grid_contour> > & 
    hoct_contour) {

    int start_level = 4;
    /* Generate the first 4 levels of the hoct_contour */
    clear_contour(hoct_contour);

    hoct_contour.resize(start_level);
    push_grid_contour(root, hoct_contour, START_LEVEL_);

}


/*===========================================================================*/
/* Generate the contour structure based on the grid (push the grid and 
 * all of its children info. to contour ) */
void push_grid_contour(par_grid * grid, 
    vector<vector<grid_contour> > & hoct_contour, int start_level){
    
    //int start_level = 4;
    if (grid->act_level() < (start_level-1)){
        for (unsigned int i = 0; i < grid->int_chi_array().size(); ++i){
            push_grid_contour(grid-> int_chi_array(i), hoct_contour,
                start_level);
        }
    }

    hoct_contour[grid->act_level()].push_back(grid_contour(grid));
}


/*===========================================================================*/
/* Assign levelid information to isothermal and constant temperature grids */
void assign_levelid_isotcont(vector<vector<grid_contour> > & hoct_contour,
    int depth){

    int need_level = START_LEVEL_-1;
    par_grid * isot_elt = NULL;
    par_grid * cont_elt = NULL;
    for (unsigned int k = 0; k < hoct_contour[need_level].size();++k){
        grid_contour & cur_gcont = hoct_contour[need_level][k];
        for (unsigned int i = 0; i < cur_gcont.nei().size(); ++i){
            for (unsigned int j = 0; j < cur_gcont.nei(i).size(); ++j){
                if (cur_gcont.nei(i,j)->type()==ISOT_){
                    isot_elt = cur_gcont.nei(i,j);
                    break;
                }
            } 
        }
    }
    for (unsigned int k = 0; k < hoct_contour[need_level].size();++k){
        grid_contour & cur_gcont = hoct_contour[need_level][k];
        for (int i = 0; i < NEI_C_; ++i){
            for (unsigned int j = 0; j < cur_gcont.nei(i).size(); ++j){
                if (cur_gcont.nei(i,j)->type()==CONT_){
                    cont_elt = cur_gcont.nei(i,j);
                    break;
                }
            } 
        }
    }
    clear_levelids(isot_elt, depth);
    clear_levelids(cont_elt, depth);
    assert(isot_elt->type()==ISOT_);
    assert(cont_elt->type()==CONT_);
    for (int i = 0; i < depth; ++i){
        isot_elt->levelid_assign(i,-2);
        cont_elt->levelid_assign(i,-1);
    }

}

 
/*===========================================================================*/
/* Update the hoct_contour levels greater than 3 after refinement */
void update_contour(par_grid * root, vector<vector<grid_contour> > & 
    hoct_contour, vector<int> & lid, int depth) {

    //assign_levelid_isotcont(hoct_contour, depth);
    
    clear_contour(hoct_contour);
    hoct_contour.resize(depth);
    clear_levelids(root, depth);
    check_levelids(root);

    int level = assign_self_level(root, depth);
    level--;
    update_contour_elts(root, hoct_contour, lid);

}


/*===========================================================================*/
/* Update the hoct_contour levels greater than 3 after refinement */
void update_contour_elts(par_grid * root, vector<vector<grid_contour> > & 
    hoct_contour, vector<int> & lid) {
    
    assert(root->self_level() >= root->act_level());
    
    if (root->act_level() >= (START_LEVEL_-1)){
        par_grid * par_grid = root->par();
        assert(par_grid);
        int begin_level = max((par_grid->self_level()+1), START_LEVEL_);
        for (int i = begin_level; i <= root->self_level(); ++i){
            root->levelid_assign(i, lid[i]);
            hoct_contour[i].push_back(grid_contour(root));
            assert(root==hoct_contour[i][lid[i]].self());
            assert(root==hoct_contour[i][root->levelid(i)].self());
            lid[i]++;
        }
    }
    for (unsigned int i = 0; i < root->int_chi_array().size(); ++i){
        update_contour_elts(root->int_chi_array(i), hoct_contour, lid);
    }
    
}

/*===========================================================================*/
/* Assign self_level for all the elements except for the first three layers
 * of elements */
int assign_self_level(par_grid * root, int depth){

    int size = root->int_chi_array().size();

    if (depth <= START_LEVEL_){
        return -1;
    }
    
    int self_level = MAX_L_;
    if (size == 0 && depth > START_LEVEL_){
        self_level = depth - 1;
        root->self_level_assign(self_level);
        return self_level;
    }  
    for (unsigned int i = 0; i < root->int_chi_array().size(); ++i){
        par_grid * cur_grid = root->int_chi_array(i);
        int chi_level = assign_self_level(cur_grid, depth);
        self_level = min(self_level, chi_level);
    }

    self_level--;
    if (root->act_level() >= (START_LEVEL_-1)){
        root->self_level_assign(self_level);
    }else{
        self_level = root->act_level();
    }
    return self_level;
}


/*===========================================================================*/
/* Assign neighbor information to the hoct_cotnour grid based on the 
 * current and neighbor information in the "upper_level" */
inline void geo_nei_assign(par_grid * curr_grid, par_grid * upper_grid, 
	int dim, int curr_level, int upper_level,
	int gid, vector<vector<grid_contour> > & hoct_contour) {

	vector<par_grid *> & upper_nei_array = 
		hoct_contour[upper_level][upper_grid->levelid(upper_level)].nei(dim);
    assert(upper_grid->levelid(upper_level) < int(hoct_contour[upper_level].size()));
    for(int k = 0; k < int(upper_nei_array.size()); k++) {
		par_grid * upp_nei = upper_nei_array[k];
        assert(upp_nei);
		/* If the upper level neighbor is not the thermal grid type, assign 
         * the same neighbor to current grid */	
		if((upp_nei->type() == CONT_)||(upp_nei->type()==ISOT_)) {
			hoct_contour[curr_level][gid].nei(dim).push_back(upp_nei);
			//upp_nei->levelid_assign(curr_level, upp_nei->levelid(upper_level));
		}/* else if the upp_nei belongs to curr_level, consider upp_nei */
		else if (upp_nei->levelid_check(curr_level)) {
			if (neighbor_check(dim/2, curr_grid, upp_nei)) {
				hoct_contour[curr_level][gid].nei(dim).push_back(upp_nei);
			}
		}/* else: the upp_nei does not belong to curr_level, then consider 
          it children for curr_level */
		else {
			for (int l = 0; l < int(upp_nei->int_chi_array().size()); l++) {
				par_grid * curr_nei = upp_nei->int_chi_array(l);
				if (neighbor_check(dim/2, curr_grid, curr_nei)) {
					hoct_contour[curr_level][gid].nei(dim).push_back(curr_nei);
				}
			}
		}//end else
	}//end for
}


/*===========================================================================*/
/* Generate the AGC structure for one specific level of the grids in 
 * the "hoct_contour" data structure */
void hoct_contour_AGC(vector<grid_contour> & hoct_contour, 
	RC_network_type & AGC, int level){ 

	int grids = hoct_contour.size();
    /* Clear the previous data stored in AGC structure */
	AGC.clear();
	AGC.resize(grids);

	//hoct_interface t_elm;
	for(int i = 0; i < grids; i++) {
		grid_contour& curr_ctour = hoct_contour[i];
		par_grid * curr_grid = curr_ctour.self();
		int nei_cnt = 0;
		int nei_dim = curr_ctour.nei().size();
		for (int nei = 0; nei < nei_dim; ++nei){
			nei_cnt += curr_ctour.nei(nei).size();		
		}
		AGC.C_[i] = cal_cap(curr_grid);
        
		vector<int> & AI = AGC.A_[i];
		vector<double> & GI = AGC.G_[i];
		AI.reserve(nei_cnt + 1);
		GI.reserve(nei_cnt + 1);
		AGC.A_[i].push_back(curr_grid -> levelid(level));
		AGC.G_[i].push_back(0.0);
		int cnt = 0; 
		for(int j = 0; j < nei_dim; j++) {
			int c_nei_cnt = curr_ctour.nei(j).size();
			for (int k = 0; k < c_nei_cnt; ++k){
				cnt ++;
				par_grid * curr_nei = curr_ctour.nei(j, k);
				double inter_cond = 0;
				if(curr_nei -> type() == GRID_) {
					inter_cond = cal_cond(curr_grid, curr_nei, j/2);
					GI[0] += inter_cond;
					AI.push_back(curr_nei->levelid(level));
					GI.push_back(-inter_cond);
				}else {
					if(curr_nei->type() == CONT_) {
					    inter_cond = cal_cond(curr_grid, j/2);
						GI[0] += inter_cond;
						AI.push_back(CONT_ID_);
						GI.push_back(-inter_cond);
					}
					else if(curr_nei -> type() == ISOT_){
						AI.push_back(ISOT_ID_);
						GI.push_back(0.0);
					}
				} 	
			}
		}
	}

	AGC.size_ = static_cast<int>(AGC.A_.size());	
	return;

}

/*============================================================================*/
/* Check whether a grid is the neighbor of the input power layer whose 
 * coordinate in Z direction is "player_zcoor" */
char layer_nei(grid_contour & grid_cont, double player_zcoor){

    par_grid * cur_grid = grid_cont.self();
    char r_char = 'a';    
    if (fabs(player_zcoor + cur_grid->sizes(Z_)/2.0-cur_grid->coors(Z_))<FMIN_){
        r_char = 'u';
    }else{
        if (fabs(player_zcoor-cur_grid->sizes(Z_)/2.0-cur_grid->coors(Z_))<FMIN_){
            r_char = 'd';
        }
    }

    return r_char;
}


/*============================================================================*/
/* Check whether a grid is the neighbor of the input power layer whose 
 * coordinate in Z direction is "player_zcoor" */
char layer_nei(par_grid * grid, double player_zcoor){

    char r_char = 'a';    
    if (fabs(player_zcoor + grid->sizes(Z_)/2.0-grid->coors(Z_))<FMIN_){
        r_char = 'u';
    }else{
        if (fabs(player_zcoor-grid->sizes(Z_)/2.0-grid->coors(Z_))<FMIN_){
            r_char = 'd';
        }
    }

    return r_char;
}


/*===========================================================================*/
/* Apply the correct neighbor information for the "hoct_contour" data 
 * stucture */
int hoct_contour_neighbor(vector<vector<grid_contour> > & hoct_contour) {

	int levels = hoct_contour.size();
    
    /* The paritioning starts to work from level 3 */
	for(int i = START_LEVEL_; i < levels; i++) {
	
		int curr_level = i;
		int upper_level = i-1;
		int grids = hoct_contour[curr_level].size();
	
        /* Grids in the current contour level */
		for(int j = 0; j < grids; j++) {
			par_grid * curr_grid = hoct_contour[curr_level][j].self();
			par_grid * upper_grid = upper_level_grid(curr_grid, upper_level);
			assert(upper_grid->self_level() >= (START_LEVEL_ - 1));
			assert(upper_grid->act_level() >= (START_LEVEL_ - 1));
            if(curr_grid != upper_grid) {
				int * u_p = upper_grid->partition_flag();
				int off_set= 1;
				for(int k = 0; k < DIM_C_; k++) {
					int index =k*2;
					if(u_p[k]) {
						if(curr_grid->coors(k) < upper_grid->coors(k)) {
							geo_nei_assign(curr_grid, upper_grid, index, 
								curr_level, upper_level, j, hoct_contour);
							hoct_contour[curr_level][j].nei(index+1).push_back
								(upper_grid->int_chi_array(curr_grid->id()+off_set));
						}else {
							geo_nei_assign(curr_grid, upper_grid, index+1, 
								curr_level, upper_level, j, hoct_contour);
							hoct_contour[curr_level][j].nei(index).push_back
								(upper_grid->int_chi_array(curr_grid->id()-off_set));
						}
						off_set *= 2;
					}else {
						geo_nei_assign(curr_grid, upper_grid, index, curr_level, 
							upper_level, j, hoct_contour);
						geo_nei_assign(curr_grid, upper_grid, index+1, curr_level, 
							upper_level, j, hoct_contour);
					}
				}//end for
			}else {
				for(int k = 0; k < DIM_C_; k++) {
					int index = k*2;
					geo_nei_assign(curr_grid, upper_grid, index, curr_level,
						upper_level, j, hoct_contour);
					geo_nei_assign(curr_grid, upper_grid, index+1, curr_level,
						upper_level, j, hoct_contour);
				}
			}
		}
	}//end for

    return 0;
}


/*===========================================================================*/
/* Check whether grid a and grid b are neighbors in dimension i */
int neighbor_check(int i, par_grid * a, par_grid * b) {

	for(int j = 0; j < DIM_C_; j++) {
		if(i == j) {
            /* If in the dimension supposed to be neighbor, a and b are not
             * neighbors, return 0 */
			if (fabs(fabs(a->coors(j) - b->coors(j)) - (a->sizes(j) +
			    b->sizes(j))/2) > FMIN_) {
				return 0;
			}
		}else {
            /* If in the dimension not supposed to be neighbor, a and b are 
             * neighbors, also return 0 */
			if ((-fabs(a->coors(j) - b->coors(j)) + (a->sizes(j) +
			    b->sizes(j))/2)<FMIN_) {
				return 0;
			}
		}
	}

	return 1;
}


/*==========================================================================*/
/* Clear the levelids for the grids in levels greater than 3 */
void clear_levelids(par_grid * root, int depth){

    int stop_level = START_LEVEL_;
    if ((root->type()==ISOT_)||(root->type()==CONT_)){
        stop_level = 0;
    }
    
	for (int i = (depth-1); i >= START_LEVEL_; --i){
	//for (int i = tot_levels; i > (START_LEVEL_+1); --i){
        if (root->levelid_check(i)){
            root->levelid().erase(i);
        }
	}

    assert(int(root->levelid().size()) <= START_LEVEL_);
    map<int,int>::iterator iter = root->levelid().begin();
    if (stop_level == 0){
        return;
    }

    for(int i = 0; i < int(root->int_chi_array().size()); i++){
		par_grid * chi = root->int_chi_array(i);
		clear_levelids(chi, depth);
	}

    return;
}


/*==========================================================================*/
/* Calculate the capacitance of grid "gA" */
double cal_cap(par_grid * gA){

    return gA -> ccap() * gA -> sizes(X_) * gA -> sizes(Y_)
        * gA -> sizes(Z_);

}


/*===========================================================================*/
/* Calculate the conductance between grids grid "gA" and grid "gB" in dimenion 
 * "dim" */
double cal_cond(par_grid * gA, par_grid * gB, int dim) {

	double la = 0.0;
	double lb = 0.0;

	double intersect_area = 0.0;
	double gA_area = 0.0;
	double gB_area = 0.0;
	switch (dim) {
	
		case Z_: 
			//gA_area = gA -> sizes(X_) * gA -> sizes(Y_);
			//gB_area = gB -> sizes(X_) * gB -> sizes(Y_);
			intersect_area = gA -> area_intersect(gB); 
			la = gA->sizes(Z_)/2; lb = gB->sizes(Z_)/2;
			break; 
		case Y_: 
			gA_area = gA -> sizes(X_) * gA -> sizes(Z_);
			gB_area = gB -> sizes(X_) * gB -> sizes(Z_);
			intersect_area = min(gA_area, gB_area);
			la = gA->sizes(Y_)/2; lb = gB->sizes(Y_)/2;
			break; 
		case X_: 
			gA_area = gA -> sizes(Y_) * gA -> sizes(Z_);
			gB_area = gB -> sizes(Y_) * gB -> sizes(Z_);
			intersect_area = min(gA_area, gB_area);
			la = gA->sizes(X_)/2; lb = gB->sizes(X_)/2;
			break; 
		default:
			break;
	}

	return (1.0/(la/(gA->cond()*intersect_area) + 
		lb/(gB->cond()* intersect_area)));

}


/*===========================================================================*/
/* Calculate the conductance of grid "gA" in dimension "dim" */
double cal_cond(par_grid * gA, int dim) {

	double area = 0.0;
	double la = 0.0;

	switch (dim) {
		case Z_: 
			area = gA->sizes(X_) * gA->sizes(Y_);
			la = gA->sizes(Z_)/2; 
			break; 
		case Y_: 
			area = gA->sizes(X_) * gA->sizes(Z_);
			la = gA->sizes(Y_)/2; 
			break; 
		case X_: 
			area = gA->sizes(Z_) * gA->sizes(Y_);
			la = gA->sizes(X_)/2;
			break; 
		default:
		break;
	}
    
	return ((gA->cond()*area)/la);
}


/*==========================================================================*/
/* Calculate the equivalent resisistance of two neighbor grids ("gA" and
 * "gB" in Z dimension, in order to calculate the power share when 2D power 
 * is transfered to 3D elements */
pair<double, double> cal_eq_resis(par_grid *gA, par_grid * gB){

	double xyintsect_area = gA -> area_intersect(gB);
	double gA_resis = gA->sizes(Z_)/(2.0*gA->cond() * xyintsect_area);
	double gB_resis = gB->sizes(Z_)/(2.0*gB->cond() * xyintsect_area); 

	pair<double, double> AB_resis = pair<double, double>(gA_resis,
		gB_resis);
	
	return AB_resis;
}


/*==========================================================================*/
/* 2D material layer's z position value is zpos
 * for 3D elements, to check whether it is the lower neighbor of the above 2D
 * material  
 *     2D layer (zpos) 
 *     3D els */
bool is_uppernei(const par_grid * cur_grid, const double & zpos){

	bool is = false;
	
	if (fabs(fabs(cur_grid -> coors(2) - cur_grid -> sizes(2)/2.0) - 
		zpos) < FMIN_) {
		is = true;	
	}

	return is;
}


/*==========================================================================*/
/* 2D material layer's z position value is zpos
 * for 3D elements, to check whether it is the lower neighbor of the above 2D
 * material  
 *     2D layer (zpos) 
 *     3D els */
bool is_lowernei(const par_grid * cur_grid, const double & zpos){

	bool is = false;
	
	if (fabs(fabs(cur_grid -> coors(2) + cur_grid -> sizes(2)/2.0) - 
		zpos) < FMIN_) {
		is = true;	
	}

    return is;
}


/*===========================================================================*/
/* Find the number of level of children for "root" */
int check_depth(par_grid * root) {

	vector<par_grid *> & chis = root -> int_chi_array();
	int sizes = chis.size();
	int depth = 0;
	if (sizes > 0) {
		for(int i = 0; i < sizes; i++) {
			int chi_depth = check_depth(chis[i]);
			if(chi_depth > depth) {
				depth = chi_depth;
			}
		}
	}
	depth ++;
	return depth;
}


/*===========================================================================*/
/* Check the sanity of the levelid for root and all its children */
void check_levelids(par_grid * root){

    int depth = check_depth(root);
    if ((root->type()==CONT_)||(root->type()==ISOT_)){
        return;
    }
    for (int i = START_LEVEL_; i < depth; ++i){
        if (root->levelid_check(i)){
            assert(0);
        }
    }
    for (unsigned int i = 0; i < root->int_chi_array().size();++i){
        check_levelids(root->int_chi_array(i));
    }

}
