/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** input.cc: input file parser
****************************************************************************/

/*
Input format:

comments:
	# (line starts with)


*/
#include "header.h"

//convert the string to an int
bool StringToInt(const string &s, int &i);
////convert the string to an int and cut the int off the string
int return_int_and_cut(string * next_part);
//convert the string to a double and cut the double off the string
double return_doub_and_cut(string * next_part);


/*==========================================================================*/
//For inputing for a single structure
int Device_structure::input(string filename) {

    //Make sure a device file is specified
    if(filename.empty()) {
        cerr << "Error: Device file not specified." << endl;
        exit(1);
    }

    ifstream fin;
    fin.open(filename.c_str()); //input file

    //Make sure the file exists
    if(fin.fail()) {
        cerr << "Error: File " << filename << " not found." << endl;
        exit(1);
    }

    string str, first_part, next_part;

	//check the input file for structure delimiters "%"
    //get information from file until the end of the file
    while(getline(fin,str)) {

		//initialize the strings
        first_part = "";
        next_part = "";

        //find where the first space is in the string
        int whereis = str.find(' ',0);
        //if the first char is a space, it will not be accepted
        if(whereis>0) {
            first_part = str.substr(0,whereis);
            //+1 so it doesnt include the space
            next_part = str.substr(whereis+1,str.length());
        }

		//number of iterations
        if(first_part=="iter") {
			//get the number of iterations for the Fourier and BTE sections
			iter_param = return_int_and_cut(&next_part);
        }
		//initial temperature for the structure
        else if(first_part=="t_ini") {
            substrate.T_ini = return_doub_and_cut(&next_part);
        }
        //factor to maintain stability between BTE and Fourier solvers
        else if(first_part=="k_stab") {
            substrate.k_stab = return_doub_and_cut(&next_part);
        }
		//number of times the mean free path
        else if(first_part=="expand_factor") {
            //get the number of iterations for the Fourier and BTE sections
            bte_region_expand_factor = return_int_and_cut(&next_part);
        }
        //set all cells to bte
        else if(first_part=="all_bte") {
            substrate.all_bte_=1;
        }
        //set all cells to fourier
        else if(first_part=="all_fourier") {
            substrate.all_fourier_=1;
        }
		//transient simulation
        else if(first_part=="transient") {
            substrate.transient_=1;
            //set the step size
            substrate.step_time = return_doub_and_cut(&next_part);
            //set the total simulation time
            substrate.total_time = return_doub_and_cut(&next_part);
        }
		//cells in each direction
        else if(first_part=="cells_x") {
            substrate.f_cells_x=return_int_and_cut(&next_part);
        }
		else if(first_part=="cells_y") {
            substrate.f_cells_y=return_int_and_cut(&next_part);
        }
		else if(first_part=="cells_z") {
            substrate.f_cells_z=return_int_and_cut(&next_part);
        }

		//this indicates the beginning of the structure or material
		else if(str[0]=='%') {
			//check if it is a material
			if(str.size()>9) {
                string tempo = str.substr(2,8);
                if((tempo=="material")||(tempo=="MATERIAL")) {

                    //create a new material and get its information from the
                    //input file
					materials.resize(materials.size()+1);
					
					//set this material's name
					materials[materials.size()-1].name = str.substr(11);	

                    //also make sure there were no errors
                    int verify = materials[materials.size()-1].input(fin);

                    //if there was a failure, indicate it and quit
                    if(verify==1) {
                        return 1;
                    }
                }
            }

			//check if it is the substrate
			if(str.size()>10) {
				string tempo = str.substr(2,9);
				if((tempo=="substrate")||(tempo=="SUBSTRATE")) {

					//get substrate information from the input file
                    //also make sure there were no errors
					int verify = substrate.input(fin);
	
					//verify if the material name specified exists
					int size = materials.size();
					int have_mat = 0; //=1 if the material was defined
					for(int i=0; i<size; i++) {
						if(substrate.material == materials[i].name) {
							have_mat = 1;
							//get all the material information
							substrate.material_full = &(materials[i]);
						}
					}

        	    	//if there was a failure, indicate it and quit
		           	if((verify==1)||(have_mat==0)) {
        		        return 1;
			        }

				}
			}
			//check if it is a device
            if(str.size()>7) {
                string tempo = str.substr(2,6);
                if((tempo=="device")||(tempo=="DEVICE")) {
					//if it is a device get all the pieces of the device

					int verify=0;
					device.resize(device.size()+1);
					int sizer = device.size()-1;

                    //set all boundaries initially to non-oxide
                    oxide_x_0.push_back(0);
                    oxide_x_end.push_back(0);
                    oxide_y_0.push_back(0);
                    oxide_y_end.push_back(0);
                    oxide_z_0.push_back(0);
                    oxide_z_end.push_back(0);

                    trans_factor.resize(trans_factor.size()+1);
                    //set tranmittivity factor for all directions 
                    //to 1.0 as default
                    trans_factor[trans_factor.size()-1].push_back(1.0);
                    trans_factor[trans_factor.size()-1].push_back(1.0);
                    trans_factor[trans_factor.size()-1].push_back(1.0);
                    trans_factor[trans_factor.size()-1].push_back(1.0);
                    trans_factor[trans_factor.size()-1].push_back(1.0);
                    trans_factor[trans_factor.size()-1].push_back(1.0);

					//get the n_angles and heat sources
					get_device_info(fin);

					//verify = 1 if error, or = 2 if end of device
					while(verify==0) {
						//make new structures for the device pieces
						device[sizer].resize(device[sizer].size()+1);

	                    //also make sure there were no errors
    	                verify=device[sizer][device[sizer].size()-1].input(fin);

						//verify if the material name specified exists
        	            int size = materials.size();
    	                int have_mat = 0; //=1 if the material was defined
            	        for(int i=0; i<size; i++) {
                	        if(device[sizer][device[sizer].size()-1].material
														==materials[i].name) {
                        	    have_mat = 1;
                               device[sizer][device[sizer].size()-
                                            1].material_index=i;
                    	    }
	                    }

        	            //if there was a failure, indicate it and quit
            	        if((verify==1)||(have_mat==0)) {
                	        return 1;
                    	}
						
					}
					
                }
            }

        }
    }

    //Do some sanity checks for the input
    if(substrate.all_bte_ == 1 && substrate.all_fourier_ == 1) {
        cerr << "Error: Both \"all_bte\" and \" all_fourier\" are set to 1 simultaneously. Please set at least one of them to 0." << endl;
        exit(1);
    }

	return 0; //indicating it got the input ok

}


/*==========================================================================*/
//For inputing for a single structure from device_structure
int BTE_structure::input(ifstream & fin) {
	//"fin" gives access to the input file

	//mark whether or not to continue the device
	int continue_dev=1;

    string str, first_part,next_part;

    //get information from file until the end of the structure
    while(getline(fin,str)) {
        
		//if the line has a structure delimiter in it, exit the function and
		//stop taking input data
		if(str[0]=='%') {
			//if this is the end of the device, then stop this device
			if(str.size()>11) {
				string usenow = str.substr(2,10);
				if(usenow == "end device") {
					continue_dev = 0;
				}
			}
			
			break;
        }

		//initialize the strings
        first_part = "";
        next_part = "";

        //find where the first space is in the string
        int whereis = str.find(' ',0);
        //if the first char is a space, it will not be accepted
        if(whereis>0) {
            first_part = str.substr(0,whereis);
            //+1 so it doesnt include the space
            next_part = str.substr(whereis+1,str.length());
        }

        //the x,y,z coordinates of the south-west-bottom corner of structure
        if (first_part=="coordinates") {
            //get x coordinate
            x_coor = return_doub_and_cut(&next_part); 
            //get y coordinate
            y_coor = return_doub_and_cut(&next_part);
            //get z coordinate
            z_coor = return_doub_and_cut(&next_part);
        }  
        //the length of the x dimension
        else if(first_part=="lengthx") {
            length_x = return_doub_and_cut(&next_part);
            //get the amount of cells in x
            //cells_x = return_int_and_cut(&next_part);
        }
        else if(str[0]=='#') {
            //dont do anything because it is a comment
        }
        //the length of the y dimension
        else if(first_part=="lengthy") {
            length_y = return_doub_and_cut(&next_part);
            //get the amount of cells in y
            //cells_y = return_int_and_cut(&next_part);
        }
        //the length of the z dimension
        else if(first_part=="lengthz") {
            length_z = return_doub_and_cut(&next_part);
            //get the amount of cells in z
            //cells_z = return_int_and_cut(&next_part);
        }
        //the amount of control angles per octant
        else if(first_part=="nphi") {
            N_phi = return_int_and_cut(&next_part);
        }
        else if(first_part=="ntheta") {
            N_theta = return_int_and_cut(&next_part);
        }
        //volumetric specific heat
        else if(first_part=="cv") {
            cv_ = return_doub_and_cut(&next_part);
        }
        //reference temperature for specific heat
        else if(first_part=="temp") {
            t_ref_ = return_doub_and_cut(&next_part);
        }
        //relaxation time
        else if(first_part=="t_rel") {
            t_rel_ = return_doub_and_cut(&next_part);
        }
        //initial temperature for the structure
        else if(first_part=="t_ini") {
            T_ini = return_doub_and_cut(&next_part);
        }
        //group velocity
        else if(first_part=="vg") {
            vg_ = return_doub_and_cut(&next_part);
        }
        //specular boundary
        else if(first_part=="spec") {
            //the specular axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the specular boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the specular boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    spec_x_0 = 1;
                }
                else if(bound == 1) {
                    spec_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    spec_y_0 = 1;
                }
                else if(bound == 1) {
                    spec_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    spec_z_0 = 1;
                }
                else if(bound == 1) {
                    spec_z_end = 1;
                }
            }

        }
        //constant temperature boundaries
        else if(first_part=="t_bound") {
            //the constant temperature axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(2);

            //the constant temperature boundary
            int bound = return_int_and_cut(&next_part);
            double value = return_doub_and_cut(&next_part);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            if(axis == x_ax) {
                if(bound == 0) {
                    tbound_x_0 = 1;
                    tbx_0_val = value;
                }
                else if(bound == 1) {
                    tbound_x_end = 1;
                    tbx_end_val = value;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    tbound_y_0 = 1;
                    tby_0_val = value;
                }
                else if(bound == 1) {
                    tbound_y_end = 1;
                    tby_end_val = value;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    tbound_z_0 = 1;
                    tbz_0_val = value;
                }
                else if(bound == 1) {
                    tbound_z_end = 1;
                    tbz_end_val = value;
                }
            }

        }
        //diffuse boundary
        else if(first_part=="diff") {
            //the diffuse axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the diffuse boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the diffuse boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    diff_x_0 = 1;
                }
                else if(bound == 1) {
                    diff_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    diff_y_0 = 1;
                }
                else if(bound == 1) {
                    diff_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    diff_z_0 = 1;
                }
                else if(bound == 1) {
                    diff_z_end = 1;
                }
            }

        }
        //get the 3d heat source information
        else if(first_part=="source") {
            int size = sources.size();

            //add one to the amount of sources
            sources.resize(size+1);

            //get the heat generation, Q value
            sources[size].Q_source = return_doub_and_cut(&next_part);

            //get the x boundaries, the y boundaries, then the z boundaries
            //and make sure l1b<l2b
            double l_1 = return_doub_and_cut(&next_part);
            double l_2 = return_doub_and_cut(&next_part);
            sources[size].l1x = min(l_1,l_2);
            sources[size].l2x = max(l_1,l_2);

            //same for 'y'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1y = min(l_1,l_2);
            sources[size].l2y = max(l_1,l_2);

            //same for 'z'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1z = min(l_1,l_2);
            sources[size].l2z = max(l_1,l_2);
        }
		//get the 3d heat source information
        else if(first_part=="trans_source") {
            int size = sources.size();

            //add one to the amount of sources
            sources.resize(size+1);

            //get the heat generation, Q value
            sources[size].Q_source = return_doub_and_cut(&next_part);

            //get the x boundaries, the y boundaries, then the z boundaries
            //and make sure l1b<l2b
            double l_1 = return_doub_and_cut(&next_part);
            double l_2 = return_doub_and_cut(&next_part);
            sources[size].l1x = min(l_1,l_2);
            sources[size].l2x = max(l_1,l_2);

            //same for 'y'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1y = min(l_1,l_2);
            sources[size].l2y = max(l_1,l_2);

            //same for 'z'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1z = min(l_1,l_2);
            sources[size].l2z = max(l_1,l_2);

            //get the time delay
            sources[size].offset_time = return_doub_and_cut(&next_part);

            //get the 'on' time
            sources[size].on_time = return_doub_and_cut(&next_part);
        }
		else if(first_part=="k_cond") {
			//get the thermal conductivity of the material
			k_cond = return_doub_and_cut(&next_part);
		}
		//insulating boundary (Fourier)
        else if(first_part=="ins") {
            //the insulating axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the insulating boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the insulating boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    ins_x_0 = 1;
                }
                else if(bound == 1) {
                    ins_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    ins_y_0 = 1;
                }
                else if(bound == 1) {
                    ins_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    ins_z_0 = 1;
                }
                else if(bound == 1) {
                    ins_z_end = 1;
                }
            }

        }
		//indicate the material of the structure
		else if(first_part=="material") {
			material = next_part;
		}

    }

	//For now, always return a 0, until other indicators are found that can
	//determine whether the input information was found to be accurate

    //if the device should stop then flag it
    if(continue_dev == 0) {
        return 2;
    }
    else
        return 0; //indicating it got the input ok

}

/*==========================================================================*/
//get information that is specific to devices
void Device_structure::get_device_info(ifstream & fin) {
    //"fin" gives access to the input file

    string str, first_part,next_part;
    
    int current_device = trans_factor.size()-1;

    //get information from file until the end of the structure
    while(getline(fin,str)) {

        //if the line has a structure delimiter in it, exit the function and
        //stop taking input data
        if(str[0]=='%') {
            break;
        }

        //initialize the strings
        first_part = "";
        next_part = "";

        //find where the first space is in the string
        int whereis = str.find(' ',0);
        //if the first char is a space, it will not be accepted
        if(whereis>0) {
            first_part = str.substr(0,whereis);
            //+1 so it doesnt include the space
            next_part = str.substr(whereis+1,str.length());
        }

        //the amount of control angles per octant
        if(first_part=="nphi") {
            N_phi.push_back(return_int_and_cut(&next_part));
			//N_phi = return_int_and_cut(&next_part);
        }
        else if(first_part=="ntheta") {
			N_theta.push_back(return_int_and_cut(&next_part));
            //N_theta = return_int_and_cut(&next_part);
        }
        else if(first_part=="trans_factor") {
            //the silicon dioxide axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the silicon dioxide boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";

            next_part = next_part.substr(2);
            //check which axis
            if(axis == x_ax) {
                if(bound == 0) {
                    trans_factor[current_device][0]= 
                        return_doub_and_cut(&next_part);
                }
                else if(bound == 1) {
                    trans_factor[current_device][1]= 
                        return_doub_and_cut(&next_part);
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    trans_factor[current_device][2]= 
                        return_doub_and_cut(&next_part);
                }
                else if(bound == 1) {
                    trans_factor[current_device][3]= 
                        return_doub_and_cut(&next_part);
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    trans_factor[current_device][4]= 
                        return_doub_and_cut(&next_part);
                }
                else if(bound == 1) {
                    trans_factor[current_device][5]= 
                        return_doub_and_cut(&next_part);
                }
            }

        }
        else if(first_part=="oxide") {
            //the silicon dioxide axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the silicon dioxide boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";

            //check which axis the oxide boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    oxide_x_0[current_device]=1;
                }
                else if(bound == 1) {
                    oxide_x_end[current_device]=1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    oxide_y_0[current_device]=1;
                }
                else if(bound == 1) {
                    oxide_y_end[current_device]=1;

                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    oxide_z_0[current_device]=1;

                }
                else if(bound == 1) {
                    oxide_z_end[current_device]=1;
                }
            }

        }
        //get the 3d heat source information
        else if(first_part=="source") {
            int size = sources.size();

            //add one to the amount of sources
            sources.resize(size+1);

            //get the heat generation, Q value
            sources[size].Q_source = return_doub_and_cut(&next_part);

            //get the x boundaries, the y boundaries, then the z boundaries
            //and make sure l1b<l2b
            double l_1 = return_doub_and_cut(&next_part);
            double l_2 = return_doub_and_cut(&next_part);
            sources[size].l1x = min(l_1,l_2);
            sources[size].l2x = max(l_1,l_2);

            //same for 'y'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1y = min(l_1,l_2);
            sources[size].l2y = max(l_1,l_2);

            //same for 'z'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1z = min(l_1,l_2);
            sources[size].l2z = max(l_1,l_2);
        }
		//get the 3d heat source information
        else if(first_part=="trans_source") {
            int size = sources.size();

            //add one to the amount of sources
            sources.resize(size+1);

            //get the heat generation, Q value
            sources[size].Q_source = return_doub_and_cut(&next_part);

            //get the x boundaries, the y boundaries, then the z boundaries
            //and make sure l1b<l2b
            double l_1 = return_doub_and_cut(&next_part);
            double l_2 = return_doub_and_cut(&next_part);
            sources[size].l1x = min(l_1,l_2);
            sources[size].l2x = max(l_1,l_2);

            //same for 'y'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1y = min(l_1,l_2);
            sources[size].l2y = max(l_1,l_2);

            //same for 'z'
            l_1 = return_doub_and_cut(&next_part);
            l_2 = return_doub_and_cut(&next_part);
            sources[size].l1z = min(l_1,l_2);
            sources[size].l2z = max(l_1,l_2);
			
			//get the time delay
            sources[size].offset_time = return_doub_and_cut(&next_part);

			//get the 'on' time
            sources[size].on_time = return_doub_and_cut(&next_part);
        }
	}
}

/*==========================================================================*/
//For inputing for a single Fourier structure
int Fourier_structure::input(ifstream & fin) {
    //"fin" gives access to the input file

    string str, first_part,next_part;

    //get information from file until the end of the structure
    while(getline(fin,str)) {

        //if the line has a structure delimiter in it, exit the function and
        //stop taking input data
        if(str[0]=='%') {
            break;
        }

        //initialize the strings
        first_part = "";
        next_part = "";

        //find where the first space is in the string
        int whereis = str.find(' ',0);
        //if the first char is a space, it will not be accepted
        if(whereis>0) {
            first_part = str.substr(0,whereis);
            //+1 so it doesnt include the space
            next_part = str.substr(whereis+1,str.length());
        }

        //the x,y,z coordinates of the south-west-bottom corner of structure
        if (first_part=="coordinates") {
            //get x coordinate
            x_coor = return_doub_and_cut(&next_part);
            //get y coordinate
            y_coor = return_doub_and_cut(&next_part);
            //get z coordinate
            z_coor = return_doub_and_cut(&next_part);
        }
        //the length of the x dimension
        else if(first_part=="lengthx") {
            length_x = return_doub_and_cut(&next_part);
        }
        else if(str[0]=='#') {
            //dont do anything because it is a comment
        }
        //the length of the y dimension
        else if(first_part=="lengthy") {
            length_y = return_doub_and_cut(&next_part);
        }
        //the length of the z dimension
        else if(first_part=="lengthz") {
            length_z = return_doub_and_cut(&next_part);
        }
        //initial temperature for the structure
        else if(first_part=="t_ini") {
            T_ini = return_doub_and_cut(&next_part);
        }
        //specular boundary
        else if(first_part=="spec") {
            //the specular axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the specular boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the specular boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    spec_x_0 = 1;
					ins_x_0 = 1;
                }
                else if(bound == 1) {
                    spec_x_end = 1;
					ins_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    spec_y_0 = 1;
					ins_y_0 = 1;
                }
                else if(bound == 1) {
                    spec_y_end = 1;
					ins_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    spec_z_0 = 1;
					ins_z_0 = 1;
                }
                else if(bound == 1) {
                    spec_z_end = 1;
					ins_z_end = 1;
                }
            }

        }
        //constant temperature boundaries
        else if(first_part=="t_bound") {
            //the constant temperature axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(2);

            //the constant temperature boundary
            int bound = return_int_and_cut(&next_part);
            double value = return_doub_and_cut(&next_part);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            if(axis == x_ax) {
                if(bound == 0) {
                    tbound_x_0 = 1;
                    tbx_0_val = value;
                }
                else if(bound == 1) {
                    tbound_x_end = 1;
                    tbx_end_val = value;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    tbound_y_0 = 1;
                    tby_0_val = value;
                }
                else if(bound == 1) {
                    tbound_y_end = 1;
                    tby_end_val = value;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    tbound_z_0 = 1;
                    tbz_0_val = value;
                }
                else if(bound == 1) {
                    tbound_z_end = 1;
                    tbz_end_val = value;
                }
            }
        }
        //diffuse boundary
        else if(first_part=="diff") {
            //the diffuse axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the diffuse boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the diffuse boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    diff_x_0 = 1;
					ins_x_0 = 1;
                }
                else if(bound == 1) {
                    diff_x_end = 1;
					ins_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    diff_y_0 = 1;
					ins_y_0 = 1;
                }
                else if(bound == 1) {
                    diff_y_end = 1;
					ins_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    diff_z_0 = 1;
					ins_z_0 = 1;
                }
                else if(bound == 1) {
                    diff_z_end = 1;
					ins_z_end = 1;
                }
            }

        }
        //insulating boundary (Fourier)
        else if(first_part=="ins") {
            //the insulating axis
            string axis = next_part.substr(0,1);
            next_part = next_part.substr(1);
            //the insulating boundary
            int bound;
            //convert the string to an int
            StringToInt(next_part, bound);

            string x_ax = "x";
            string y_ax = "y";
            string z_ax = "z";
            //check which axis the insulating boundaries are and make note
            if(axis == x_ax) {
                if(bound == 0) {
                    ins_x_0 = 1;
                }
                else if(bound == 1) {
                    ins_x_end = 1;
                }
            }
            else if(axis == y_ax) {
                if(bound == 0) {
                    ins_y_0 = 1;
                }
                else if(bound == 1) {
                    ins_y_end = 1;
                }
            }
            else if(axis == z_ax) {
                if(bound == 0) {
                    ins_z_0 = 1;
                }
                else if(bound == 1) {
                    ins_z_end = 1;
                }
            }

        }
        //indicate the material of the structure
        else if(first_part=="material") {
            material = next_part;
        }
    }

    //For now, always return a 0, until other indicators are found that can
    //determine whether the input information was found to be accurate

   	return 0; //indicating it got the input ok

}
  
/*==========================================================================*/
//For inputing for a single material
int Material::input(ifstream & fin) {
    //"fin" gives access to the input file

    string str, first_part,next_part;

    //get information from file until the end of the structure
    while(getline(fin,str)) {

        //if the line has a structure delimiter in it, exit the function and
        //stop taking input data
        if(str[0]=='%') {
            break;
        }

        //initialize the strings
        first_part = "";
        next_part = "";

        //find where the first space is in the string
        int whereis = str.find(' ',0);
        //if the first char is a space, it will not be accepted
        if(whereis>0) {
            first_part = str.substr(0,whereis);
            //+1 so it doesnt include the space
            next_part = str.substr(whereis+1,str.length());
        }
		
		//volumetric specific heat
        if(first_part=="cv") {
            cv_ = return_doub_and_cut(&next_part);
        }
        //reference temperature for specific heat
        else if(first_part=="temp") {
            t_ref_ = return_doub_and_cut(&next_part);
        }
        //relaxation time
        else if(first_part=="t_rel") {
            t_rel_ = return_doub_and_cut(&next_part);
        }
        //group velocity
        else if(first_part=="vg") {
            vg_ = return_doub_and_cut(&next_part);
        }
		else if(first_part=="k_cond") {
            //get the thermal conductivity of the material
            k_cond = return_doub_and_cut(&next_part);
        }
	}
	
	return 0; //indicating it got the input ok

}
	

/*==========================================================================*/
int return_int_and_cut(string * next_part) {
	int whereis; //to find where the next space is
	string temp; //temporary string
	int value;	//the desired value in the string

	whereis = (*next_part).find(' ',0);
    //get the next value in the string
    temp = (*next_part).substr(0,whereis);
    //convert the string to an int
	StringToInt(temp, value);
	//cut the string to find the next part
    *next_part = (*next_part).substr(whereis+1,(*next_part).length());	
	
	return value;
}

/*==========================================================================*/
double return_doub_and_cut(string * next_part) {
	int whereis; //to find where the next space is
	string temp; //temporary string
	double value;  //the desired value in the string
     
	whereis = (*next_part).find(' ',0);
    //get the next value in the string
    temp = (*next_part).substr(0,whereis);
    //convert the string to a double
	value = strtod((*next_part).c_str(),NULL);
    //cut the string to find the next part
    *next_part = (*next_part).substr(whereis+1,(*next_part).length());
     
    return value;
}

/*==========================================================================*/
bool StringToInt(const string &s, int &i){
  istringstream myStream(s);
  
  if (myStream>>i)
    return true;
  else
    return false;
}

