/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** isac.cc: code for the ISAC class
****************************************************************************/

#include <cstdlib>
#include "isac.h"
#include "util.h"
#include "sol.h"
#include "configuration.h"

using namespace isac;
using namespace std;

//*==========================================================================*/
//* Form the object of isac class based on the input chip package
// * ("chippkg"), floorplan("flrplan") and boundary condition ("boundary").
// * Output: 
// *      "chip3D": Object of Chippkg_type class. 
// * Input: 
// *      "chippkg": Chip package definition;
// *      "flrplan": floorplan;
// *      "boundary": boundary condition definition. */
ISAC form_isac(ChipPkg_type & chip3D, const vector<Layer_type> & chippkg, 
	const vector<vector<FlrplanElt_type> > & flrplan, const Boundary & boundary){

    parser(chip3D.components, chippkg, flrplan);

    ISAC isac(chip3D, boundary);
    check_input(chip3D);

	isac.retn_solution()->form_solution(boundary.ambient_resis,
        boundary.ambient_heatcap);
		
	return isac;
}


                                                                               
//*==========================================================================*/
//* Form the object of isac class based on the input chip package file 
// * ("chippkg_file"), floorplan file ("floorplan_file") and 
// * and boundary condition ("boundary").
// * Output: 
// *      "chip3D": Object of Chippkg_type class. 
// * Input: 
// *      "chippkg_file": Chip package definition file;
// *      "floorplan_file": floorplan file;
// *      "boundary": boundary condition definition. */
ISAC form_isac(ChipPkg_type & chip3D, const char * chippkg_file, 
	const char * floorplan_file, const Boundary & boundary){

    parser(chip3D.components, chippkg_file, floorplan_file);

    ISAC isac(chip3D, boundary);
    check_input(chip3D);

	isac.retn_solution()->form_solution(boundary.ambient_resis,
        boundary.ambient_heatcap);
		
	return isac;
}

//*===========================================================================*/
//* Generate the object of Solution class based on the input chip package 
// * information ("chip_3D").
// * Input:
// *      "chip3D": Object of Chippkg_type class. */ 
Solution * ISAC::build_solution(const ChipPkg_type & chip_3D){
    
    return new Solution(chip_3D);
}

//*===========================================================================*/
//* Do the steady-state thermal simulation. 
// * Calculate the steady thermal profile for the given power profile
// * ("power_profile") when time proceeds to infinity. 
// * Input:
// *      "power_profile": power profile. */
vector<Element_type> ISAC::solve_static
    (const vector<Element_type> & power_profile){

    vector<vector<double> > power;
    power.resize(1);    //For now, assume a single power layer

    for(unsigned int i = 0; i < power_profile.size(); i++) {
        power[0].push_back(power_profile[i].value);
    }

    vector<Element_type> thermal_profile;
	double err_thres = 1E-6;
    thermal_profile = solution_->static_solver(power, err_thres,
        boundary_.ambient_resis, boundary_.ambient_heatcap); 

    //* Adjust the thermal profile according to the boundary information. */
	temp_adj(thermal_profile, boundary_);	

    return thermal_profile;
}

//*===========================================================================*/
//* Do the steady-state thermal simulation. 
// * Calculate the steady thermal profile for the given power profile
// * ("power") when time proceeds to infinity. 
// * Input:
// *      "power": power profile. */
vector<Element_type> ISAC::solve_static
    (const vector<vector<double> > & power){

    vector<Element_type> thermal_profile;
	double err_thres = 1E-6;
    thermal_profile = solution_->static_solver(power, err_thres,
        boundary_.ambient_resis, boundary_.ambient_heatcap); 

    //* Adjust the thermal profile according to the boundary information. */
	temp_adj(thermal_profile, boundary_);	

    return thermal_profile;
}


//*===========================================================================*///*
//Do the steady-state thermal simulation.
// * Calculate the steady thermal profile for the given power profile
// * ("power") when time proceeds to infinity.
// * Input:
// *      "power": power profile. */
vector<Element_type> ISAC::solve_static_adaptive
    (const vector<vector<double> > & power){

    vector<Element_type> thermal_profile;
    double err_thres = 1E-6;
    thermal_profile = solution_->static_solver_adaptive(power, err_thres,
        boundary_.ambient_resis, boundary_.ambient_heatcap);

    //* Adjust the thermal profile according to the boundary information. */
    temp_adj(thermal_profile, boundary_);

    return thermal_profile;
}

//*===========================================================================*/
//* Initial dynamic simulation.
// * It actually calls the steady-state simulation function, and its output 
// * thermal profile serves as the initial condition for later dynamic 
// * simulation. 
// * Input:
// *      "power": power profile. */
vector<Element_type> ISAC::init_dynamic
    (const vector<vector<double> > & power){ 

    //* Check the validity of the input power vector. */
    check_inputpower(power);

    vector<Element_type> thermal_profile;
	double err_thres = 1E-6;
	thermal_profile = solution_->static_solver(power, err_thres,
        boundary_.ambient_resis, boundary_.ambient_heatcap);

    //* Adjust the thermal profile according to the boundary information. */
	temp_adj(thermal_profile, boundary_);	

    return thermal_profile;
}

//*===========================================================================*/
//* Dynamic simulation function. 
// * Calculate the thermal profile when time duration is "duration". 
// * input: 
// *      "duration": The simulation time of the dynamic simulation; 
// *      "power_profile": power profile;
vector<Element_type> ISAC::step_dynamic(
    const vector<Element_type> & power_profile, double duration) {

    vector<vector<double> > power;
    power.resize(1);    //For now, assume a single power layer

    for(unsigned int i = 0; i < power_profile.size(); i++) {
        power[0].push_back(power_profile[i].value);
    }

    vector<vector<double> > all_temp;
    double tempreport_interval = 1E-6;
    char select_domain = configuration::ap().dyn_solver_type();

    check_inputpower(power);
    vector<Element_type> thermal_profile;
	thermal_profile = solution_->dynamic_solver(all_temp, duration, 
		tempreport_interval, power, select_domain);	
	
	for (int i = 0; i < int(all_temp.size()); ++i){
		for (int j = 0; j < int(all_temp[i].size()); ++j){
			all_temp[i][j] += boundary_.ambient_temp;
		}
	}
	
	temp_adj(thermal_profile, boundary_);	

    return thermal_profile;
}

//*===========================================================================*/
//* Dynamic simulation function. 
// * Calculate the thermal profile when time duration is "duration". 
// * input: 
// *      "duration": The simulation time of the dynamic simulation; 
// *      "power_profile": power profile;
vector<Element_type> ISAC::step_dynamic(
    const vector<vector<double> > & power, double duration) {

    vector<vector<double> > all_temp;
    double tempreport_interval = 1E-6;
    char select_domain = configuration::ap().dyn_solver_type();

    check_inputpower(power);
    vector<Element_type> thermal_profile;
    thermal_profile = solution_->dynamic_solver(all_temp, duration, 
        tempreport_interval, power, select_domain); 
    
    for (int i = 0; i < int(all_temp.size()); ++i){
        for (int j = 0; j < int(all_temp[i].size()); ++j){
            all_temp[i][j] += boundary_.ambient_temp;
        }
    }
    
    temp_adj(thermal_profile, boundary_);   

    return thermal_profile;
}

//*===========================================================================*/
//* Dynamic simulation function. 
// * Calculate the thermal profile when time duration is "duration". 
// * Output: 
// *      "all_temp": Output temperature vectors. 
// *          The ratio ("duration/tempreport_interval") determins the number  
// *          of the output temperature vectors.
// * input: 
// *      "duration": The simulation time of the dynamic simulation; 
// *      "tempreport_interval": The interval the tempeartures are reported;
// *      "power": power profile;
// *      "select_domain": choose whether to use time_domain solver or 
//*          frequency domain solver. */
vector<Element_type> ISAC::step_dynamic(vector<vector<double> > & all_temp,
	double duration, double tempreport_interval, 
    const vector<vector<double> > & power, char select_domain) {
	
    check_inputpower(power);
    vector<Element_type> thermal_profile;
	thermal_profile = solution_->dynamic_solver(all_temp, duration, 
		tempreport_interval, power, select_domain);	
	
	for (int i = 0; i < int(all_temp.size()); ++i){
		for (int j = 0; j < int(all_temp[i].size()); ++j){
			all_temp[i][j] += boundary_.ambient_temp;
		}
	}
	
	temp_adj(thermal_profile, boundary_);	

    return thermal_profile;
}

//*==========================================================================*/
//Zoom in a box in the chip defined by four edges and an active layer
//and perform steady-state simulation. The active layers are in the same order
//as the chip package file (0 for first power layer, 1 for second power layer,
//etc.).
vector<Element_type> ISAC::zoom_and_solve_static(double x_min, double x_max,
    double y_min, double y_max, const vector<vector<double> > & power, 
    double plyr_z) {

    //Zoom in the box
    solution_->zoom_box(x_min, x_max, y_min, y_max, plyr_z, power);

    vector<Element_type> thermal_profile;

    //Call solve static
    thermal_profile = solve_static(power);

    //Get the elements only inside the box
    //filter_elements_in_box(thermal_profile);

    return thermal_profile;
    
}

//*==========================================================================*/
//Zoom in a box in the chip defined by four edges and an active layer
//and perform dynamic simulation. The active layers are in the same order
//as the chip package file (0 for first power layer, 1 for second power layer,
//etc.).
vector<Element_type> ISAC::zoom_and_step_dynamic(double x_min, double x_max,
    double y_min, double y_max,const vector<vector<double> > & power, 
    double duration, double plyr_z) {

    //Zoom in the box
    solution_->zoom_box(x_min, x_max, y_min, y_max, plyr_z, power);

    vector<Element_type> thermal_profile;

    //Call solve static
    thermal_profile = step_dynamic(power, duration);

    return thermal_profile;
    
}

//*==========================================================================*/
//Report the temperatures of all devices inside a box
std::vector<Device_type> ISAC::report_device_temperatures(
    vector<Element_type> & thermal_profile, string layoutfilename,
    double x_min, double x_max, double y_min, double y_max, double plyr_z) {

    //Open the layout file
    if(layoutfilename.empty()) {
        cerr << "Error: No layout file provided" << endl;
        exit(1);
    }
    ifstream layoutfile;
    layoutfile.open(layoutfilename.c_str());
    if(layoutfile.fail()) {
        cout << "Error: Layout file \"" << layoutfilename << "\" not found.";
        cout << endl;
        exit(1);
    }

    vector<Device_type> chip_devices;
    Device_type device;

    while(layoutfile >> device.loc[X_]) {
        layoutfile >> device.loc[Y_];
        layoutfile >> device.loc[Z_];
        layoutfile >> device.power;
        chip_devices.push_back(device);
    }
    
    //Find the devices inside the box
    vector<Device_type> devices_in_box;
    int num_devices = chip_devices.size();
    for(int i = 0; i < num_devices; i++) {
        if(chip_devices[i].loc[X_] >= x_min && 
        chip_devices[i].loc[X_] <= x_max && chip_devices[i].loc[Y_] >= y_min &&
        chip_devices[i].loc[Y_] <= y_max) {
            
            devices_in_box.push_back(chip_devices[i]);
        }
    }

    //Filter out the elements in thermal_profile by only taking elements inside 
    //the desired box. This saves search time in the next step.
    vector<Element_type> thermal_profile_box;
    int num_elements = thermal_profile.size();
    for(int i = 0; i < num_elements; i++) {
        if(areasIntersect(x_min, x_max, y_min, y_max, 
        thermal_profile[i].center[X_] - thermal_profile[i].size[X_]/2.0,
        thermal_profile[i].center[X_] + thermal_profile[i].size[X_]/2.0,
        thermal_profile[i].center[Y_] - thermal_profile[i].size[Y_]/2.0,
        thermal_profile[i].center[Y_] + thermal_profile[i].size[Y_]/2.0)) {
            //Assuming the power layer is below the element
            if(thermal_profile[i].center[Z_] - thermal_profile[i].size[Z_]/2.0 -            plyr_z < FMIN_) {
                thermal_profile_box.push_back(thermal_profile[i]);
            }
        }
    }
    
    //Now for each device in the box, find the thermal element it belongs to,
    //and compute its temperature
    num_devices = devices_in_box.size();
    num_elements = thermal_profile_box.size();

    for(int i = 0; i < num_devices; i++) {
        for(int j = 0; j < num_elements; j++) {
            if(devices_in_box[i].loc[X_] >= thermal_profile_box[j].center[X_] - 
            thermal_profile_box[j].size[X_]/2.0 &&
            devices_in_box[i].loc[X_] <= thermal_profile_box[j].center[X_] +
            thermal_profile_box[j].size[X_]/2.0 &&
            devices_in_box[i].loc[Y_] >= thermal_profile_box[j].center[Y_] - 
            thermal_profile_box[j].size[Y_]/2.0 &&
            devices_in_box[i].loc[Y_] <= thermal_profile_box[j].center[Y_] + 
            thermal_profile_box[j].size[Y_]/2.0) {
        
                devices_in_box[i].temp = thermal_profile_box[j].value; 
                break;
            }
        }
    }

    //Now add the BTE temperature using the value from the lookup-table file
    ifstream LUT;
    string LUTfile = configuration::ap().LUT_file();
    if(LUTfile.empty()) {
        cout << "Warning: No lookup-table file provided. Temperatures reported will be based on the Fourier model and not the BTE." << endl;
    }
    else {
        LUT.open(LUTfile.c_str());
        if(LUT.fail()) {
            cout << "Error: Lookup-table file \"" << LUTfile << "\" not found.";
            cout << endl;
            exit(1);
        }
        //Read the thermal resistance values
        double BTE_res, Fourier_res;

        LUT >> BTE_res;
        LUT >> Fourier_res;

        //Now adjust the temperatures of the devices with the BTE temperatures
        for(int i = 0; i < num_devices; i++) {
            double BTE_temp = devices_in_box[i].power * BTE_res;
            double Fourier_temp = devices_in_box[i].power * Fourier_res;
            devices_in_box[i].temp -= Fourier_temp;
            devices_in_box[i].temp += BTE_temp;
        }
    }

    return devices_in_box;

}

//*===========================================================================*/
//* Map temperature from 3D elements to 2D elements (such as elements with the 
// * same geometry information as the power profile elements).
// * Output & input:
// *      "temp_profile": Temperature profile.
// * Input: 
// *      "power": Power profile. */
void ISAC::Map_temperature(vector<Element_type> & temp_profile,
    const vector<vector<double> > & power){
    
    vector<double> cond_vec;
    cond_vec.resize(temp_profile.size());
    for (unsigned int i = 0; i < cond_vec.size();++i){
        cond_vec[i] = 0.0;
    }
    //* Convert to correct values when unit for length is millimeter */
    /*
    for (unsigned int i = 0; i < temp_profile.size(); ++i){
        for (int j = 0; j < DIM_C_; ++j){
            temp_profile[i].size[j] *= 1E3;
            temp_profile[i].center[j] *= 1E3;
        }
    }
    */
    
    Map_basic_temperature(temp_profile, cond_vec);
    Map_lumped_temperature(temp_profile, cond_vec, power); 
	temp_adj(temp_profile, boundary_);	

    //* Convert back when unit for length is meter */
    /*
    for (unsigned int i = 0; i < temp_profile.size(); ++i){
        for (int j = 0; j < DIM_C_; ++j){
            temp_profile[i].size[j] *= 1E-3;
            temp_profile[i].center[j] *= 1E-3;
        }
    }
    */

}


//*===========================================================================*/
//* Subroutine for the Map_temperature function call(). */
void ISAC::Map_basic_temperature(vector<Element_type> & temp_profile,
    vector<double> & cond_vec){

    assert(temp_profile.size()>0);
    double zcoor = temp_profile[0].center[Z_];
    int level = (solution_->hoct_contour().size()-1);
    vector<double> temp = solution_->temp();
    vector<grid_contour> & grid_cont = solution_->hoct_contour(level);
    
    for (unsigned int i = 0; i < grid_cont.size(); ++i){ 
        char p_nei = layer_nei(grid_cont[i], zcoor);
        int t_index = grid_cont[i].self()->levelid(level);
        if (p_nei == 'u'){
            for (unsigned int j = 0; j < temp_profile.size(); ++j){
                double area_overlap = 
                    area_intersect(grid_cont[i].self(), temp_profile[j]);
                double telt_area = 
                    temp_profile[j].size[X_] * temp_profile[j].size[Y_];
                double area_share = area_overlap/telt_area;
                temp_profile[j].value += temp[t_index]*area_share;
                cond_vec[j] += grid_cont[i].self()-> cond() *
                    area_overlap/(grid_cont[i].self()->sizes(Z_)/2.0); 
            }
        }
    }
    /*
    for (unsigned int i = 0; i < grid_cont.size(); ++i){
        cout << "temp "<<temp_profile[i].value << endl;
    } 
    */
} 
  

//*===========================================================================*/
//* Subroutine for the Map_temperature function call(). */
void ISAC::Map_lumped_temperature(vector<Element_type> & temp_profile,
    const vector<double> & cond_vec, const vector<vector<double> > & power){

    int player_num = 0;
    vector<pair<int,int> > player_index;
    vector<int> power_index;
    for (unsigned int i = 0; i < solution_->chip()->components.size(); ++i){
        for (unsigned int j =0; j<solution_->chip()->components[i].layers.size(); 
            ++j){
            if (solution_->chip()->components[i].layers[j].property == POWER_){
                player_index.push_back(pair<int,int>(i,j));
                power_index.push_back(player_num);
                player_num ++;
            }
        }
    }
    assert(player_num>0);
    
    for (unsigned int i = 0; i < temp_profile.size(); ++i){
        double pshare = 0.0;
        for (unsigned int j = 0; j < player_index.size(); ++j){ 
            int comp_id = player_index[j].first;
            int layer_id = player_index[j].second;
            double player_zcoor = 
                solution_->chip()->
                    components[comp_id].layers[layer_id].coor[Z_];
            if (!(player_zcoor > temp_profile[i].center[Z_])){
                Layer_type & cur_player = 
                    solution_->chip()->components[comp_id].layers[layer_id];
                for (unsigned int k = 0;k<cur_player.power_profile.size();++k){
                    double area_overlap = area_intersect
                        (temp_profile[i],cur_player.power_profile[k]);
                    double pelt_area = cur_player.power_profile[k].size[X_] *
                        cur_player.power_profile[k].size[Y_];
                    double area_share = area_overlap/pelt_area;
                    pshare += power[power_index[j]][k] * area_share; 
                }
            } 
        }
        temp_profile[i].value += pshare/cond_vec[i];
    }
    
}


//*===========================================================================*/
//* Check the input power vector, to see whether it complies with the 
//* information in the floorplan file. */
void ISAC::check_inputpower(const vector<vector<double> > & power){

    int players_inchip = 0;
    int players_inpower = power.size();
    ChipPkg_type & chip = (ChipPkg_type &)(*solution_->chip());

    for (int i = 0; i < static_cast<int>(chip.components.size()); ++i){
        for (int j = 0; j < static_cast<int>(chip.components[i].
            layers.size()); ++j){
            if (chip.components[i].layers[j].property == POWER_){
                players_inchip ++;
                assert(players_inpower >= players_inchip);
                int pelts = chip.components[i].layers[j].power_profile.size();
                int input_pelts = power[players_inchip-1].size();
                assert(pelts == input_pelts);
            }
        }
    }

    assert(players_inpower == players_inchip);
}

//*==========================================================================*/
//* Constructor for FlrplanElt_type class. */
FlrplanElt_type::FlrplanElt_type():
    x_min(),
    x_max(),
    y_min(),
    y_max()
{

}


//*==========================================================================*/
//* Constructor for FlrplanElt_type class. */
FlrplanElt_type::FlrplanElt_type(double xmin, double xmax, double ymin, double ymax):
    x_min(xmin),
    x_max(xmax),
    y_min(ymin),
    y_max(ymax)
{

}


//*===========================================================================*/
//* Constructor for Element_type class. */
Element_type::Element_type():
	size(),
	center(),
	value(0)
{
	value = 0.0;
}


//*===========================================================================*/
//* Constructor for Element_type class. */
Element_type::Element_type(double size_x, double size_y, double center_x,
double center_y, double value_a) :
	size(),
	center(),
	value(value_a)
{
	size[0] = size_x;
	size[1] = size_y;
	size[2] = 0.0;

	center[0] = center_x;
	center[1] = center_y;
	center[2] = 0.0;
}


//*===========================================================================*/
//* Constructor for Element_type class. */
Element_type::Element_type(double size_x, double size_y, double size_z, double 
    center_x, double center_y, double center_z, double value_a) :
	size(),
	center(),
	value(value_a)
{
	size[0] = size_x;
	size[1] = size_y;
	size[2] = size_z;
	center[0] = center_x;
	center[1] = center_y;
	center[2] = center_z;
}

//*===========================================================================*/
//* Constructor for Device_type class. */
Device_type::Device_type() 
{
    temp = 0.0;
}

//*===========================================================================*/
//* Constructor for Device_type class. */
Device_type::Device_type(double loc_x, double loc_y,
    double loc_z, double power_val, double temp_val)
{
    loc[0] = loc_x;
    loc[1] = loc_y;
    loc[2] = loc_z;
    power = power_val;
    temp = temp_val;
}


//*===========================================================================*/
//* Constructor for Boundary class. */
Boundary::Boundary(){

}
 

//*===========================================================================*/
//* Constructor for Boundary class. */
Boundary::Boundary(double a_t, double a_r, double a_hc) :
	ambient_temp(a_t),
    // The length unit of the input is meter, and unit is converted to 
    // milimeter to be compatible with the inside code.
	ambient_resis(a_r * 1E6),
	ambient_heatcap(a_hc * 1E-6)
{}

Boundary Boundary::cu45deg(318.15, 3.6E-4, 3.9E4);
Boundary Boundary::cu25deg(298.15, 3.6E-4, 3.9E4);
Boundary Boundary::cu00deg(273.15, 3.6E-4, 3.9E4);


//*===========================================================================*/
//* Constructor for ChipPkg_type class. */
ChipPkg_type::ChipPkg_type():
    components(),
    isot(),
    cont(),
    init_partition(){

    cont.id_assign(CONT_ID_);
    isot.id_assign(ISOT_ID_);
}


//*===========================================================================*/
//* Constructor for Layer_type class. */
Layer_type::Layer_type():

property(),
size(),
coor(),
par(),
cond(),
cap(),
homog_cond(true),
homog_cap(true)
{

}

//*===========================================================================*/
//* Adjust the absoluate temperature value of the thermal profile according 
// * to the boundary condition. */
void temp_adj(vector<Element_type> & thermal_profile, 
	const Boundary & boundary){

	int ther_elts = thermal_profile.size();

	for (int i = 0; i < ther_elts; ++i){
		thermal_profile[i].value += boundary.ambient_temp;
	}
}


//*===========================================================================*/
//* Adjust the absolute temperature value of the vector "temp" based on the 
// * boundary condition. */
void temp_adj(vector<double> & temp, const Boundary & boundary){

	for (int i = 0; i < static_cast<int>(temp.size()); ++i){
		temp[i] += boundary.ambient_temp;
	}
}


//*===========================================================================*/
//* Check the sanity of the input chip package description. */
void check_input(const ChipPkg_type & chip_3D){

	int comp_num = chip_3D.components.size();
	double ozpos = 0.0;
	double zpos = 0.0;
	for (int i = 0; i < comp_num; ++i){
		const Component_type & cur_comp = chip_3D.components[i];
		int layer_num = cur_comp.layers.size();
        double sizex = 0.0; double sizey = 0.0;
        for (int j = 0; j < layer_num; j++){
            if (cur_comp.layers[j].property == POWER_){
                continue;
            }
		    sizex = cur_comp.layers[j].size[X_];
		    sizey = cur_comp.layers[j].size[Y_];
        }
		for (int j = 0; j< layer_num; ++j){
            if (cur_comp.layers[j].property == POWER_){
                continue;
            }
			double cur_sizex = cur_comp.layers[j].size[X_];
			double cur_sizey = cur_comp.layers[j].size[Y_];
			if ((fabs(sizex - cur_sizex) > FMIN_) ||
				(fabs(sizey - cur_sizey) > FMIN_)){
				cerr << "The x or y size for the layers in one component ";
				cerr <<	"does not equal to each other " << endl; 
				exit(1);
			} 
		}
		for (int j = 0; j< layer_num; ++j){
			zpos += cur_comp.layers[j].size[Z_];
			double zcoor = cur_comp.layers[j].coor[Z_];
			double sup_zcoor = ozpos + (zpos - ozpos )/2.0;
			if (fabs(sup_zcoor - zcoor) > FMIN_){
				cerr << "The size and/or coordinate in z dimension of layer ";
				cerr << j << " in component " <<i<< " is not correct "<<endl;
				exit(1);
			} 
			ozpos = zpos;
		}
		int power_num = 0;
		for (int j = 0; j< layer_num; ++j){
			if (cur_comp.layers[j].property == POWER_){
				power_num ++;
			}
			if (power_num > 1){
				cerr<<"There are more than one power layers inside component ";
				cerr << i << endl;
				exit(1);
			}
		}
		for (int j = 0; j< layer_num; ++j){
			if (cur_comp.layers[j].property == POWER_){	
				int pelts = cur_comp.layers[j].power_profile.size();
				if (pelts == 0){
					cerr << "The power layer has 0 power consumption elements " 
                        << endl;
					exit(1);
				}
				const vector<Element_type> & power_layer = 
                    cur_comp.layers[j].power_profile;
				for (int k = 0; k < pelts; ++k){
					if (power_layer[k].center[Z_] 
                        != cur_comp.layers[j].coor[Z_]){
						cerr << "The z coordinate of the power elment " << k ;
						cerr << " does not equal to that of the power layer" 
                            << endl;
						exit(1);
					}
				}	
			}
		}	
	}

}


//*===========================================================================*/
//* Copy information from "src_layer" to "dst_layer". */
void copy_layer(Layer_type & dst_layer, const Layer_type src_layer){

	for (int i = 0; i < DIM_C_; ++i){
		dst_layer.coor[i] = src_layer.coor[i];
		dst_layer.size[i] = src_layer.size[i];
	}
	dst_layer.cond = src_layer.cond;
	dst_layer.cap = src_layer.cap;
	int layer_num = src_layer.power_profile.size();
	dst_layer.power_profile.resize(layer_num);
	for (int i = 0; i < layer_num; ++i){ 
		copy_elts(dst_layer.power_profile[i], src_layer.power_profile[i]);
	}	
}

//*===========================================================================*/
//* Copy information from "src_elt" to "dst_elt". */
void copy_elts(Element_type & dst_elt, const Element_type src_elt){

	for (int i = 0; i < DIM_C_; ++i){
		dst_elt.center[i] = src_elt.center[i];
		dst_elt.size[i] = src_elt.size[i];
	}
	dst_elt.value = src_elt.value;
}

/*==========================================================================*/
/* Constructor for ISAC class */
ISAC::ISAC():

solution_(),
boundary_(){

}

/*===========================================================================*/
/* Constructor for ISAC class */
ISAC::ISAC(ChipPkg_type & chip_3D, const Boundary & boundary):
	boundary_(boundary)
{

    /* This is the initial partition */
    //gene_input(chip_3D);
	gene_partition(chip_3D, chip_3D.components);
    
    solution_ = build_solution(chip_3D);
}

/*===========================================================================*/
/* Convert a vector of Element_type to a vector of Map_block */
vector<Map_block> toMap(vector<Element_type> & a) {

    vector<Map_block> b;

    int size = a.size();
    b.resize(size);
    for(int i = 0; i < size; i++) {
        b[i].size[0] = a[i].size[0];
        b[i].size[1] = a[i].size[1];
        b[i].center[0] = a[i].center[0];
        b[i].center[1] = a[i].center[1];
        b[i].value = a[i].value;
    }
    return b;
}
