/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** isac.h: the ISAC class
****************************************************************************/

#ifndef ISAC_H_
#define ISAC_H_

#include <memory>
#include <algorithm>
#include <iterator>
#include <vector>
#include <iostream>
#include "hoct_model.h" 

namespace isac {

class ISAC;
class Solution;
class FlrplanElt_type;
class Element_type;
class Device_type;
class Component_type;
class Layer_type;
class ChipPkg_type;
class Boundary;

//* Units:
//*   W-watt
//*   m-meter
//*   K-Kelvin temperature
//*   J-Joule
//*   Kg-Kilogram */

//*===========================================================================*/
//* Consctructor of Boudary class. */
//* "ambient_temp" is the ambient temperature, "ambient_resist" is ambient 
//* resistivity in K-m^2/W of convection effect, while "abient_heatcap" 
//* is the heat capacitivity in J/(m^2-K) between the heatsink and the 
//* enviorment. */
struct Boundary {
    Boundary();
    Boundary(double ambient_temp, double ambient_resistance, double
        ambient_heatcap);

	static Boundary cu45deg;
	static Boundary cu25deg;
    static Boundary cu00deg;

	double ambient_temp;
	double ambient_resis;
	double ambient_heatcap;
};

//*===========================================================================*/
class ISAC{

public: 
    //* Constructor for ISAC class */
	ISAC(); 

    //* Constructor of ISAC class.
    // * Output:
    // *      "chip3D": Object of Chippkg_type class. 
    // * Input:
    // *      "boundary": boundary condition */
    ISAC(ChipPkg_type & chip_3D, const Boundary & boundary);

    //* Generate the object of Solution class based on the input chip package 
    // * information ("chip_3D").
    // * Input:
    // *      "chip3D": Object of Chippkg_type class. */ 
    Solution * build_solution(const ChipPkg_type & chip_3D);

    /* Return a private parameter. */
    Solution * retn_solution(){return solution_;}

    //* Do the steady-state thermal simulation. 
    // * Calculate the steady thermal profile for the given power profile
    // * ("power_profile") when time proceeds to infinity. 
    // * Input:
    // *      "power": power profile. */
    std::vector<Element_type> solve_static(const 
        std::vector<isac::Element_type> & power_profile);
 
    //* Do the steady-state thermal simulation. 
    // * Calculate the steady thermal profile for the given power profile
    // * ("power") when time proceeds to infinity. 
    // * Input:
    // *      "power": power profile. */
    std::vector<Element_type> solve_static(const 
        std::vector<std::vector<double> > & power);
    
    //* Do the steady-state thermal simulation. 
    // * Calculate the steady thermal profile for the given power profile
    // * ("power") when time proceeds to infinity. 
    // * Input:
    // *      "power": power profile. */
    std::vector<Element_type> solve_static_adaptive(const 
        std::vector<std::vector<double> > & power);

    //* Initial dynamic simulation.
    // * It actually calls the steady-state simulation function, and its output 
    // * thermal profile serves as the initial condition for later dynamic 
    // * simulation. 
    // * Input:
    // *      "power": power profile. */
	std::vector<Element_type> init_dynamic(const 
        std::vector<std::vector<double> > & power);

    //* Dynamic simulation function. 
    // * Calculate the thermal profile when time duration is "duration". 
    // * input: 
    // *      "duration": The simulation time of the dynamic simulation; 
    // *      "power_profile": power profile;
	std::vector<Element_type> step_dynamic(
        const std::vector<isac::Element_type> & power_profile, double duration);

     //* Dynamic simulation function. 
    // * Calculate the thermal profile when time duration is "duration". 
    // * input: 
    // *      "duration": The simulation time of the dynamic simulation; 
    // *      "power_profile": power profile;
	std::vector<Element_type> step_dynamic(
        const std::vector<std::vector<double> > & power, double duration);
 
    //* Dynamic simulation function. 
    // * Calculate the thermal profile when time duratio is "duration". 
    // * Output: 
    // *      "all_temp": Output temperature vectors. 
    // *          The ratio ("duration/tempreport_interval") determins the number of 
    // *          the output temperature vectors.
    // * input: 
    // *      "duration": The simulation time of the dynamic simulation; 
    // *      "tempreport_interval": The interval the tempeartures are reported;
    // *      "power": power profile;
    // *      "select_domain": choose whether to use time_domain solver or 
    // *          frequency domain solver. */
	std::vector<Element_type> step_dynamic(
		std::vector<std::vector<double> > & all_temp, double duration,  
		double tempreport_intval, 
        const std::vector<std::vector<double> > & power, char select_domain);
   
    //Zoom in a box in the chip defined by four edges and perform steady-state
    //thermal simulation
    std::vector<Element_type> zoom_and_solve_static(double x_min, double x_max,
        double y_min, double y_max, 
        const std::vector<std::vector<double> > & power, double plyr_z = 0.0);

    //Zoom in a box in the chip defined by four edges and perform dynamic
    //thermal simulation
    std::vector<Element_type> zoom_and_step_dynamic(double x_min, double x_max,
        double y_min, double y_max, 
        const std::vector<std::vector<double> > & power, double duration,
        double plyr_z = 0.0);

    //Report the temperatures of all devices inside a box
    std::vector<Device_type> report_device_temperatures(
        std::vector<Element_type> & thermal_profile,
        std::string layoutfilename, double x_min, double x_max,
        double y_min, double y_max, double plyr_z = 0.0);

    //* Map temperature from 3D elements to 2D elements (such as elements with the 
    // * same geometry information as the power profile elements).
    // * Output & input:
    // *      "temp_profile": Temperature profile.
    // * Input: 
    // *      "power": Power profile. */
    void Map_temperature(std::vector<Element_type> & temp_profile,
        const std::vector<std::vector<double> > & power);

    //* Subroutine for the Map_temperature function call(). */
    void Map_basic_temperature(std::vector<Element_type> & temp_profile,
        std::vector<double> & cond_vec);

    //* Subroutine for the Map_temperature function call(). */
    void Map_lumped_temperature(std::vector<Element_type> & temp_profile,
        const std::vector<double> & cond_vec,
        const std::vector<std::vector<double> > & power);
	
    //* Check the input power vector, to see whether it complies with the 
    //  information in the floorplan file. */
    void check_inputpower(const std::vector<std::vector<double> > & power);

private:
    Solution * solution_;
	Boundary boundary_;
};

/*===========================================================================*/
struct FlrplanElt_type {
    FlrplanElt_type();

    /* Constructors for convenience only */
    /* Sizes and center locations are in meter */
    FlrplanElt_type(double xmin, double xmax, double ymin, double ymax);

    double x_min;
    double x_max;
    double y_min;
    double y_max;
};


/*===========================================================================*/
struct Element_type {
	Element_type();

    /* Constructors for convenience only */
    /* Sizes and center locations are in meter. Z = 0 is the active layer */
	Element_type(double size_x, double size_y,
		double center_x, double center_y, double value = 0.0);

    Element_type(double size_x, double size_y, double size_z, double center_x,
		double center_y, double center_z, double value = 0.0);
	
		double size[3];
		double center[3];
		double value;
};

/*===========================================================================*/
struct Device_type {
    Device_type();

    Device_type(double loc_x, double loc_y, double loc_z,
        double power_val, double temp_val = 0.0);

    double loc[3];
    double power;
    double temp;
    
};


/*===========================================================================*/
struct Component_type{

	std::vector<Layer_type> layers;
};


/*===========================================================================*/
/*enum material_property {power_, interface_, material_}; */
enum property_type {POWER_, INTERFACE_, MATERIAL_};


/*===========================================================================*/
struct Layer_type{
	public:

		Layer_type();
		property_type property;/* define what layer it is  */
		double size[3];
		double coor[3];
        int par[3]; //The initial partition of the layer
		double cond;  /* W/m-K for 3D material layer, and W/m^2-K for 
					interface layer*/
		double cap;
		std::vector<Element_type> power_profile;

        bool homog_cond; /* a flag to indicate if the layer has a homogoneous
                    thermal conductivity or not, by default true */
        std::vector<Element_type> cond_map; /* defines the thermal conductivity map
                                for a non-homogeneous conductivity layer */
        bool homog_cap; /* a flag to indicate if the layer has a homogoneous
                   heat capacity or not, by default true */
        std::vector<Element_type> cap_map; /* defines the heat capacity map
                                for a non-homogeneous capacity layer */

};


/*===========================================================================*/
struct ChipPkg_type{
    ChipPkg_type();

    std::vector<Component_type> components;
    par_grid isot;
    par_grid cont;
	//* The initial partition of each component based on the power profile. */
	par_grid init_partition;
	//* The basic parition obtained by the smallest x size and smallest y size
    //	of all the components. */
	par_grid basic_partition;
	//* Power grid info. (only the size coor info. neighbor info 
	//* is not necessary), this is in accordance with the floorplan file
	//*	(total number of grids equals to that of the power blocks). */
	par_grid power_grid;
};

} /* namespace isac */

//* Form the object of isac class based on the input chip package
// * ("chippkg"), floorplan("flrplan") and boundary condition ("boundary").
// * Output: 
// *      "chip3D": Object of Chippkg_type class. 
// * Input: 
// *      "chippkg": Chip package definition;
// *      "flrplan": floorplan;
// *      "boundary": boundary condition definition. */
isac::ISAC form_isac(isac::ChipPkg_type & chip3D,
    const std::vector<isac::Layer_type> & chippkg, 
	const std::vector<std::vector<isac::FlrplanElt_type> > & flrplan,
    const isac::Boundary & boundary);

//* Form the object of isac class based on the input chip package file 
// * ("chippkg_file"), floorplan file ("floorplan_file") and 
// * and boundary condition ("boundary").
// * Output: 
// *      "chip3D": Object of Chippkg_type class. 
// * Input: 
// *      "chippkg_file": Chip package definition file;
// *      "floorplan_file": floorplan file;
// *      "boundary": boundary condition definition. */
isac::ISAC form_isac(isac::ChipPkg_type & chip3D, const char * chippkg_file, 
	const char * floorplan_file, const isac::Boundary & boundary);

//* Adjust the absoluate temperature value of the thermal profile according 
// * to the boundary condition. */
void temp_adj(std::vector<isac::Element_type> & thermal_profile, 
	const isac::Boundary & boundary);

//* Adjust the absolute temperature value of the vector "temp" based on the 
// * boundary condition. */
void temp_adj(std::vector<double> & temp, const isac::Boundary 
	& boundary);

//* Check the validity of the input chip package description */
void check_input(const isac::ChipPkg_type & chip_3D);

//* Copy information from "src_layer" to "dst_layer". */
void copy_layer(isac::Layer_type & dst_layer, 
	const isac::Layer_type src_layer);

//* Copy information from "src_elt" to "dst_elt". */
void copy_elts(isac::Element_type & dst_elt,const isac::Element_type src_elt);

//* Convert a vector of Element_type to a vector of Map_block. */
std::vector<Map_block> toMap(std::vector<isac::Element_type> & a);


#endif /* isac.h */
