/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** main.cc
****************************************************************************/

#include <iostream>
#include <sstream>
#include <algorithm>
#include <vector>
#include "hoct_model.h"
#include "isac.h"
#include "grids.h"
#include "util.h"
#include "assert.h"
#include <fstream>
#include "header.h"
#include "configuration.h"

using namespace isac;
using namespace std;

/*===========================================================================*/
int main(int argc, char* argv[]) {

    configuration config(argc, argv);

    //cout << config << endl;
    /*======================================================================*/
    /* Example Code for using the device solver and generating a look-up table
    file */
    /* Uncomment this code, and run the solver using: ./solver -d EXAMPLE_DEVICE_FILE */
    /*
    Device_structure device;

    string device_file = configuration::ap().device_file();
    int verify = device.input(device_file);
    if(verify != 0) { 
        cerr << "Error in reading device information" << endl;
        exit(1);
    }

    device.solve();

    string LUTfile = "LUT";
    device.generateLUT(LUTfile.c_str());

    //(Optional)
    //device.output("output");
    */
    /*======================================================================*/


    /*======================================================================*/
    /* Example code for using the multiscale solver */
    /* Run the solver using: ./solver -c EXAMPLE_CHIP_FILE -f EXAMPLE_FLOORPLAN_FILE -p EXAMPLE_POWER_FILE -y EXAMPLE_LAYOUT_FILE -l LUT */
    ChipPkg_type chip3D;
    //Boundary bond(318.15, 7.43E-4, 1.0);
    Boundary bond(328.15, 0.0, 1.0);

    ISAC isac;

    //create and configure the chip from the provided information
    string chip_file = configuration::ap().chip_file();
    string floorplan_file = configuration::ap().floorplan_file();
    isac = form_isac(chip3D, chip_file.c_str(), floorplan_file.c_str(), bond);

    /*
    vector<Layer_type> chipPkg;
    chipPkg.resize(4);
    chipPkg[0].property = POWER_;
    chipPkg[0].size[X_] = 16.3E-3;
    chipPkg[0].size[Y_] = 16.3E-3;
    chipPkg[0].size[Z_] = 0.0;
    chipPkg[0].cond = 142.3;
    chipPkg[0].cap = 1.66E6;

    chipPkg[1].property = MATERIAL_;
    chipPkg[1].size[X_] = 16.3E-3;
    chipPkg[1].size[Y_] = 16.3E-3;
    chipPkg[1].size[Z_] = 0.725E-3;
    chipPkg[1].cond = 142.3;
    chipPkg[1].cap = 1.66E6;
    chipPkg[1].par[X_] = 16;
    chipPkg[1].par[Y_] = 16;
    chipPkg[1].par[Z_] = 8;
    
    chipPkg[2].property = MATERIAL_;
    chipPkg[2].size[X_] = 16.3E-3;
    chipPkg[2].size[Y_] = 16.3E-3;
    chipPkg[2].size[Z_] = 2E-3;
    chipPkg[2].cond = 400.0;
    chipPkg[2].cap = 3.55E6;
    chipPkg[2].par[X_] = 16;
    chipPkg[2].par[Y_] = 16;
    chipPkg[2].par[Z_] = 8;

    chipPkg[3].property = MATERIAL_;
    chipPkg[3].size[X_] = 34.0E-3;
    chipPkg[3].size[Y_] = 34.0E-3;
    chipPkg[3].size[Z_] = 2E-3;
    chipPkg[3].cond = 237;
    chipPkg[3].cap = 2.43E6;
    chipPkg[3].par[X_] = 16;
    chipPkg[3].par[Y_] = 16;
    chipPkg[3].par[Z_] = 4;

    string floorplan_file = configuration::ap().floorplan_file();
    ifstream fin;
    fin.open(floorplan_file.c_str());
    vector<vector<FlrplanElt_type> > flrplan;
    flrplan.resize(1);
    double x_min, x_max, y_min, y_max;
    while(fin >> x_min) {
        fin >> x_max;
        fin >> y_min;
        fin >> y_max;
        flrplan[0].push_back(FlrplanElt_type(x_min, x_max, y_min, y_max));
    }

    isac = form_isac(chip3D, chipPkg, flrplan, bond);
    */

    vector<Element_type> thermal_profile;
    vector<vector<double> > power;
    power.resize(1);
    power[0].resize(4096);

    ifstream powerfile;
    powerfile.open(configuration::ap().powerfile().c_str());

    double cur_timepoint = 0.0;
    double next_timepoint = 0.0;
    double duration = 0.0;
    char unit[10];

    // Scan in the power profile
    for(unsigned int i = 0; i < power[0].size(); i++) {
        powerfile >> power[0][i];
    }
    powerfile >> next_timepoint;
    powerfile >> unit;
    if (!powerfile)
      err_sys("File format error");

    // Init the dynamic solver with power profile at 0
    thermal_profile = isac.init_dynamic(power);

    while (powerfile) {

      // calculate next simulation duration
      convert_unit(next_timepoint, unit); 
      duration = next_timepoint - cur_timepoint;
      cur_timepoint = next_timepoint;

      // simulate for the duration
      thermal_profile = isac.step_dynamic(power, duration);

      for(unsigned int i = 0; i < power[0].size(); i++) {
        powerfile >> power[0][i];
      }

      powerfile >> next_timepoint;
      powerfile >> unit;
    }

    //cout << "Simulation ending time:" << next_timepoint << endl;
    powerfile.close();

    //* print out each element's temperature
    for(unsigned int i = 0; i < thermal_profile.size(); i++) {
        cout << thermal_profile[i].center[X_] << " ";
        cout << thermal_profile[i].center[Y_] << " ";
        cout << thermal_profile[i].center[Z_] << " ";
        cout << thermal_profile[i].value << endl;
    }
    

    //Print out the resulting temperature profile for visualizing using gnuplot
    //print_hetero_eltvec_gnuplot(thermal_profile, 64);

    //Zoom-in a box of the chip defined by x = -4mm to x = 0mm, and y = -2mm to     
    //y = 4mm and recompute the steady-state temperature profile
    //thermal_profile = 
    //            isac.zoom_and_solve_static(-4.0, 0.0, -2.0, 4.0, power);

    //Reprint out the resulting temperature profile for gnuplot
    //print_hetero_eltvec_gnuplot(thermal_profile, 64);

    //Report the temperatures of devices inside the box defined by x = -1mm to x = 0mm and y = -1mm to y = 0mm
    //vector<Device_type> devices;

    //string layout_file = configuration::ap().layout_file();
    //devices = isac.report_device_temperatures(thermal_profile, layout_file, -1.0, 0.0, -1.0, 0.0);

    //Print out the device temperatures
    //for(unsigned int i = 0; i < devices.size(); i++) {
    //    cout << devices[i].loc[X_] << " " << devices[i].loc[Y_] << " ";
    //    cout << devices[i].loc[Z_] << " " << devices[i].temp << endl;
    //}
    /*======================================================================*/

    return 0;
}

