/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** sol.h: the solution class in ISAC
****************************************************************************/

#ifndef SOL_H_
#define SOL_H_

#include <iostream>
#include <vector>
#include <ctime>
#include "hoct_model.h"
#include "init.h"
#include "isac.h"
#include "dysol.h"
#include "nhar.h"

namespace isac {
/*===========================================================================*/
/* The class for 3D thermal modeling */
class Solution{
		
public:
	
	Solution();
	Solution(const std::vector<Component_type> & chip_3D); 
    Solution(const ChipPkg_type & chip_3D);

    //* Form the related objects inside Solution class. */
    void form_solution
        (const double ambient_resis, const double ambient_heatcap);
    
    //* Steady-state thermal simulation.
    // * Calculate the thermal profile when time goes to infinity. */
    std::vector<Element_type> static_solver(
        const std::vector<std::vector<double> > & power, double err_thres,
        const double ambient_resis, const double ambient_heatcap); 
    //* Steady-state thermal simulation.
    // * Calculate the thermal profile when time goes to infinity. */
    std::vector<Element_type> static_solver_adaptive(
        const std::vector<std::vector<double> > & power, double err_thres,
        const double ambient_resis, const double ambient_heatcap); 
    //* Spatial refinement and static thermal simulation.
    // * Do spatial refinement and steady-state thermal simulation until the
    // * thermal gradient meets the predefined threshold. */
	void spatial_multigrid_solver(
        const std::vector<std::vector<double> > & power, 
        const double ambient_resis, const double ambient_heatcap);
    //* Generate the Solution object and call iterative smoother function
    // * call to calculate the steady state thermal profile. */
	void static_temp_compute(double err_thres);
    //* Dynamic thermal simulation function. */
    std::vector<Element_type> dynamic_solver(std::vector<std::vector<double> >
		& all_temp, double duration, double tempreport_interval, 
        const std::vector<std::vector<double> > & power, 
		char select_domain);
    //* Time domain dynamic thermal analysis. */
    void time_domain_solver(double duration);
    //* Frequency domain dynamic thermal analysis based on the NHAR method. */
    void frequency_domain_solver(std::vector<std::vector<double> > & all_temp, 
		double duration, double tempreport_interval);
    //* The NHAR moment-mathcing thermal analysis method. */
    void nhar_mm(std::vector<std::vector<double> > & all_temp,
        double duration, double tempreport_interval);
    //* The NHAR moment-matching thermal analysis method with adaptive moment
    //*  determination. */
    void nhar_mm_adap(std::vector<std::vector<double> > & all_temp,
        double duration, double tempreport_interval);
    //* Parition all the elements inside a box homogeneously and map the power
    //profile to the new grid
    void zoom_box(double x_min, double x_max, double y_min, double y_max,
    double plyr_z, const std::vector<std::vector<double> > & power);
    //* Generate the object of Solution class. 
    // * Setup the first level element of the hoct_contour data structure  
    // * (root_). */
	void gene_solu();
    // * Setup the second level elements of the hoct_contour data structure  
    // * (compnonents). */
	void gene_comps();
    //* Setup the third level elements of the hoct_contour data structure  
    //* (layers). */
    void gene_layers(par_grid & root, Component_type & comp, 
        int & layer_id, int & elem_id);
    //* Setup the third level elements of the hoct_contour data structure  
    //* (initial partitioned thermal elements). */
    //* Generate the finest grid elements for the initial partition. */
    void gene_elements(par_grid & root, Layer_type & layer, int & elem_id);
    //* Generate the neighbor information for each partition level (the whole
    //* packge level, the component level, the layer level) of the chip package 
    //* based on the init_partition input */
    void gene_neis();
    //* Generate the neighbor relationship for the components. */
    void gene_comp_neis(par_grid & init_partition);
    //* Generates the neighbor information for the layers and/or the initial
    //* finest elements. */
    void gene_layers_neis(par_grid & root, par_grid & init_partition);
    //* Generate the neighbor information for the elements in the chip package 
    //* based on the init_partition input. */
    void gene_elts_neis(par_grid & init_partition);
    //* Find the elements that have upper/lower neighbor of power
    //* for 3D thermal elements. */ 
    void find_powernei(int level, char direction);
    //* Further partition hoct_tree here as needed */
    void refine_par();
    void refine_par(int comp, int layer, int x_div, int y_div, int z_div);
    void refine_par(int comp, int layer, int elt, int x_div,
        int y_div, int z_div);
    //* Check whether the children of the par grid are the neighbor of the
    //* power layer(a vector of par grids) with a coor of z_pos. */
    //* push the according grids into power_neis_ when correct
    //* neighbors are found. */
    //* only check neighbor in Z_ direction. */
    void check_nei(par_grid * par, std::vector<par_grid *> & cur_power_layer,
		int direction);

    //* Calcuate the mapped power value to the 3D elements from the 2D power 
    //* input. */
    void cal_power_value(int level, 
        const std::vector<std::vector<double> > & power);
    //* Find the 3D elements that should be mapped with power values from the
    //* 2D power input. */
    void find_power_rela();
    //* Find the 3D elements that should be mapped with power values from the
    //* 2D power input. */
    void find_power_rela_elt(grid_contour & grid_cont, int level);
    //* Calculate the coefficient for power value calculation for one grid, 
    //* which is the neighbor of power_layer in Z direction. */
    void cal_power_rela_coeff(grid_contour & grid_cont, char nei_rela,
        const Layer_type & power_layer, int pelt_index_offset, int level);
    //* Set the initial temperature to be 0.0 for all the thermal grids */
	void set_init_temp();
	//* Put the temperature valus back from temp_ to hoct_tree grids */
	void temperature_back_tree();
    //* Spatial refinement based on the thermal gradient.
    //* Check the thermal gradient between each two neighbor elements, and refine
    //* them if the thermal gradient between them are larger than the input 
    //* thermal gradient threshold ("thermal_thre"). */
	int spatial_refinement(double thermal_thre);
    //* Update the data structures after the spatial refinement. */
	void spatial_update();

    /* Calculate the hcoeff_conv_. */
    void cal_hcoeff();

    //* Find out the smallest sizes for all the layers. */
    double min_zsize();
    
    //* Run the ssmoother iteratively among multiple AGCs, or several parts of 
    //* AGCs. */
	void multiple_ssmoother(double err_thres);

    //* Here adjust the convection effect(resistance) into the thermal
    //* implementation by adding nodes into the AGC data structure(not
    //* real grids to the tree or the hoct_contour). The added node also
    //* presents resistance and heat capacity as other normal nodes. 
    //* Since the power_ and temp_ vectors also need to be revised to conduct
    //* the thermal modeling, according revisions needs to be done when the 
    //* temp_ vector was mapped back to the tree (contour) */
    //* Change the power_ and temp_ vectors here by adding the
    // * additional elements. */
    void broaden_AGC(const double ambient_resis, const double
        ambient_heatcap);
    void broaden_AGC_elt(); 
    //* Generate the correct power_ and temp_ information when considering 
    //* convection effect. /
    void broaden_temp_power(double temp_value);
    /* Initialize the power_ and temp_ with the correct size and 0.0 values*/
    void init_power_temp();
	
    //* Do custom spatial refinement for any required element. */
    void FC_spatial_refinement(int elem_index, int x_div, int y_div, int z_div);
    //* Partition all the elements into 2 in Z dimension */
    void refine_zpar();
	void FC_spatial_update(const std::vector<std::vector<double> > & power);

    par_grid root(){return root_;}
		
    std::vector<std::vector<grid_contour> > & hoct_contour() 
		{ return hoct_contour_;}
	const std::vector<std::vector<grid_contour> > & hoct_contour() const 
	 	{ return hoct_contour_;}
	std::vector<grid_contour> & hoct_contour(int i) 
		{ return hoct_contour_[i];}
	const std::vector<grid_contour> & hoct_contour(int i) 
		const { return hoct_contour_[i];}
	grid_contour hoct_contour(int i, int j) { return hoct_contour_[i][j];}

    RC_network_type & AGC() {return AGC_;}

    std::vector<std::pair<int, par_grid *> > & power_pos()
		{return power_pos_;}
	std::pair<int, par_grid *> & power_pos(int i)
		{return power_pos_[i];}
    
    std::vector<double> & temp() {return temp_;}
	double temp(int i) {return temp_[i];}
	std::vector<double> & power() {return power_;}
	double power(int i) {return power_[i];}

	int init_level() {return init_level_;}
	void init_level_assign(int a) {init_level_ = a;}

	ChipPkg_type * chip() {return chip_;}

	std::vector<std::vector<par_grid *> > & player_btmnei()
		{return player_btmnei_;}
	std::vector<par_grid *> & player_btmnei(int i)
		{return player_btmnei_[i];}
	par_grid * player_btmnei(int i, int j)
		{return player_btmnei_[i][j];}

private:

	par_grid root_;
    std::vector<std::vector<grid_contour> > hoct_contour_;
    RC_network_type AGC_;
    std::vector<std::pair<int, par_grid *> > power_pos_;
    std::vector<double> temp_;
 std::vector<std::vector<double> > old_temp;
	std::vector<double> power_;
	//* Total number of thermal elements. */
	int N_;
	//* The number of thermal elements in each component. */
	std::vector<int> elts_incomp_;
	int init_level_;
	par_grid isot_; 
	par_grid cont_; 

	//* Pointer to the 3D chip information so that the Solution can be called 
    // * multiple times without losing the chip package information. */
	ChipPkg_type * chip_;

	//* The bottom neighbor of all the power layers when the simulation
	// * is done, so that the temperature of the power layer can be reported. */
	std::vector<std::vector<par_grid *> > player_btmnei_;
	
    //* h coefficient for convection effect, the unit is W/(m^2-K). */
    double hcoeff_conv_;

    int btm_elts_;
    double temp_conv_;
    bool AGC_expanded_;
};
} //* namespace isac */


void cal_sizecoor(std::vector<double> & sizes, 
    std::vector<double> & coors, 
    const isac::Component_type & component, const double & top_zpos);

#endif
