/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** solve.cc: solve the BTE equations using line by line TDMA
****************************************************************************/

#include "header.h"

/*==========================================================================*/
double BTE_structure::lblTdma(int angle, double max_rel_change)
{

    int i,j,k;	//counters for 3 directions
	
    int l;

    vector<double> coeffVector1,coeffVector2,coeffVector3,bVector,energyVector;
	
    double rel_change; //the relative change of energy after each iteration
	
    double energy_new; //the value of energy after iteration
    double energy_old; //the value of energy before iteration
	
    //reserve memory space for the vectors
    coeffVector1.resize(cells_z-2);	//boundaries are not taken
    coeffVector2.resize(cells_z-2);
    coeffVector3.resize(cells_z-2);
	
    bVector.resize(cells_z-2);
	
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(cells_z-2);

    //loop for all cartesian directions

    //Sweeping the directions (6 loops)
    /* ------------------------  First Loop  ------------------------ */

    for(i=1;i<cells_x-1;i++)
    //loop for all cells in x-direction (forward)
    {
        for(j=1;j<cells_y-1;j++)
        //loop for all cells in y-direction (forward)
        {
            for(k=1;k<cells_z-1;k++)
            //loop for all cells in z-direction
            {
                l=k-1;
                coeffVector1[l]=cell[i][j][k].AP_; //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].ATop_; //AT1D(k) = AT(i,j,k);
                coeffVector3[l]=-cell[i][j][k].ABottom_; //AB1D(k) = AB(i,j,k);
                bVector[l]=cell[i][j][k].B_+
                        cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]+
                        cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]+
                        cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle];



                bVector[l] = bVector[l] + 
						cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);
                    //B1D(k) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                    //                  + AE(i,j,k)∗X(i+1,j,k)
                    //                  + AS(i,j,k)∗X(i,j-1,k)
                    //                  + AN(i,j,k)∗X(i,j+1,k);
                    
                    //add effects of two cells at the boundary to B term
                if(k==1) {
                    bVector[l] = bVector[l]
                        +cell[i][j][1].ABottom_*cell[i][j][0].f_angle[angle];
                }
                if(k==(cells_z-2)) {
                    bVector[l] = bVector[l]
                        +cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];
                }
            }
            // Call TDMA(AP1D,AT1D,AB1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            // Write elements back in structure
            for(k=1;k<cells_z-1;k++)
            {
                //X(i,j,k) = X1D(k);
                cell[i][j][k].f_angle[angle]=energyVector[k-1];
            }	
        }
    }
        
    //	/ ------------------------  Second Loop  ------------------------ /
    
    
    for(i=cells_x-2;i>=1;i--)
    //loop for all cells in x-direction (backward)
    {
        for(j=cells_y-2;j>=1;j--)
        //loop for all cells in y-direction (backward)
        {
            for(k=1;k<cells_z-1;k++)
            //loop for all cells in z-direction
            {
                l=k-1;
                coeffVector1[l]=cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].ATop_;  //AT1D(k) = AT(i,j,k);
                coeffVector3[l]=-cell[i][j][k].ABottom_;  //AB1D(k) = AB(i,j,k);
                bVector[l]=cell[i][j][k].B_
                        +cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]
                        +cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]
                        +cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]
                        +cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle];

                bVector[l] = bVector[l] +
			        cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

                //B1D(k) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                  + AE(i,j,k)∗X(i+1,j,k)
                //                  + AS(i,j,k)∗X(i,j-1,k)
                //                  + AN(i,j,k)∗X(i,j+1,k);
                    
                //add effects of two cells at the boundary to B term
                if(k==1) {
                    bVector[l] = bVector[l]
                        +cell[i][j][1].ABottom_*cell[i][j][0].f_angle[angle];
                }
                if(k==(cells_z-2)) {
                    bVector[l] = bVector[l]
                        +cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];
                }
            }
            // Call TDMA(AP1D,AT1D,AB1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            // Write elements back in structure
            for(k=1;k<cells_z-1;k++)
            {
                //X(i,j,k) = X1D(k);
                cell[i][j][k].f_angle[angle]=energyVector[k-1];
            }	
        }
    }	
    
    
    //reserve memory space for the vectors
    coeffVector1.resize(cells_y-2);
    coeffVector2.resize(cells_y-2);
    coeffVector3.resize(cells_y-2);
    bVector.resize(cells_y-2);
            
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(cells_y-2);
    
    /* ------------------------  Third Loop  ------------------------ */
    
    for(k=1;k<cells_z-1;k++)
    //loop for all cells in z-direction (forward)
    {
        for(i=1;i<cells_x-1;i++)
        //loop for all cells in x-direction (forward)
        {
            for(j=1;j<cells_y-1;j++)
            //loop for all cells in y-direction
            {
                l=j-1;
                coeffVector1[l]=cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].ANorth_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-cell[i][j][k].ASouth_;  //AS1D(j) = AS(i,j,k);
                bVector[l]=cell[i][j][k].B_+
                        cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]+
                        cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle]+
                        cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];

                bVector[l] = bVector[l] +
                    cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

                //   B1D(j) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                     + AE(i,j,k)∗X(i+1,j,k)
                //                     + AB(i,j,k)∗X(i,j,k-1)
                //                     + AT(i,j,k)∗X(i,j,k+1);
                
                //add effects of two cells at the boundary to B term
                if(j==1) {
                    bVector[l] = bVector[l]
                        +cell[i][1][k].ASouth_*cell[i][0][k].f_angle[angle];
                }
                if(j==(cells_y-2)) {
                    bVector[l] = bVector[l]
                        +cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle];
                }	
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            // Write elements back in structure
            for(j=1;j<cells_y-1;j++)
            {
                //X(i,j,k) = X1D(k);
                cell[i][j][k].f_angle[angle]=energyVector[j-1];
            }	
        }
    }
    
    /* ------------------------  Fourth Loop  ------------------------ */
    
    for(k=cells_z-2;k>=1;k--)
    //loop for all cells in z-direction (backward)
    {
        for(i=cells_x-2;i>=1;i--)
        //loop for all cells in x-direction (backward)
        {
            for(j=1;j<cells_y-1;j++)
            //loop for all cells in y-direction
            {
                l=j-1;
                coeffVector1[l]=cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].ANorth_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-cell[i][j][k].ASouth_;  //AS1D(j) = AS(i,j,k);
                bVector[l]=cell[i][j][k].B_+
                        cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]+
                        cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle]+
                        cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];


                bVector[l] = bVector[l] + 
                cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

                //   B1D(j) = B(i,j,k) + AW(i,j,k)∗X(i-1,j,k)
                //                     + AE(i,j,k)∗X(i+1,j,k)
                //                     + AB(i,j,k)∗X(i,j,k-1)
                //                     + AT(i,j,k)∗X(i,j,k+1);
                
                //add effects of two cells at the boundary to B term
                if(j==1) {
                    bVector[l] = bVector[l]
                        +cell[i][1][k].ASouth_*cell[i][0][k].f_angle[angle];
                }
                if(j==(cells_y-2)) {
                    bVector[l] = bVector[l]
                        +cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle];
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            // Write elements back in structure
            for(j=1;j<cells_y-1;j++)
            {
                //X(i,j,k) = X1D(k);
                cell[i][j][k].f_angle[angle]=energyVector[j-1];
            }
        }
    }
    
    //reserve memory space for the vectors
    coeffVector1.resize(cells_x-2);
    coeffVector2.resize(cells_x-2);
    coeffVector3.resize(cells_x-2);
    bVector.resize(cells_x-2);
            
    //reserve memory space for solution vector and fill with zeros
    //to be able to insert elements anywhere
    energyVector.resize(cells_x-2);
        
    /* ------------------------  Fifth Loop  ------------------------ */
    
    for(j=1;j<cells_y-1;j++)
    //loop for all cells in y-direction (forward)
    {
        for(k=1;k<cells_z-1;k++)
        //loop for all cells in z-direction (forward)
        {
            for(i=1;i<cells_x-1;i++)
            //loop for all cells in x-direction
            {
                l=i-1;
                coeffVector1[l]=cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].AEast_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-cell[i][j][k].AWest_;  //AS1D(j) = AS(i,j,k);
                bVector[l]=cell[i][j][k].B_+
                        cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]+
                        cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle]+
                        cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle]+
                        cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];

                bVector[l] = bVector[l] +
                    cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

                //add effects of two cells at the boundary to B term
                if(i==1) {
                    bVector[l] = bVector[l]+
                        cell[1][j][k].AWest_*cell[0][j][k].f_angle[angle];
                }
                if(i==(cells_x-2)) {
                    bVector[l] = bVector[l]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle];
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            for(i=1;i<cells_x-1;i++)
            {
                //X(i,j,k) = X1D(k);
                cell[i][j][k].f_angle[angle]=energyVector[i-1];
            }	
        }
    }
    
    
    /* ------------------------  Sixth Loop  ------------------------ */
    
    for(j=cells_y-2;j>=1;j--)
    //loop for all cells in y-direction (backward)
    {
        for(k=cells_z-2;k>=1;k--)
        //loop for all cells in z-direction (backward)
        {
            for(i=1;i<cells_x-1;i++)
            //loop for all cells in x-direction
            {
                l=i-1;
                coeffVector1[l]=cell[i][j][k].AP_;  //AP1D(k) = AP(i,j,k);
                coeffVector2[l]=-cell[i][j][k].AEast_;  //AN1D(j) = AN(i,j,k);
                coeffVector3[l]=-cell[i][j][k].AWest_;  //AS1D(j) = AS(i,j,k);
                bVector[l]=cell[i][j][k].B_+
                        cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]+
                        cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle]+
                        cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle]+
                        cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle];

                bVector[l] = bVector[l] +
                    cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

                //add effects of two cells at the boundary to B term
                if(i==1) {
                    bVector[l] = bVector[l]+
                        cell[1][j][k].AWest_*cell[0][j][k].f_angle[angle];
                }
                if(i==(cells_x-2)) {
                    bVector[l] = bVector[l]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle];
                }
            }
            // Call TDMA(AP1D,AN1D,AS1D,B1D,X1D);
            Tdma(coeffVector1,coeffVector2,coeffVector3,bVector,energyVector);
            for(i=1;i<cells_x-1;i++)
            {
                //the value of energy after iteration
                energy_new=energyVector[i-1];
                //the value of energy before iteration
                energy_old=cell[i][j][k].f_angle[angle];
                //the relative change
                rel_change=fabs((energy_new-energy_old)/energy_old);
                //max_rel_change is used to stop the iterations
                //if its value goes below the required accuracy
                if(rel_change>max_rel_change)
                {
                    max_rel_change=rel_change;   		     
                }
                // Write elements back in structure
                cell[i][j][k].f_angle[angle]=energy_new;
            }
        }
    }

	return max_rel_change;

}

/*==========================================================================*/
void Tdma(vector<double> &coeffVector1,vector<double> &coeffVector2,vector<double> &coeffVector3,vector<double> &bVector,vector<double> &energyVector)
{
	
    int cell;	//counter for the cell number in the line
    double r;
    int size;	

    size=coeffVector1.size();	//size of the vectors

    for(cell=1;cell<size;cell++)	//start elimination from the second line
    {
        // r = AW(i)/AP(i-1);
        r=coeffVector3[cell]/coeffVector1[cell-1];
        // AP(i) = AP(i) - r*AE(i-1);
        coeffVector1[cell]=coeffVector1[cell]-r*coeffVector2[cell-1];
        //B(i) = B(i) - r*B(i-1);
        bVector[cell]=bVector[cell]-r*bVector[cell-1];
    }
    //--back substitution--
    //X(N) = B(N)/AP(N);
    energyVector[size-1]=bVector[size-1]/coeffVector1[size-1];	
    for(cell=size-2;cell>=0;cell--)
    {
        //X(i) = (B(i) - AE(i)*X(i+1))/AP(i);
        energyVector[cell]=(bVector[cell]-coeffVector2[cell]*energyVector[cell+1])/coeffVector1[cell];	
    }
}


/*==========================================================================*/
//obtain the matrix coefficients
//input angular direction (as variable 'l')
void BTE_structure::obtain_coeffs(int l) {

    for(int i=1; i<cells_x-1; i++) {
        for(int j=1; j<cells_y-1; j++) {
            for(int k=1; k<cells_z-1; k++) {
                cell[i][j][k].reset_coeffs();
            }
        }
    }

	double flow; //temporary variable

	//get north and south coefficients
	for(int i=1; i<cells_x-1; i++) {
		for(int k=1; k<cells_z-1; k++) {
			//currently set up for uniform grid (width is constant)
			//flow = del_x*del_z*propvec[l].S_j()*vg_;
			
			for(int j=2; j<cells_y-1; j++) {
				//currently set up for uniform grid (width is constant)
	            flow=del_x*del_z*propvec[l].S_j()*(cell[i][j][k].material->vg_);
				cell[i][j][k].ASouth_ = max(0.0,flow);
				cell[i][j-1][k].ANorth_ = max(0.0,-flow);
			}

			//get the boundary coefficients
			flow=del_x*del_z*propvec[l].S_j()*(cell[i][1][k].material->vg_);
			cell[i][1][k].ASouth_ = max(0.0,flow);
			flow=del_x*del_z*propvec[l].S_j()*
										(cell[i][cells_y-2][k].material->vg_);
			cell[i][cells_y - 2][k].ANorth_ = max(0.0,-flow);	

		}
	}

    //get east and west coefficients
    for(int j=1; j<cells_y-1; j++) {
        for(int k=1; k<cells_z-1; k++) {
            //currently set up for uniform grid (width is constant)
            //flow = del_y*del_z*propvec[l].S_i()*vg_;

            for(int i=2; i<cells_x-1; i++) {
				//currently set up for uniform grid (width is constant)
                flow=del_y*del_z*propvec[l].S_i()*(cell[i][j][k].material->vg_);
                cell[i][j][k].AWest_ = max(0.0,flow);
                cell[i-1][j][k].AEast_ = max(0.0,-flow);

            }
             
			//get the boundary coefficients
			flow=del_y*del_z*propvec[l].S_i()*(cell[1][j][k].material->vg_);
            cell[1][j][k].AWest_ = max(0.0,flow);
			flow=del_y*del_z*propvec[l].S_i()*
										(cell[cells_x-2][j][k].material->vg_);
            cell[cells_x - 2][j][k].AEast_ = max(0.0,-flow);

        }
    }

    //get top and bottom coefficients
    for(int i=1; i<cells_x-1; i++) {
        for(int j=1; j<cells_y-1; j++) {
            //currently set up for uniform grid (width is constant)
            //flow = del_x*del_y*propvec[l].S_k()*vg_;

            for(int k=2; k<cells_z-1; k++) {
				//currently set up for uniform grid (width is constant)
                flow=del_x*del_y*propvec[l].S_k()*(cell[i][j][k].material->vg_);
                cell[i][j][k].ABottom_ = max(0.0,flow);
                cell[i][j][k-1].ATop_ = max(0.0,-flow);

            }
             
			//get the boundary coefficients
			flow=del_x*del_y*propvec[l].S_k()*(cell[i][j][1].material->vg_);
            cell[i][j][1].ABottom_ = max(0.0,flow);
			flow=del_x*del_y*propvec[l].S_k()*
                                        (cell[i][j][cells_z-2].material->vg_);
            cell[i][j][cells_z -2].ATop_ = max(0.0,-flow);

        }
    }

	double volume; //temporary variable
	//obtain centroid point coefficient Ap and b term
	for(int i=1; i<cells_x-1; i++) {
		for(int j=1; j<cells_y-1; j++) {
			for(int k=1; k<cells_z-1; k++) {
				volume = del_x*del_y*del_z; //the volume of a cell
				//get the centroid coeff.
				cell[i][j][k].AP_ = 
				volume*(propvec[l].extent_)/(cell[i][j][k].material->t_rel_) +
										cell[i][j][k].coeffs();

				//get the b term
				cell[i][j][k].B_ = (cell[i][j][k].e0_*volume* 
				propvec[l].extent_)/(cell[i][j][k].material->t_rel_);
			}
		}
	}

}

/*==========================================================================*/
//calculate e0 and T (in update section)
void BTE_structure::calc_e0_t() {

	//dont recalculate the boundary T or e0 if they are const temp. boundaries
		//the same goes if they are specular boundaries
    int x_0, x_end, y_0, y_end, z_0, z_end;
    if((tbound_x_0==1)||(spec_x_0==1))
        x_0 = 1;
    else
        x_0 = 0;
    if((tbound_x_end==1)||(spec_x_end==1))
        x_end = cells_x-1;
    else
        x_end = cells_x;
    if((tbound_y_0==1)||(spec_y_0==1))
        y_0 = 1;
    else
        y_0 = 0;
    if((tbound_y_end==1)||(spec_y_end==1))
        y_end = cells_y-1;
    else
        y_end = cells_y;
    if((tbound_z_0==1)||(spec_z_0==1))
        z_0 = 1;
    else
        z_0 = 0;
    if((tbound_z_end==1)||(spec_z_end==1))
        z_end = cells_z-1;
    else
        z_end = cells_z;

    //go over all cells
    for(int i = x_0; i<x_end; i++) {
        for(int j = y_0; j<y_end; j++) {
            for(int k = z_0; k<z_end; k++) {

                //calc this cell's e0 and T
                cell[i][j][k].calc_cell_e0_t(propvec);
            }
        }
    }

	//update the e0 and temp for the specular and diffuse boundaries
	calc_spec_diff_e0_t();

}

/*==========================================================================*/
//calculate the sum of the residuals
double BTE_structure::calc_residual(int angle)
{
	int i,j,k;	//counters for 3 directions
	
	double residual=0; //residuals are accumulated in this variable
	
	//loop on all blocks and sum residuals
	for(i=1;i<cells_x-1;i++)
	{
		for(j=1;j<cells_y-1;j++)
		{
			for(k=1;k<cells_z-1;k++)
			{
				//res=ApXp-(AwXw+AeXe+AsXs+AnXn+AbXb+AtXt+B)
				residual=residual+fabs(cell[i][j][k].AP_*
												cell[i][j][k].f_angle[angle]-
		                 (cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]+
		                  cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]+
		                  cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]+
		                  cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle]+
		                  cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle]+
		                  cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle]+
		                  cell[i][j][k].B_+
                          cell[i][j][k].q_vol*propvec[angle].extent_));
			}
		}
	}
	return residual;
}

/*==========================================================================*/
//set the initial guesses and conditions
void BTE_structure::ini_solve(void) {

    //set the boundary temperatures to the initial temperature if constant
	    //boundary temperatures were not specified
    double temper;
    if(tbound_x_0==0) {
        tbx_0_val = T_ini;
    }
    if(tbound_x_end==0) {
        tbx_end_val = T_ini;
    }
    if(tbound_y_0==0) {
        tby_0_val = T_ini;
    }
    if(tbound_y_end==0) {
        tby_end_val = T_ini;
    }
    if(tbound_z_0==0) {
        tbz_0_val = T_ini;
    }
    if(tbound_z_end==0) {
        tbz_end_val = T_ini;
    }	

	//go over all cells
	for(int i=0; i<cells_x; i++) {
		for(int j=0; j<cells_y; j++) {
			for(int k=0; k<cells_z; k++) {

				if(i==0) {
                    temper = tbx_0_val;
                }
                else if(i==(cells_x-1)) {
                    temper = tbx_end_val;
                }
                else if(j==0) {
                    temper = tby_0_val;
                }
                else if(j==(cells_y-1)) {
                    temper = tby_end_val;
                }
                else if(k==0) {
                    temper = tbz_0_val;
                }
                else if(k==(cells_z-1)) {
                    temper = tbz_end_val;
                }
                else
                    temper = T_ini;

				//set the guesses and conditions of the cell
				cell[i][j][k].ini_cell_solve(*this, temper);
			}
		}
	}
}

/*==========================================================================*/
//update boundary conditions for constant temperature boundaries
void BTE_structure::const_temp() {

	//if x axis has constant temperature boundaries
	if((tbound_x_0==1)||( tbound_x_end==1)) {
		//for all cells in y and z
		for(int j=0; j<cells_y; j++) {
			for(int k=0; k<cells_z; k++) {
				//for all angular directions
				for(int l=0; l<Nmax_angle; l++) {
					//if the vector is going in the boundary
					if(propvec[l].S_i_<0) {
						if(tbound_x_0==1) {
							cell[0][j][k].f_angle[l] = cell[1][j][k].f_angle[l];
						}
					}
					else {
						if(tbound_x_end==1) {
							cell[cells_x-1][j][k].f_angle[l] = 
									cell[cells_x-2][j][k].f_angle[l];
						}
					}
				}				
			}
		}
	}

	//if y axis has constant temperature boundaries
    if((tbound_y_0==1)||( tbound_y_end==1)) {
		//for all cells in x and z
        for(int i=0; i<cells_x; i++) {
            for(int k=0; k<cells_z; k++) {
                //for all angular directions
                for(int l=0; l<Nmax_angle; l++) {
                    //if the vector is going in the boundary
                    if(propvec[l].S_j_<0) {
						if(tbound_y_0==1) {
	                        cell[i][0][k].f_angle[l] = cell[i][1][k].f_angle[l];
						}
                    }
                    else {
                    	if(tbound_y_end==1)    
							cell[i][cells_y-1][k].f_angle[l] =
									cell[i][cells_y-2][k].f_angle[l];
					}
                }
            }
        }
    }

	//if z axis has constant temperature boundaries
    if((tbound_z_0==1)||( tbound_z_end==1)) {
		//for all cells in x and y
        for(int i=0; i<cells_x; i++) {
            for(int j=0; j<cells_y; j++) {
                //for all angular directions
                for(int l=0; l<Nmax_angle; l++) {
                    //if the vector is going in the boundary
                    if(propvec[l].S_k_<0) {
						if(tbound_z_0==1)
	                        cell[i][j][0].f_angle[l] = cell[i][j][1].f_angle[l];
                    }
                    else {
						if(tbound_z_end==1) {
	                        cell[i][j][cells_z-1].f_angle[l] =
                                    cell[i][j][cells_z-2].f_angle[l];
						}
					}
                }
            }
        }
    }

}

/*==========================================================================*/
//calculate the output heat flowing in direction dir
double BTE_structure::calc_qout(int index1, int index2, int dir)
{
    double qout=0;

    switch(dir)
    {
        case E_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_i_>0)
                {

                    qout = qout +
                           cell[cells_x-2][index1][index2].f_angle[angle]*
                           cell[cells_x-2][index1][index2].material->vg_*
                           propvec[angle].S_i_;
                }
                else
                {
                    qout = qout +
                           (-propvec[angle].S_i_)*
                           cell[cells_x-1][index1][index2].material->vg_*
                           cell[cells_x-1][index1][index2].e0_;
                }
            }
            qout = qout*del_y*del_z;
            break;

        case W_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_i_<0)
                {

                    qout = qout +
                           cell[1][index1][index2].f_angle[angle]*
                           cell[1][index1][index2].material->vg_*
                           (-propvec[angle].S_i_);
                }
                else
                {
                    qout = qout +
                           propvec[angle].S_i_*
                           cell[0][index1][index2].material->vg_*
                           cell[0][index1][index2].e0_;
                }
            }
            qout = qout*del_y*del_z;
            break;

        case N_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_j_>0)
                {

                    qout = qout +
                           cell[index1][cells_y-2][index2].f_angle[angle]*
                           cell[index1][cells_y-2][index2].material->vg_*
                           propvec[angle].S_j_;
                }
                else
                {
                    qout = qout +
                           (-propvec[angle].S_j_)*
                           cell[index1][cells_y-1][index2].material->vg_*
                           cell[index1][cells_y-1][index2].e0_;
                }
            }
            qout = qout*del_x*del_z;
            break;

        case S_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_j_<0)
                {

                    qout = qout +
                           cell[index1][1][index2].f_angle[angle]*
                           cell[index1][1][index2].material->vg_*
                           (-propvec[angle].S_j_);
                }
                else
                {
                    qout = qout +
                           propvec[angle].S_j_*
                           cell[index1][0][index2].material->vg_*
                           cell[index1][0][index2].e0_;
                }
            }
            qout = qout*del_x*del_z;
            break;

        case T_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_k_>0)
                {

                    qout = qout +
                           cell[index1][index2][cells_z-2].f_angle[angle]*
                           cell[index1][index2][cells_z-2].material->vg_*
                           propvec[angle].S_k_;
                }
                else
                {
                    qout = qout +
                           (-propvec[angle].S_k_)*
                           cell[index1][index2][cells_z-1].material->vg_*
                           cell[index1][index2][cells_z-1].e0_;
                }
            }
            qout = qout*del_x*del_y;
            break;

        case B_:
            for(int angle=0;angle<Nmax_angle;angle++)
            {
                if(propvec[angle].S_k_<0)
                {

                    qout = qout +
                           cell[index1][index2][1].f_angle[angle]*
                           cell[index1][index2][1].material->vg_*
                           (-propvec[angle].S_k_);
                }
                else
                {
                    qout = qout +
                           propvec[angle].S_k_*
                           cell[index1][index2][0].material->vg_*
                           cell[index1][index2][0].e0_;
                }
            }
            qout = qout*del_x*del_y;
            break;

        default:
            break;
    }

    return qout;

}

/*==========================================================================*/
//solve the BTE for the system
int BTE_structure::solve(int iter_param,Device_structure* device_structure) {

    f_error=(cell[0][0][0].material->cv_)*TEMP_ERROR/(4*PI);

     double residual=f_error*10.0; //the residual of the system
    //initialized with any value greater than ACCURACY in order for the loop to
    //execute the first time

	//the maximum relative change of energy of all cells
    double max_rel_change=f_error*10.0;
    //initialized with any value greater than ACCURACY in order for the loop to
    //execute the first time
    double residual_old=0;

    int count=0;

    //quit the loop when it has completed the desired amount of solutions
    while(count<iter_param) {

        count++;

        residual_old=residual;
        residual=0;

        //update boundary conditions for specular boundaries
        specular();
        //update boundary conditions for diffuse boundaries
        diffuse();
        //update boundary conditions for constant temperature boundaries
        const_temp();
       
        //calculate e0 and T (in update section)
        calc_e0_t();
        
        max_rel_change = 0;
        //start loop for all angular directions
        for(int angle=0;angle<Nmax_angle;angle++)
        {
	        //obtain matrix coefficients
    	    obtain_coeffs(angle);
	
    	    //execute lbl-tdma
            max_rel_change=lblTdma(angle,max_rel_change);
	
    	    //calculate residual
            residual=residual+calc_residual(angle);
        }
        //calculate e0 and T (in update section)
        calc_e0_t();

    }

    //return a 0 if solution converged
    //if((max_rel_change>f_error)||((residual>1e-6)&&(residual_old!=residual)))
    if(max_rel_change>f_error)
        return 1;
    else
        return 0;

}

