/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** transient_bte.cc: contains functions for transient BTE analysis
****************************************************************************/

#include "header.h"

/*==========================================================================*/
//solve the bte system for transient analysis
void BTE_structure::solve_transient(Device_structure* device_structure, 
															double del_t) {
	
    //update boundary conditions for specular boundaries
    specular();
    //update boundary conditions for diffuse boundaries
    diffuse();
    //update boundary conditions for constant temperature boundaries
    const_temp();
    //calculate e0 and T
    calc_e0_t();

	//go over all angles (and get the correct coefficients)
	//start loop for all angular directions
    for(int angle=0;angle<Nmax_angle;angle++) {
        //obtain matrix coefficients
        obtain_coeffs(angle);

        //solve the system for the current time step using 4th order RK
        //method
		trans_RK(angle,del_t);
    }
		
	//update all the energy densities with the new time values
	update_f_angle_trans();

	//calculate e0 and T
    calc_e0_t();

}

/*==========================================================================*/
//using runge-kutta with constant step time
void BTE_structure::trans_RK(int angle, double del_t) {

	double del_vol = del_x*del_y*del_z;
	
	//go over all cells not counting boundaries
	//loop for all cells in x-direction
	for(int i=1;i<cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<cells_y-1;j++) {
        	//loop for all cells in z-direction
	        for(int k=1;k<cells_z-1;k++) {
			
			    double b_term = cell[i][j][k].B_+
                        cell[i][j][k].AWest_*cell[i-1][j][k].f_angle[angle]+
                        cell[i][j][k].AEast_*cell[i+1][j][k].f_angle[angle]+
                        cell[i][j][k].ASouth_*cell[i][j-1][k].f_angle[angle]+
                        cell[i][j][k].ANorth_*cell[i][j+1][k].f_angle[angle]+
						cell[i][j][k].ATop_*cell[i][j][k+1].f_angle[angle]+
                        cell[i][j][k].ABottom_*cell[i][j][k-1].f_angle[angle];

                //add the heat source term
                b_term = b_term +
                        cell[i][j][k].q_vol*propvec[angle].extent_/(4*PI);

				//get k1
				double k1 =  (1/(del_vol*propvec[angle].extent_))*
                        (-cell[i][j][k].AP_*cell[i][j][k].f_angle[angle]
                        + b_term);

				//get k2
				double k2 = (1/(del_vol*propvec[angle].extent_))*
                        (-cell[i][j][k].AP_*
                            (cell[i][j][k].f_angle[angle] + k1*del_t/2)
                        	+b_term);

				//get k3
				double k3 = (1/(del_vol*propvec[angle].extent_))*
                        (-cell[i][j][k].AP_*
                            (cell[i][j][k].f_angle[angle] + k2*del_t/2)
                        	+b_term);

				//get k4
				double k4 = (1/(del_vol*propvec[angle].extent_))*
                        (-cell[i][j][k].AP_*
                            (cell[i][j][k].f_angle[angle] + k3*del_t)
                        	+b_term);

				//get new value
				cell[i][j][k].f_angle_next_t[angle] 
							= cell[i][j][k].f_angle[angle] +
                            	del_t*(k1+2*k2+2*k3+k4)/6;
			}
		}
	}
	
	//now the values for the angles at the new time are in f_angle_next_t
}

/*==========================================================================*/
//update all the energy densities with the new time values
void BTE_structure::update_f_angle_trans() {
	
	//loop over all angles
	for(int angle=0;angle<Nmax_angle;angle++) {

		//loop over all cells
    	//loop for all cells in x-direction
	    for(int i=1;i<cells_x-1;i++) {
    	    //loop for all cells in y-direction
        	for(int j=1;j<cells_y-1;j++) {
            	//loop for all cells in z-direction
	            for(int k=1;k<cells_z-1;k++) {

					//set the f_angles to the ones at the new time
					cell[i][j][k].f_angle[angle] = 
										cell[i][j][k].f_angle_next_t[angle];
				}
			}
		}
	}
}

/*==========================================================================*/
//update all the energy densities with the new time values
void Fourier_structure::update_T_trans() {
	
		//loop over all cells
    	//loop for all cells in x-direction
	    for(int i=1;i<f_cells_x-1;i++) {
    	    //loop for all cells in y-direction
        	for(int j=1;j<f_cells_y-1;j++) {
            	//loop for all cells in z-direction
	            for(int k=1;k<f_cells_z-1;k++) {

					//set the f_angles to the ones at the new time
					f_cell[i][j][k].T_ = 
										f_cell[i][j][k].T_next_;
				}
			}
		}
}

/*==========================================================================*/
//solve the bte system for transient analysis
void Fourier_structure::solve_transient(Device_structure* device_structure, 
															double del_t) {
	for(int i = 1; i < f_cells_x; i++) {
        for(int k = 1; k < f_cells_z; k++) {
            f_cell[i][0][k].T_ = 300.0;
        }
    }



    //solve the system for the current time step using 4th order RK
    //method
	trans_RK_fourier(del_t);

    update_T_trans();		

    //update the constant temperature boundary, y0
    
    for(int i = 1; i < f_cells_x; i++) {
        for(int k = 1; k < f_cells_z; k++) {
            f_cell[i][0][k].T_ = 300.0;
        }
    }

}



/*==========================================================================*/
//using runge-kutta with constant step time
void Fourier_structure::trans_RK_fourier(double del_t) {

	double del_vol = del_x*del_y*del_z;
	
	//go over all cells not counting boundaries
	//loop for all cells in x-direction
	for(int i=1;i<f_cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<f_cells_y-1;j++) {
        	//loop for all cells in z-direction
	        for(int k=1;k<f_cells_z-1;k++) {
			
			    double b_term = f_cell[i][j][k].B_+
                    f_cell[i-1][j][k].AEast_*f_cell[i-1][j][k].T_+
                    f_cell[i][j][k].AEast_*f_cell[i+1][j][k].T_+
                    f_cell[i][j-1][k].ANorth_*f_cell[i][j-1][k].T_+
                    f_cell[i][j][k].ANorth_*f_cell[i][j+1][k].T_+
		            f_cell[i][j][k].ATop_*f_cell[i][j][k+1].T_+
                    f_cell[i][j][k-1].ATop_*f_cell[i][j][k-1].T_;

                //add the heat source term
                b_term = b_term +
                        f_cell[i][j][k].q_vol;
                //cout << "qvol = " << f_cell[i][j][k].q_vol << endl;

				//get k1
				//double k1 =  (1/(del_vol*f_cell[i][j][k].material->cv_))*
				double k1 =  (1/(del_vol*f_cell[i][j][k].material->cv_))*
                        (-f_cell[i][j][k].AP_*f_cell[i][j][k].T_
                        + b_term);

/*
                cout << "AP = " << f_cell[i][j][k].AP_ << endl;
                cout << "T_ = " << f_cell[i][j][k].T_ << endl;
                cout << "cv_ = " << f_cell[i][j][k].material->cv_ << endl;
                cout << "b_term = " << b_term << endl;
                cout << "del_vol = " << del_vol << endl;
*/
                //cout << "k1 = " << k1 << endl;

				//get k2
				double k2 = (1/(del_vol*f_cell[i][j][k].material->cv_))*
                        (-f_cell[i][j][k].AP_*
                            (f_cell[i][j][k].T_             + k1*del_t/2.0)
                        	+b_term);

                //cout << "k2 = " << k2 << endl;
				//get k3
				double k3 = (1/(del_vol*f_cell[i][j][k].material->cv_))*
                        (-f_cell[i][j][k].AP_*
                            (f_cell[i][j][k].T_             + k2*del_t/2.0)
                        	+b_term);

                //cout << "k3 = " << k3 << endl;
				//get k4
				double k4 = (1/(del_vol*f_cell[i][j][k].material->cv_))*
                        (-f_cell[i][j][k].AP_*
                            (f_cell[i][j][k].T_             + k3*del_t)
                        	+b_term);
                //cout << "k4 = " << k4 << endl;

				//get new value
				f_cell[i][j][k].T_next_
							= f_cell[i][j][k].T_ +
                            	del_t*(k1+2*k2+2*k3+k4)/6;
                //cout << f_cell[i][j][k].T_next_ << endl;
			}
		}
	}
	
	//now the values for the angles at the new time are in f_angle_next_t
}

/*==========================================================================*/
//resize all of the storage for the future
void Device_structure::resize_trans_f_angle() {
	int size = bte_system.size();
	//loop over all bte structures
     for(int h=0; h<size; h++) {
		bte_system[h].resize_trans_f_angle();
	}
}

/*==========================================================================*/
//resize all of the storage for the future
void BTE_structure::resize_trans_f_angle() {

	//loop over all cells
    //loop for all cells in x-direction
	for(int i=1;i<cells_x-1;i++) {
        //loop for all cells in y-direction
       	for(int j=1;j<cells_y-1;j++) {
           	//loop for all cells in z-direction
	        for(int k=1;k<cells_z-1;k++) {
                //set the f_angles to the ones at the new time
                cell[i][j][k].f_angle_next_t.resize(Nmax_angle);
            }
        }
    }

}

/*==========================================================================*/
void Device_structure::output_yflux( ofstream & fout_flux) {

    //find the x-centerline
    int xcell = substrate.f_cells_x;
    xcell /= 2;

    cout << "Total number of x-cells = " << substrate.f_cells_x;
    cout << "Printing yflux from x-cell " << xcell << endl;

    int ycells = substrate.f_cells_y;
    int zcells = substrate.f_cells_z;

    double qout_bottom;

    //Loop over all ycells
    for(int i = 1; i < ycells - 1; i++) {
        //Calculate the total heat emitted from the bottom surface of the
        //cells
        qout_bottom = 0;
        //cout << "X: " << xcell-bte_coords[0].x+1 << endl;
        for(int j = 1; j < zcells - 1; j++) {
            //cout << "Z: " << j - bte_coords[0].z+1 << endl;
            //try to add more than one x cell
            for(int k = xcell - 5; k <= xcell + 5; k++) {
                qout_bottom = qout_bottom +
                        //bte_system[0].calc_qsouth(xcell-bte_coords[0].x+1,
                        bte_system[0].calc_qsouth(k-bte_coords[0].x+1,
                                             i-bte_coords[0].y+1,
                                             j-bte_coords[0].z+1);
            }
        }
        //Get the heat per unit area
        //qout_bottom = qout_bottom/(substrate.length_z * substrate.del_x);
        qout_bottom = qout_bottom/(substrate.length_z * (11.0*substrate.del_x));
        //fout_flux << i * substrate.del_y << " " << qout_bottom << endl;
        fout_flux << (ycells-i) * substrate.del_y <<" "<< qout_bottom << endl;

    }

}

/*==========================================================================*/
double BTE_structure::calc_qsouth(int index1, int index2, int index3) {

    double qout = 0;

    for(int angle=0;angle<Nmax_angle;angle++)
    {
        if(propvec[angle].S_j_<0)
        {

            qout = qout +
                   cell[index1][index2][index3].f_angle[angle]*
                   cell[index1][index2][index3].material->vg_*
                   (-propvec[angle].S_j_);
        }
        else
        {
            qout = qout +
                   cell[index1][index2-1][index3].f_angle[angle]*
                   cell[index1][index2-1][index3].material->vg_*
                   (-propvec[angle].S_j_);
        }
    }
    qout = qout*del_x*del_z;

    return qout;
}


/*==========================================================================*/
//get the peak temperature of the structure and print it, along with time
void Device_structure::output_peak_trans( ofstream & fout_tran, double time) {
	
	//go over all cells, see which has the peak temperature, then print that
	//one
	double peak_temp=0;
	int size = bte_system.size();
	//loop over all bte structures
	for(int h=0; h<size; h++) {
		//get the peak for this structure
		double temp = bte_system[h].get_peak();

		//replace the peak temperature if necessary
		if(temp > peak_temp) {
			peak_temp = temp;
		}
    //can also keep information about cell with peak, like location
	}
	
	//print out the results in a file
	//fout_tran<<peak_temp<<" "<<time<<endl;
	fout_tran<<time<<" "<<peak_temp<<endl;
	
}

/*==========================================================================*/
//get the peak temperature of the structure and print it, along with time
void Device_structure::output_peak_trans_fourier( ofstream & fout_tran, double time) {
	
	//go over all cells, see which has the peak temperature, then print that
	//one
    //get the peak for this structure
	double peak_temp=0;
	
	//loop over all cells
    //loop for all cells in x-direction
    for(int i=1;i<substrate.f_cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<substrate.f_cells_y-1;j++) {
            //loop for all cells in z-direction
            for(int k=1;k<substrate.f_cells_z-1;k++) {
                //set the f_angles to the ones at the new time
                if(substrate.f_cell[i][j][k].T_ > peak_temp) {
					peak_temp = substrate.f_cell[i][j][k].T_;
				}
                //can also keep information about cell with peak, like location
            }
        }
    }


	//print out the results in a file
	//fout_tran<<peak_temp<<" "<<time<<endl;
	fout_tran<<time<<" "<<peak_temp<<endl;
	
}


/*==========================================================================*/
//get the peak temperature of this bte structure
double BTE_structure::get_peak() {
	
	double peak_temp=0;
	
	//loop over all cells
    //loop for all cells in x-direction
    for(int i=1;i<cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<cells_y-1;j++) {
            //loop for all cells in z-direction
            for(int k=1;k<cells_z-1;k++) {
                //set the f_angles to the ones at the new time
                if(cell[i][j][k].T_ > peak_temp) {
					peak_temp = cell[i][j][k].T_;
				}
    //can also keep information about cell with peak, like location
            }
        }
    }

	return peak_temp;
}

/*==========================================================================*/
//update all the power sources depending on the time
void Device_structure::set_Q(double time) {
	int size = bte_system.size();
    //loop over all bte structures
     for(int h=0; h<size; h++) {
        bte_system[h].set_Q(time);
    }
}


/*==========================================================================*/
//update all the power sources depending on the time
void BTE_structure::set_Q(double time) {
	//loop over all cells
    //loop for all cells in x-direction
    for(int i=1;i<cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<cells_y-1;j++) {
            //loop for all cells in z-direction
            for(int k=1;k<cells_z-1;k++) {

               
                //if the time is past when the heat source is on, shut it off
                if(cell[i][j][k].on_time + cell[i][j][k].offset_time < time) {
                    cell[i][j][k].q_vol = 0;
                }
				//or if the time is still lower than the offset time, shut it
				//off
				else if(time<cell[i][j][k].offset_time) {
					cell[i][j][k].q_vol = 0;
				}
				//or if the time is between the delay and when it is off, turn
				//it on
				else {
					cell[i][j][k].q_vol = cell[i][j][k].q_vol_on;
				}
            }
        }
    }

}

/*==========================================================================*/
//update all the power sources depending on the time
void Device_structure::set_Q_fourier(double time) {
    
	//loop over all cells
    //loop for all cells in x-direction
    for(int i=1;i<substrate.f_cells_x-1;i++) {
        //loop for all cells in y-direction
        for(int j=1;j<substrate.f_cells_y-1;j++) {
            //loop for all cells in z-direction
            for(int k=1;k<substrate.f_cells_z-1;k++) {

               
                //if the time is past when the heat source is on, shut it off
                if(substrate.f_cell[i][j][k].on_time + substrate.f_cell[i][j][k].offset_time < time) {
                    substrate.f_cell[i][j][k].q_vol = 0;
                }
				//or if the time is still lower than the offset time, shut it
				//off
				else if(time<substrate.f_cell[i][j][k].offset_time) {
					substrate.f_cell[i][j][k].q_vol = 0;
				}
				//or if the time is between the delay and when it is off, turn
				//it on
				else {
					substrate.f_cell[i][j][k].q_vol = substrate.f_cell[i][j][k].q_vol_on;
                    //cout << substrate.f_cell[i][j][k].q_vol << endl;
				}
                
            }
        }
    }

}

