/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** util.cc
****************************************************************************/

#include <iostream>
#include <string>
#include <set>
#include "util.h"
#include "grids.h"
#include "init.h"

using namespace std;
using namespace isac;

/*==========================================================================*/
/* Constuctor for error_type class */
error_type::error_type():
flag_(0),
avg_(0),
max_(0),
min_(FMAX_),
thre_(0)
{
}

/*==========================================================================*/
/* Constuctor for error_type class */
error_type::error_type(double thre_a):
flag_(0),
avg_(0),
max_(0),
min_(FMAX_),
thre_(thre_a)
{
}

/*==========================================================================*/
/* Record the current time */
void timer_set(timeval *t) {

    gettimeofday(t, 0);
}


/*==========================================================================*/
/* Report the CPU time lapsed from the time defined by
 * timer_set function */
void timer_report(timeval *t) {

    time_t o_sec = t->tv_sec;
	time_t o_usec = t->tv_usec;
	gettimeofday(t, 0);

	cout << "cpu time is: " << t -> tv_sec - 
		o_sec + (t->tv_usec - o_usec) * 1e-6
		<< " seconds" << endl;
}


/*===========================================================================*/
/* Check whether two elements are neighbor or not in one dimension by given
 * the information of two elements' coordinates and sizes information */
int adj_check(double ac, double as, double bc, double bs) {

    double a = ac - as/2;
    double b = ac + as/2;
    double c = bc - bs/2;
    double d = bc + bs/2;

    double e = fmax(a,c);
    double f = fmin(b,d);
    // if( e < f) {
    // if( ((f-e) > DMIN_) ){
    if( ((f-e) > FMIN_) ) {
        return 1;
    }else {
        return 0;
    }
}


/*============================================================================*/
/* Convert the coordinate, size information of the lowest level grids to 
 * the vector of Element_type */
void pargrid_to_element(vector<Element_type> & grid_Element_type,
    vector<grid_contour> & hoct_contour_level,
    const vector<double> & value){

    int tot_size = static_cast<int>(hoct_contour_level.size());

    grid_Element_type.resize(tot_size);
    for (int i = 0; i < tot_size; ++i){
        grid_Element_type[i].size[0] = hoct_contour_level[i].self()
            -> sizes(X_);
        grid_Element_type[i].size[1] = hoct_contour_level[i].self()
            -> sizes(Y_);
        grid_Element_type[i].size[2] = hoct_contour_level[i].self()
            -> sizes(Z_);
        grid_Element_type[i].center[0] = hoct_contour_level[i].self()
            -> coors(X_);
        grid_Element_type[i].center[1] = hoct_contour_level[i].self()
            -> coors(Y_);
        grid_Element_type[i].center[2] = hoct_contour_level[i].self()
            -> coors(Z_);
        grid_Element_type[i].value = value[i];
    }
}


/*===========================================================================*/
/* Clear the contour levels that are greater than START_ELVEL_ 3 in 
 * hoct_contour. Hoct_contour levels (0,1,2,3) are always kept the same,
 * they are never cleared */
void clear_contour(vector<vector<grid_contour> > & hoct_contour){

    /* If the root is further partitioned before the contour is
     * constructed, then we should do nothing about this contour */
    
	int levels = hoct_contour.size();
	int keep_level = 4;
    if (levels <= keep_level){
        return;
    }else{
		for (int i = (levels-1); i > (keep_level - 1); --i){
			hoct_contour[i].clear();
		}
		for (int i = (levels-1); i >= keep_level ; --i){
			hoct_contour.erase(hoct_contour.end());
		}
	} 
    assert(int(hoct_contour.size()) == keep_level);
}

/*===========================================================================*/
/* Based on the input chip packge info "chippkg" and the floorplan info
 * "flrplan", read in all the package and floorplan information 
 * to parameter "components", which will be used later to construct
 * the ISAC class; */
void parser(vector<isac::Component_type> & components, 
	const vector<Layer_type> & chippkg,
    const vector<vector<FlrplanElt_type> > & flrplan){

	double zpos = 0.0;
	
	vector<int> pelts;
	unsigned int power_layer_num = 0;
	int comp_num = 0;
	int layer_num = 0;
	int player = 0;

    /* Read all the information from chip package and floorplan vectors to
    components */
    for(unsigned int i = 0; i < chippkg.size(); i++) {
		property_type layer_type;
		layer_type = chippkg[i].property;

		double size[DIM_C_];			
        size[X_] = chippkg[i].size[X_];
        size[Y_] = chippkg[i].size[Y_];
        size[Z_] = chippkg[i].size[Z_];
		double cond;
        cond = chippkg[i].cond;
		double cap;
        cap = chippkg[i].cap;
        int par[DIM_C_];
        //Read the partitoning information if this is not a power layer
        if (layer_type != POWER_) {
            par[X_] = chippkg[i].par[X_];
            par[Y_] = chippkg[i].par[Y_];
            par[Z_] = chippkg[i].par[Z_];
        }
		//if (!(comp_num == 0 && layer_num == 0)){
        /* A component only contains one power layer, therefore, a 
         * new power layer means the beginning of a new component */
		if ((layer_type == POWER_) && (player == 1)){
			comp_num ++;
			layer_num = 0;
			player = 0;
		}
        
        /* A material/interface layer with different size(s) than that of 
         * its previous layer (material layer or interface layer) also 
         * means the beginning of a new component */
        else if ( (layer_type != POWER_) && 
        (!((comp_num == 0) && (layer_num == 0))) ){
            if (components[comp_num].layers[layer_num-1].property != POWER_){
			    if ((fabs(size[X_] - 
                components[comp_num].layers[layer_num-1].size[X_])>FMIN_) || 
			    (fabs(size[Y_] - 
                components[comp_num].layers[layer_num-1].size[Y_])>FMIN_)){
				    comp_num ++;
				    layer_num = 0;
				    player = 0;
                }
			}
		}
		//}
		components.resize(comp_num + 1);
		components[comp_num].layers.resize(layer_num + 1);

		Layer_type & cur_layer = components[comp_num].layers[layer_num];
		switch (layer_type){
			case POWER_:{
				cur_layer.property = POWER_;
				player ++;
				break;
			}
			case INTERFACE_:{
				cur_layer.property = INTERFACE_;
				break;
			}
			case MATERIAL_:{
				cur_layer.property = MATERIAL_;
				break;
			}
			default:
				cerr << "layer type is not defined " << endl;
				exit(1);
		}
		for (int ii = 0; ii < DIM_C_; ++ii){
			cur_layer.size[ii] = size[ii];
			cur_layer.coor[ii] = 0.0;
            cur_layer.par[ii] = par[ii];
		}
		cur_layer.coor[Z_] = zpos + size[Z_]/2.0;
		zpos += size[Z_];
		cur_layer.cond = cond;
        if(chippkg[i].cond_map.size() != 0) {
            cur_layer.homog_cond = false;
        }
        cur_layer.cond_map = chippkg[i].cond_map;
		cur_layer.cap = cap;
        if(chippkg[i].cap_map.size() != 0) {
            cur_layer.homog_cap = false;
        }
        cur_layer.cap_map = chippkg[i].cap_map;
		if (layer_type == POWER_){
            /* The thickness of power layer must be 0 */
			if (size[Z_] > FMIN_){
				cerr << "The thickness of the power layer is not 0" << endl;
				exit(1);
			}
            //Make sure a floorplan for this power layer is provided
            if (flrplan.size() < power_layer_num + 1) {
                cerr << "error: Floorplan for power layer ";
                cerr << power_layer_num << " is not provided";
                exit(1);
            }
            for(unsigned int j = 0; j < flrplan[power_layer_num].size(); j++) {
				double x_min;
                x_min = flrplan[power_layer_num][j].x_min;
				double x_max;
                x_max = flrplan[power_layer_num][j].x_max;
				double y_min;
                y_min = flrplan[power_layer_num][j].y_min;
				double y_max;
                y_max = flrplan[power_layer_num][j].y_max;
                /* Check the sanity of the floorplan */
				if (!(x_max > x_min )){
					cerr << "error: The length in x dimension for some element " ;
					cerr << " in floorplan " << power_layer_num << " is negtive ";
					exit(1);
				}
				if (!(y_max > y_min )){
					cerr << "error: The length in y dimension for some element " ;
					cerr << " in floorplan " << power_layer_num << " is negtive ";
					exit(1);
				}
                Element_type elt;
                elt.size[X_]=(x_max-x_min);
				elt.size[Y_] = (y_max - y_min);
				elt.size[Z_] = 0.0;
				elt.center[X_] = (x_min + (x_max - x_min)/2.0);
				elt.center[Y_] = (y_min + (y_max - y_min)/2.0);
				elt.center[Z_] = cur_layer.coor[Z_];
                cur_layer.power_profile.push_back(elt); 
			}
			power_layer_num ++;
		}	
		layer_num ++;
 	}

}



/*===========================================================================*/
/* Based on the input chip packge file "chippkg_file" and floorplan file
 * "flrplan_file", read in all the package and floorplan information 
 * to parameter "components", which will be used later to construct
 * the ISAC class; 
 * parser() function does not require the user to specify the total 
 * number of layers in first line of the package file, the parser will
 * automatically calculate the total number of layers based on the input
 * files */
void parser(vector<isac::Component_type> & components, 
	const char *chippkg_file, const char *flrplan_file){

	fstream chipfile;
	fstream flrplanfile;

    /*Open chip package and floorplan files. Give error information
     * and exit the program if one/two of these two files do not exist*/
	chipfile.open(chippkg_file, ios::in);
	flrplanfile.open(flrplan_file, ios::in);
    if(!chipfile) {
        cerr<<"error: can not open chip package definition file.\n";
        exit(1);
	}
    if(!flrplanfile) {
        cerr<<"error:can not open flrplan definition file.\n";
        exit(1);
	}

	double zpos = 0.0;
	
	vector<int> pelts;
	unsigned int power_layer_num = 0;
	bool flag = true;
	int comp_num = 0;
	int layer_num = 0;
	int player = 0;

    /* Read all the information from chip package file and floorplan 
     * files to components */
	while (!chipfile.eof() && flag){
		char layer_type;
		flag = chipfile >> layer_type ;
		if (!flag){
			continue;
		}
		double size[DIM_C_];			
		chipfile >> size[X_];
		chipfile >> size[Y_];
		chipfile >> size[Z_];
		double cond;
		chipfile >> cond;
        vector<Element_type> cond_map;
        bool homog_cond = true;
        //Check to see if a non-homogeneous thermal conductivity for this layer
        //is desired (indicated by a negative number supplied in the chipfile)
        if(cond < 0) {
            //If so, the filename containing the thermal conductivity map should
            //be supplied
            //Read the thermal conductivity filename
            string thcondfilename;
            chipfile >> thcondfilename;
            fstream thcondfile;
            //Open the file, and give error and exit if the file doesn't exist
            thcondfile.open(thcondfilename.c_str(), ios::in);
            if(!thcondfile) {
                cerr << "error: a heterogeneous thermal conductivity ";
                cerr << "is desired, but the thermal conductivity map file ";
                cerr << "could not be opened.\n";
                exit(1);
            }
            bool thcondflag = true;
            //Read the thermal conductivity map file
            //First, read the bulk conductivity
            thcondfile >> cond;
            //Next, read the locations that have a different conductivity than
            //the bulk, and their values
            while(!thcondfile.eof()) {
                Element_type elt;
                thcondflag = thcondfile >> elt.center[X_];
                //Check if the file has ended
                if(!thcondflag) {
                    continue;
                }
                thcondfile >> elt.center[Y_];
                thcondfile >> elt.size[X_];
                thcondfile >> elt.size[Y_];
                thcondfile >> elt.value;
                //Check the sanity of the thermal conductivity value supplied
                if(elt.value < 0) {
                    cerr << "error: A negative thermal conductivity is ";
                    cerr << "given in the thermal conductivity map file.\n";
                    exit(1);
                }
                cond_map.push_back(elt);
            }
            thcondfile.close();    
            //If the thermal conductivity map contains at least one element,
            //then the thermal conductivity of this layer is non-homogeneous
            //set the corresponding flag to indicate that
            if(cond_map.size() > 0) {
                homog_cond = false;
            }
        }
		double cap;
		chipfile >> cap;
        //Do the same for the heat capacity
        vector<Element_type> cap_map;
        bool homog_cap = true;
        //Check to see if a non-homogeneous heat capacity for this layer
        //is desired (indicated by a negative number supplied in the chipfile)
        if(cap < 0) {
            //If so, the filename containing the heat capacity map should
            //be supplied
            //Read the heat capacity filename
            string heatcapfilename;
            chipfile >> heatcapfilename;
            fstream heatcapfile;
            //Open the file, and give error and exit if the file doesn't exist
            heatcapfile.open(heatcapfilename.c_str(), ios::in);
            if(!heatcapfile) {
                cerr << "error: a heterogeneous heat capacity ";
                cerr << "is desired, but the heat capacity map file ";
                cerr << "could not be opened.\n";
                exit(1);
            }
            bool heatcapflag = true;
            //Read the heat capacity map file
            //First, read the bulk heat capacity
            heatcapfile >> cap;
            //Next, read the locations that have a different heat capacity than
            //the bulk, and their values
            while(!heatcapfile.eof()) {
                Element_type elt;
                heatcapflag = heatcapfile >> elt.center[X_];
                //Check if the file has ended
                if(!heatcapflag) {
                    continue;
                }
                heatcapfile >> elt.center[Y_];
                heatcapfile >> elt.size[X_];
                heatcapfile >> elt.size[Y_];
                heatcapfile >> elt.value;
                //Check the sanity of the heat capacity value supplied
                if(elt.value < 0) {
                    cerr << "error: A negative heat capacity is ";
                    cerr << "given in the heat capacity map file.\n";
                    exit(1);
                }
                cap_map.push_back(elt);
            }
            heatcapfile.close();
            //If the heat capacity map contains at least one element,
            //then the heat capacity of this layer is non-homogeneous
            //set the corresponding flag to indicate that
            if(cap_map.size() > 0) {
                homog_cap = false;
            }
        }
            

        int par[DIM_C_];
        //Read the partitoning information if this is not a power layer
        if (layer_type != 'p') {
            chipfile >> par[X_];
            chipfile >> par[Y_];
            chipfile >> par[Z_];
        }
		//if (!(comp_num == 0 && layer_num == 0)){
        /* A component only contains one power layer, therefore, a 
         * new power layer means the beginning of a new component */
		if ((layer_type == 'p') && (player == 1)){
			comp_num ++;
			layer_num = 0;
			player = 0;
		}
        
        /* A material/interface layer with different size(s) than that of 
         * its previous layer (material layer or interface layer) also 
         * means the beginning of a new component */
        else if ( (layer_type != 'p') && 
        (!((comp_num == 0) && (layer_num == 0))) ){
            if (components[comp_num].layers[layer_num-1].property != POWER_){
			    if ((fabs(size[X_] - 
                components[comp_num].layers[layer_num-1].size[X_])>FMIN_) || 
			    (fabs(size[Y_] - 
                components[comp_num].layers[layer_num-1].size[Y_])>FMIN_)){
				    comp_num ++;
				    layer_num = 0;
				    player = 0;
                }
			}
		}
		//}
		components.resize(comp_num + 1);
		components[comp_num].layers.resize(layer_num + 1);

		Layer_type & cur_layer = components[comp_num].layers[layer_num];
		switch (layer_type){
			case 'p':{
				cur_layer.property = POWER_;
				player ++;
				break;
			}
			case 'i':{
				cur_layer.property = INTERFACE_;
				break;
			}
			case 'm':{
				cur_layer.property = MATERIAL_;
				break;
			}
			default:
				cerr << "layer type is not defined " << endl;
				exit(1);
		}
		for (int ii = 0; ii < DIM_C_; ++ii){
			cur_layer.size[ii] = size[ii];
			cur_layer.coor[ii] = 0.0;
            cur_layer.par[ii] = par[ii];
		}
		cur_layer.coor[Z_] = zpos + size[Z_]/2.0;
		zpos += size[Z_];
		cur_layer.cond = cond;
        cur_layer.homog_cond = homog_cond;
        cur_layer.cond_map = cond_map;
		cur_layer.cap = cap;
        cur_layer.homog_cap = homog_cap;
        cur_layer.cap_map = cap_map;
		if (layer_type == 'p'){
            /* The thickness of power layer must be 0 */
			if (size[Z_] > FMIN_){
				cerr << "The thickness of the power layer is not 0" << endl;
				exit(1);
			}
            bool flrplanflag = true;
            while (!flrplanfile.eof()){
				double x_min;
				flrplanflag = flrplanfile >> x_min; 
		        if (!flrplanflag){
			        continue;
		        }
				double x_max;
				flrplanfile >> x_max;
				double y_min;
				flrplanfile >> y_min;
				double y_max;
				flrplanfile >> y_max;
                /* Check the sanity of the floorplan file */
				if (!(x_max > x_min )){
					cerr << "error: The length in x dimension for some element " ;
					cerr << " in floorplan " << power_layer_num << " is negtive ";
					exit(1);
				}
				if (!(y_max > y_min )){
					cerr << "error: The length in y dimension for some element " ;
					cerr << " in floorplan " << power_layer_num << " is negtive ";
					exit(1);
				}
                Element_type elt;
                elt.size[X_]=(x_max-x_min);
				elt.size[Y_] = (y_max - y_min);
				elt.size[Z_] = 0.0;
				elt.center[X_] = (x_min + (x_max - x_min)/2.0);
				elt.center[Y_] = (y_min + (y_max - y_min)/2.0);
				elt.center[Z_] = cur_layer.coor[Z_];
                cur_layer.power_profile.push_back(elt); 
			}
			power_layer_num ++;
		}	
		layer_num ++;
 	}

	chipfile.close();
	flrplanfile.close();
}

/*===========================================================================*/
/* Generate the intial partition for "chip_3D" based on the information from 
 * "input_components"(one partition for each component). Later, each layer
 * inside one component gets the same partition. */
void gene_partition(ChipPkg_type & chip_3D, 
	const vector<Component_type> & input_components){

    /* Copy the layers from input_components to chip_3D */
    //copy_comps(chip_3D,input_components); 

	/* Do the unit conversion here (meter to millimeter) */	
	unit_conversion(chip_3D);
	/* Finish doing the unit conversion here (meter to millimeter) */	

    /* Generate components */
    gene_comps_par(chip_3D, input_components); 

	/* Generate the basic_partition based on the smallest input
    * power sizes info.*/
    /* The components are partitioned in order of their sizes
     * The smallest input power sizes are used as reference to
     * decide the partition number of each component */
	par_grid * basicpar_grid = & (chip_3D.basic_partition);
	pair<double, double> min_comp_xysizes;
	pair<double, double> min_pelts_xysizes;
	pair<int, int> basic_divs;
	basic_par_para(min_comp_xysizes, min_pelts_xysizes, basic_divs,
		chip_3D.components);
	gene_basicpar_grid(basicpar_grid, min_comp_xysizes);
	int order = 0;
	vector<par_logic_pos> par_pos;
    //Force an initial partition of 1x1 for all the layers
    //The actual initial parition will be applied later using the
    //chip-package information supplied by the user
    int par_no = 1;
    double silicon_size_x = min_comp_xysizes.first; 
    double silicon_size_y = min_comp_xysizes.second; 
    min_pelts_xysizes.first = silicon_size_x/par_no;
    basic_divs.first = par_no;
    min_pelts_xysizes.second = silicon_size_y/par_no;
    basic_divs.second = par_no;
    gene_homo_child(basicpar_grid, &(chip_3D.isot), &(chip_3D.cont), 
		par_pos, basic_divs.first, basic_divs.second, order);

    vector<pair<double, double> > sorted_comp_sizes;
    vector<int> comp_order;
	sort_comp_sizes(chip_3D.components, sorted_comp_sizes, comp_order);
	/* Partition the components */
	vector<int> base_divs;
	base_divs.push_back(basic_divs.first);
	base_divs.push_back(basic_divs.second);
	par_grid * ref_grid = &(chip_3D.basic_partition);
    int comp_num = input_components.size();
	par_grid * initpar_grid = &(chip_3D.init_partition);
	pair<double, double> ref_layer_sizes = min_comp_xysizes; 
    for (int i = 0; i < comp_num; ++i){
        int index = comp_order[i];
        if (i > 0){
	        ref_layer_sizes =  
                sorted_comp_sizes[comp_order[i-1]]; 
        }
		/* Find the partition granularity for the current child(current_grid) 
		 * of init_partition */
		vector<int> cur_divs = gene_divs(min_pelts_xysizes,
			ref_layer_sizes, chip_3D.components[index], 
			base_divs[0], base_divs[1]);	
		/* Generate the children of the current_grid */
		gene_homo_child(initpar_grid->int_chi_array(index),&(chip_3D.isot),
			&(chip_3D.cont), par_pos, cur_divs[0], cur_divs[1], order);
		/* Adjust the sizes/coors of the according partition */
		adj_init_par(initpar_grid->int_chi_array(index), ref_grid,
			par_pos, order); 
		for (int div_l = 0; div_l < int(base_divs.size()); ++div_l){
			base_divs[div_l] = cur_divs[div_l];
		}
		ref_grid = initpar_grid->int_chi_array(index);
		order++;
	}

	/* Form the neighbors in Z dimension for the current leave nodes  
     * (the initial partition of each component) * init_partition in chip_3D */
	form_Z_neighbor(chip_3D);

	/* Generate the simple info. for power_grid(only sizes, coors, no neighbor info)
	 * under ChipPke_type, it may contains more than one input power layer */
    gene_pgrid(chip_3D, input_components);
}

/*===========================================================================*/
/* Copy the layers from input_components to ChipPkg chip_3D */
void copy_comps(ChipPkg_type & chip_3D, 
	const vector<Component_type> & input_components){
	
    int comp_num = input_components.size();

	chip_3D.components.resize(comp_num);
	for (int i = 0; i < comp_num; ++i){
		int layer_num = input_components[i].layers.size();
		for (int j = 0; j < layer_num; ++j){
			copy_layer(chip_3D.components[i].layers[j], 
				input_components[i].layers[j]);
		}
	}
}

/*===========================================================================*/
/* Generate the components grids in init_partition in "chip_3D" based on 
 * information from "input_components" */
void gene_comps_par(ChipPkg_type & chip_3D, 
	const vector<Component_type> & input_components){

    int comp_num = input_components.size();
    /* Generate the top grid (chip package grid ) and component 
     * grids */
	par_grid * initpar_grid = &(chip_3D.init_partition);
	isotcont_init(chip_3D.isot, chip_3D.cont);	
	gene_chip_grid(initpar_grid, &(chip_3D.isot), 
		&(chip_3D.cont));
	for (int i = 0; i < comp_num; ++i){
		par_grid * cmp = new par_grid;
		initpar_grid->int_chi_array_assign(cmp);
	}
	/* Finish generating the component(init_partition's direct children) grids*/

	/* Generate the sizes info. for the grids representing the components */
	for (int i = 0; i < comp_num; ++i){
		initpar_grid -> int_chi_array(i)->sizes_assign
			(X_, chip_3D.components[i].layers[0].size[X_]);
		initpar_grid -> int_chi_array(i)->sizes_assign
			(Y_, chip_3D.components[i].layers[0].size[Y_]);
	}	
}
	
/*===========================================================================*/
/* Generate the simple info. for power_grid (only sizes, coors, no neighbor 
 * info) in "chip_3D", neighbor information is generated later */
void gene_pgrid(ChipPkg_type & chip_3D, 
	const vector<Component_type> & input_components){

    int comp_num = input_components.size();
    for (int i = 0; i < comp_num; ++i){
        par_grid * a = new par_grid;
		a -> par_assign(&(chip_3D.power_grid));
		chip_3D.power_grid.int_chi_array_assign(a);
		for (int j = 0; j < int(chip_3D.components[i].layers.size()); ++j){
			if (chip_3D.components[i].layers[j].property == POWER_){
				gene_simp_power_grid(a, &(chip_3D.isot), &(chip_3D.cont),
					chip_3D.components[i].layers[j], i); 
				continue;	
			}

		} 
	}
}

/*===========================================================================*/
/* Generate the direct children of the init_partition in "chip_3D" */
void gene_initpar_children(ChipPkg_type & chip_3D){

	int comp_num = chip_3D.components.size();
	par_grid * initpar_grid = &chip_3D.init_partition;
	int comp_level = 1;
	double sizez = 0.0;
	double osizez = 0.0;

    for (int i = 0; i < comp_num; ++i){
		Component_type & cur_comp = chip_3D.components[i];
		par_grid * cur_grid = initpar_grid->int_chi_array(i);
		cur_grid -> id_assign(i);
		cur_grid -> levelid_assign(comp_level, i);
		cur_grid -> act_level_assign(comp_level);
		cur_grid -> par_assign(initpar_grid);
		int layer_num = cur_comp.layers.size();
		osizez = sizez;
		for (int j = 0; j < layer_num; ++j){
			sizez += cur_comp.layers[j].size[Z_];
		}
		cur_grid -> sizes_assign(cur_comp.layers[0].size[X_],
			cur_comp.layers[0].size[Y_], sizez);	
		cur_grid -> coors_assign(0.0, 0.0, osizez + sizez/2.0);
    	cur_grid->nei_assign(X_NEG_, &chip_3D.isot);
    	cur_grid->nei_assign(X_PLU_, &chip_3D.isot);
    	cur_grid->nei_assign(Y_NEG_, &chip_3D.isot);
    	cur_grid->nei_assign(Y_PLU_, &chip_3D.isot);
	}

	/* Generate the neighbors in the Z directions for components
	 * (init_paritition's children) here */
	for (int i = 0; i < comp_num; ++i){
		par_grid * cur_grid = initpar_grid->int_chi_array(i);
		if ( i == 0){
    		cur_grid->nei_assign(Z_NEG_, &chip_3D.isot);
		} 
		if ( i == (comp_num - 1)){
    		cur_grid->nei_assign(Z_PLU_, &chip_3D.isot);
		} 
		else if ((i != (comp_num - 1))){
			double cur_comp_xyarea = cur_grid -> sizes(X_) * 
				cur_grid -> sizes(Y_);
			par_grid * nex_grid = initpar_grid->int_chi_array(i+1);
			double nex_comp_xyarea = nex_grid -> sizes(X_) * 
				nex_grid -> sizes(Y_);
			cur_grid -> nei_assign(Z_PLU_, nex_grid);
			nex_grid -> nei_assign(Z_NEG_, cur_grid);
			if (cur_comp_xyarea < nex_comp_xyarea){
				cur_grid -> nei_assign(Z_PLU_, &chip_3D.isot);
			}else if (cur_comp_xyarea > nex_comp_xyarea){
				nex_grid -> nei_assign(Z_NEG_, &chip_3D.isot);
			}
		}
	}
}


/*===========================================================================*/
/* Form the neighbor information in Z direction for the init_parition 
 * children grids in "chip_3D" */
void form_Z_neighbor(ChipPkg_type & chip_3D){

	int comp_num = chip_3D.components.size();
	par_grid * initpar_grid = &chip_3D.init_partition;
	/* Form Z direction neighbor info. based on the generated power */
	int levelid = 0;
	int layer_elt_level = 2;

    for (int i = 0; i < comp_num; ++i){
		par_grid * cur_comp_grid = initpar_grid->int_chi_array(i);
		par_grid * pre_comp_grid = NULL;
		par_grid * nex_comp_grid = NULL;
		if ( i != 0){
			pre_comp_grid = initpar_grid->int_chi_array(i-1);
		}	
		if ( i != (comp_num - 1)){
			nex_comp_grid = initpar_grid->int_chi_array(i+1);
		}	
		if ( i != (comp_num - 1)){/* deal with the Element_type in between 
                                  * layers */
			int cur_comp_pelts = cur_comp_grid->int_chi_array().size();
			int nex_comp_pelts = nex_comp_grid->int_chi_array().size();
			//bool clear_flag_nex[nex_comp_pelts];
			vector<bool> clear_flag_nex;
			clear_flag_nex.resize(nex_comp_pelts);
			for (int k = 0; k < nex_comp_pelts; ++k){
				clear_flag_nex[k] = true;
			} 
			for (int j = 0; j < cur_comp_pelts; j++){
				bool clear_flag = true;
				par_grid * cur_grid = cur_comp_grid->int_chi_array(j);
				cur_grid -> levelid_assign(layer_elt_level, levelid);
				levelid ++;
				for (int k = 0; k < nex_comp_pelts; ++k){
					par_grid * test_grid = nex_comp_grid->int_chi_array(k);
					//int dir_flag = cur_grid -> nei_check(test_grid);
					int dir_flag = cur_grid -> xyintersect_check(test_grid);
					//if (dir_flag == Z_PLU_){ 
					if (dir_flag){ 
						if (clear_flag){
							cur_grid->nei(Z_PLU_).clear();
							clear_flag = false;
						}
						if (clear_flag_nex[k]){
							test_grid->nei(Z_NEG_).clear();
							clear_flag_nex[k] = false;
						}
						cur_grid->nei_assign(Z_PLU_, test_grid);
						test_grid->nei_assign(Z_NEG_, cur_grid);
					}
				} 
			}
		}
		if ( i == (comp_num - 1)){
            //i == (comp_num - 1) indicates that this is the last component
            //(usually the heat sink), so it should have the constant
            //temperature boundary
			int cur_comp_pelts = cur_comp_grid->int_chi_array().size();
			for (int j = 0; j < cur_comp_pelts; j++){
				par_grid * cur_grid = cur_comp_grid->int_chi_array(j);
				cur_grid -> levelid_assign(layer_elt_level, levelid);
				levelid ++;
				cur_grid->nei(Z_PLU_).clear();
				cur_grid->nei_assign(Z_PLU_, &chip_3D.cont);	
			}
		}	
	}	

}


/*===========================================================================*/
/* Generate the power grid based on the input power profile */
/* The sizes for the power_profile has already been converted to 
 * millimeter */
void gene_power_grid (par_grid * root, par_grid * isot, par_grid * cont,
    Layer_type & power_layer, int order){

	double zsize = 1.0;
	double zcoor = order * 1.0 + 0.5;
	int pelts_level = 2;

	vector<Element_type> & cur_power = power_layer.power_profile;
    int pelts_num = cur_power.size();
    /* Set the children for the root */
    for (int i = 0; i < pelts_num; ++i){
        par_grid *a = new par_grid;
        //a->nei().clear();
        a->sizes_assign(cur_power[i].size[X_],
            cur_power[i].size[Y_], zsize);
        a->coors_assign(cur_power[i].center[X_],
            cur_power[i].center[Y_], zcoor);
        a->par_assign(root);
        a->id_assign(i);
        a->act_level_assign(pelts_level);
        a -> power_assign(cur_power[i].value);
        root -> int_chi_array_assign(a);
    }

    /* Set the correct neighbor information for the children of the root
     * only one level of children exist */
    for (int i = 0; i < pelts_num; ++i){
        par_grid * cur_grid = root -> int_chi_array(i);
        for (int k = 0; k < (DIM_C_-1); ++k) {
            if ((fabs((cur_grid -> coors(k) - cur_grid->sizes(k)/2.0)
                + root->sizes(k)/2.0) < FMIN_) &&
            	(cur_grid -> coors(k) < 0)){
                cur_grid -> nei_assign(k*2, isot);
            }
            if ((fabs((cur_grid -> coors(k) + cur_grid->sizes(k)/2.0)
                - root->sizes(k)/2.0)< FMIN_ ) &&
            	(cur_grid -> coors(k) > 0)) {
                cur_grid -> nei_assign(k*2+1, isot);
            }
		}
        for (int j = 0; j < pelts_num; ++j){
            par_grid* test_grid = root -> int_chi_array(j);
			int dir_flag = cur_grid -> nei_check(test_grid);
			if (dir_flag >= 0 && dir_flag < ((DIM_C_ - 1)*2)){
				
				cur_grid -> nei_assign(dir_flag, test_grid);
			}	
        }
		
        /* Add the isolate and constant temperature neighbors in the z 
	     * direction */
		cur_grid -> nei_assign(Z_NEG_, isot);
        cur_grid -> nei_assign(Z_PLU_, cont);
    }
}


/*===========================================================================*/
/* Generate the power grid simple info.(only sizes and coors) based on the 
 * input power profile */
void gene_simp_power_grid (par_grid * root, par_grid * isot, par_grid * cont,
    Layer_type & power_layer, int order){

	double zsize = 1.0;
	double zcoor = order * 1.0 + 0.5;
	int pelts_level = 2;

	vector<Element_type> & cur_power = power_layer.power_profile;
    int pelts_num = cur_power.size();
    /* Set the children for the root */
    for (int i = 0; i < pelts_num; ++i){
        par_grid *a = new par_grid;
        //a->nei().clear();
        a->sizes_assign(cur_power[i].size[X_],
            cur_power[i].size[Y_], zsize);
        a->coors_assign(cur_power[i].center[X_],
            cur_power[i].center[Y_], zcoor);
        a->par_assign(root);
        a->id_assign(i);
        a->act_level_assign(pelts_level);
        a -> power_assign(cur_power[i].value);
        root -> int_chi_array_assign(a);
    }
}


/*===========================================================================*/
/* Generate the sizes and coors info. for the cur_grid based on 
 * "comp_xysizes" */
void gene_basicpar_grid(par_grid * cur_grid, 
    const pair<double, double> & comp_xysizes){
	
	cur_grid -> sizes_assign(comp_xysizes.first, comp_xysizes.second, 0.0);	
	cur_grid -> coors_assign(0.0, 0.0, 0.0);	

}


/*===========================================================================*/
/* Generate homogeneous children partition in x/y dimension for "parent_grid"
 * "x_num" children in X direction and "y_num" children in Y direction*/
void gene_homo_child(par_grid * parent_grid, par_grid * isot, 
    par_grid * cont, vector<par_logic_pos> & par_pos, 
    int x_num, int y_num, int order){

	double zsize = 1.0;
	double zcoor = order * 1.0 + 0.5;
 
	double par_sizex = parent_grid -> sizes(X_);	
	double par_sizey = parent_grid -> sizes(Y_);	
	double chi_sizex = par_sizex/x_num;
	double chi_sizey = par_sizey/y_num;

	par_logic_pos logic_pos;
	logic_pos.divs_.push_back(x_num);
	logic_pos.divs_.push_back(y_num);
	logic_pos.sizes_.push_back(parent_grid->sizes(X_));
	logic_pos.sizes_.push_back(parent_grid->sizes(Y_));

	int id = 0;
/* The sequence of the generated elements corresponding to the position
 * information is as follows(describe according to the sequence below)

00 01 02 .. 0N 10 11 ... 1N ...  N1 N2 ... NN*/
/*
N1 N2 N3 ... NN
.
.
.
10 11 12 ...  1N
00 01 02 .... 0N
*/
    
 //   for (int j = (y_num-1); j > -y_num; j=(j-2)) {
    for (int j = (-y_num+1); j < y_num; j=(j+2)) {
        for (int i = (-x_num + 1); i < x_num; i=(i+2)) {
        	par_grid *a = new par_grid;
            a -> coors_assign((0.5*i)*chi_sizex, (0.5 * j) * chi_sizey, zcoor);
			a -> sizes_assign(chi_sizex, chi_sizey, zsize);
			a -> par_assign(parent_grid);
			a->id_assign(id);
			parent_grid -> int_chi_array_assign(a);
			id++;
			vector<int> t_pos;
			t_pos.resize(2);
			if (i%2!=0){
				if (i>0){
					t_pos[X_] = (i+1)/2;
				} else{
					t_pos[X_] = -(int(fabs(i))+1)/2;
				}
			}else{
				t_pos[X_] = i/2;
			}
			if (j%2!=0){
				if (j>0){
					t_pos[Y_] = (j+1)/2;
				} else{
					t_pos[Y_] = -(int(fabs(j))+1)/2;
				}
			}else{
				t_pos[Y_] = j/2;
			}
			logic_pos.lpos_.push_back(t_pos);
			logic_pos.grid_.push_back(a);
        }
    }

    /* Set the correct neighbor information in X/Y dimensions for the 
     * children of "parent_grid", only one level of children exist */
	int pelts_num = x_num * y_num;
    for (int i = 0; i < pelts_num; ++i){
        par_grid * cur_grid = parent_grid -> int_chi_array(i);
        for (int k = 0; k < (DIM_C_-1); ++k) {
            cur_grid -> nei_assign(k*2, isot);
            cur_grid -> nei_assign(k*2+1, isot);
		}
        assert(cur_grid->id() == i);
        int nei_dims = 4;
        int id_offset[4] = {-1, 1, -x_num, x_num};
        for (int j = 0; j < nei_dims; ++j){
            int nei_id = i + id_offset[j];
            if ((nei_id >= 0) && (nei_id < pelts_num)){
                bool new_nei = true;
                switch (j){
                    case 0:
                        if (((nei_id + 1)%x_num) == 0){new_nei = false;}
                        break;
                    case 1:
                        if ((nei_id%x_num) == 0){new_nei = false;}
                        break;
                    default:
                        break;
                }
                if (new_nei){
                    cur_grid -> nei(j).clear();
                    cur_grid -> nei_assign(j, 
                        parent_grid->int_chi_array(nei_id));
                }
            } 

        }
    }
    /* Add the isolate and constant temperature neighbors in Z 
	 * direction */
    for (int i = 0; i < pelts_num; ++i){
        par_grid * cur_grid = parent_grid -> int_chi_array(i);
		cur_grid -> nei_assign(Z_NEG_, isot);
        cur_grid -> nei_assign(Z_PLU_, cont);
    }

    par_pos.push_back(logic_pos);
}


/*==========================================================================*/
/* Determine the partition */
vector<int> gene_divs(pair<double, double> min_pelts_xysizes, 
pair<double, double> ref_layer_sizes, const Component_type & comp, 
int ref_xdiv, int ref_ydiv){

	vector<int> divs;	
	int cur_xdiv = int(ceil((comp.layers[0].size[X_]-ref_layer_sizes.first)
		/min_pelts_xysizes.first));
	int cur_ydiv = int(ceil((comp.layers[0].size[Y_]-ref_layer_sizes.second)
		/min_pelts_xysizes.second));

	if (cur_xdiv%2 != 0){
		cur_xdiv += (ref_xdiv + 1);
	}else{
		cur_xdiv += ref_xdiv;
	}	
	if (cur_ydiv%2 != 0){
		cur_ydiv += (ref_ydiv + 1);
	}else{
		cur_ydiv += ref_ydiv;
	}

	divs.push_back(cur_xdiv);
	divs.push_back(cur_ydiv);

	return 	divs;
}


/*==========================================================================*/
/* Adjust the size and coordinate information and size information of the 
 * children if necessary (when two layers have different sizes, the init_partition
 * of the layers with larger size needs to be adjusted based on the information
 * of the smallest layer) */
/* Only adjust the sizes in the x/y dimension since this function
 * will only be used by init_partition (Z size is 0)*/
void adj_init_par(par_grid * adj_grid, par_grid * ref_grid,
    int ref_xelt_num, int ref_yelt_num, int adj_xelt_num,
    int adj_yelt_num){
	
	double ref_sizex = ref_grid -> sizes(X_);
	double ref_sizey = ref_grid -> sizes(Y_);
	double adj_sizex = adj_grid -> sizes(X_);
	double adj_sizey = adj_grid -> sizes(Y_);
	double ref_elt_sizex = ref_sizex/ref_xelt_num; 
	double ref_elt_sizey = ref_sizey/ref_xelt_num; 

//	double xsize_bound = ref_sizex/2.0;
//	double ysize_bound = ref_sizey/2.0;
	double xsize_bound = ref_xelt_num/2.0 * adj_sizex;
	double ysize_bound = ref_yelt_num/2.0 * adj_sizey;

	int if_cnt = 0;
	int ifelse_cnt = 0;
	int adj_elt_num = adj_grid -> int_chi_array().size();
	for (int i = 0; i < adj_elt_num; ++i){
		par_grid * cur_grid = adj_grid -> int_chi_array(i);
		double coorx = cur_grid -> coors(X_);
		double coory = cur_grid -> coors(Y_);
		/* If the cur_grid is within the ref_grid projected x/y size 
		 * adjust the shape of the cur_grid based on the ref_grid */
		if ((fabs(coorx) < xsize_bound) &&
			(fabs(coory) < ysize_bound)){
			cur_grid -> sizes_assign(X_, ref_elt_sizex);
			cur_grid -> sizes_assign(Y_, ref_elt_sizey);
			cur_grid -> coors_assign(X_, cur_grid -> coors(X_)*
				(ref_elt_sizex/cur_grid->sizes(X_)));
			cur_grid -> coors_assign(Y_, cur_grid -> coors(Y_)*
				(ref_elt_sizey/cur_grid->sizes(Y_)));
			if_cnt ++;
		}
		/* The grid must be out of the project x/y size area */
		else{
			if (fabs(coorx) < xsize_bound){
				int times = int(floor(cur_grid ->coors(Y_) - ysize_bound)
					/cur_grid->sizes(Y_));
				cur_grid -> sizes_assign(X_, ref_elt_sizex);
				cur_grid -> sizes_assign(Y_, 
					(adj_sizey-ref_sizey)/(adj_yelt_num-ref_yelt_num));
				cur_grid -> coors_assign(X_, cur_grid -> coors(X_)*
					(ref_elt_sizex/cur_grid->sizes(X_)));
				cur_grid -> coors_assign(Y_, ysize_bound + (2*times+1) * 
					(adj_sizey - ref_sizey)/(adj_yelt_num-ref_yelt_num)/2.0);
			} else if (fabs(coory) < ysize_bound){
				int times = int(floor(cur_grid ->coors(X_) - xsize_bound)
					/cur_grid->sizes(Y_));
				cur_grid -> sizes_assign(Y_, ref_elt_sizey);
				cur_grid -> sizes_assign(X_, 
					(adj_sizex-ref_sizex)/(adj_xelt_num-ref_xelt_num));
				cur_grid -> coors_assign(Y_, cur_grid -> coors(Y_)*
					(ref_elt_sizey/cur_grid->sizes(Y_)));
				cur_grid -> coors_assign(X_, xsize_bound + (2*times+1) * 
					(adj_sizex - ref_sizex)/(adj_xelt_num-ref_xelt_num)/2.0);
			}else {
				int x_times = int(floor(cur_grid ->coors(X_) - xsize_bound)
					/cur_grid->sizes(X_));
				int y_times = int(floor(cur_grid ->coors(Y_) - ysize_bound)
					/cur_grid->sizes(Y_));
				cur_grid -> sizes_assign(X_, 
					(adj_sizex-ref_sizex)/(adj_xelt_num-ref_xelt_num));
				cur_grid -> sizes_assign(Y_, 
					(adj_sizey-ref_sizey)/(adj_yelt_num-ref_yelt_num));
				cur_grid -> coors_assign(X_, xsize_bound + (2*x_times+1) * 
					(adj_sizex - cur_grid->sizes(X_))/
					(adj_xelt_num-ref_xelt_num)/2.0);
				cur_grid -> coors_assign(Y_, ysize_bound + (2*y_times+1) * 
					(adj_sizey - cur_grid->sizes(Y_))/
					(adj_yelt_num-ref_yelt_num)/2.0);
			}
			ifelse_cnt ++;
		} 
	}
}


/*==========================================================================*/
/* Adjust the size and coordinate information and size information of the 
 * children if necessary (when two layers has different sizes, the init_partition
 * of the layers with larger size needs to be adjusted based on the information
 * of the smallest layer) */
/* Only consider adjust the sizes in the x/y dimension since this function
 * will only be used by init_partition (Z size is 0) */
void adj_init_par(par_grid * adj_par, par_grid * ref_par,
    const vector<par_logic_pos> & par_pos, int order){

	const par_logic_pos & ref_pos = par_pos[order];
	const par_logic_pos & cur_pos = par_pos[order+1];
	int adj_eltnum = adj_par -> int_chi_array().size();
	int ref_eltnum = ref_par -> int_chi_array().size();
	assert((order+1) <= int(par_pos.size()));
	double xsize_bound = ref_pos.sizes_[X_];
	double ysize_bound = ref_pos.sizes_[Y_];
	double adj_sizex = adj_par->sizes(X_);
	double adj_sizey = adj_par->sizes(Y_);
	double ref_sizex = ref_par->sizes(X_);
	double ref_sizey = ref_par->sizes(Y_);

    for (int i = 0; i < adj_eltnum; ++i){
		par_grid * cur_grid = adj_par -> int_chi_array(i);

		if ((fabs(cur_pos.lpos_[i][X_]) <= ref_pos.divs_[X_]/2) &&
			(fabs(cur_pos.lpos_[i][Y_]) <= ref_pos.divs_[Y_]/2)){
			for (int j = 0; j < ref_eltnum; ++j){
				if ((ref_pos.lpos_[j][X_] == cur_pos.lpos_[i][X_])&&
					(ref_pos.lpos_[j][Y_] == cur_pos.lpos_[i][Y_])){

					par_grid * ref_grid = ref_pos.grid_[j];
					cur_grid -> sizes_assign(ref_grid->sizes(X_), 
						ref_grid->sizes(Y_), cur_grid->sizes(Z_));
					cur_grid -> coors_assign(ref_grid->coors(X_), 
						ref_grid->coors(Y_), cur_grid->coors(Z_));
				}
			}
		}
		else{
			if (fabs(cur_pos.lpos_[i][X_]) <= ref_pos.divs_[X_]/2){
				int times = int((fabs(cur_grid ->coors(Y_))-ysize_bound/2.0)
					/cur_grid->sizes(Y_))+1;
				int ref_min_divy = (ref_pos.divs_[Y_]+1)%2;
				par_grid * ref_grid = NULL;
				for (int k = 0; k < ref_eltnum; ++k){
					if ((ref_pos.lpos_[k][X_] == cur_pos.lpos_[i][X_]) &&
					 	(ref_pos.lpos_[k][Y_] == ref_min_divy)){ 
						ref_grid = ref_pos.grid_[k];
						continue;
					}
				} 
				cur_grid -> sizes_assign(X_, ref_grid->	sizes(X_));
				cur_grid -> sizes_assign(Y_, 
					(adj_sizey-ref_sizey)/(cur_pos.divs_[Y_]-ref_pos.divs_[Y_]));
				cur_grid -> coors_assign(X_, ref_grid->	coors(X_));
				double coory = ysize_bound/2.0 + (2*times-1) * 
					(adj_sizey - ref_sizey)/((cur_pos.divs_[Y_]-ref_pos.divs_[Y_]))/2.0;
				if (cur_pos.lpos_[i][Y_] < 0){
					coory = -coory;
				}
				cur_grid -> coors_assign(Y_, coory); 

			} else if (fabs(cur_pos.lpos_[i][Y_]) <= ref_pos.divs_[Y_]/2){
				int times = int((fabs(cur_grid ->coors(X_))-xsize_bound/2.0)
					/cur_grid->sizes(X_))+1;
				int ref_min_divx = (ref_pos.divs_[X_]-1)%2; 
				par_grid * ref_grid = NULL;	
				for (int k = 0; k < ref_eltnum; ++k){
					if ((ref_pos.lpos_[k][Y_] == cur_pos.lpos_[i][Y_]) &&
					 	(ref_pos.lpos_[k][X_] == ref_min_divx)){ 
						ref_grid = ref_pos.grid_[k];
						continue;
					}
				} 
				cur_grid -> sizes_assign(Y_, ref_grid->sizes(Y_));
				cur_grid -> sizes_assign(X_, 
					(adj_sizey-ref_sizex)/(cur_pos.divs_[X_]-ref_pos.divs_[X_]));
				cur_grid -> coors_assign(Y_, ref_grid->coors(Y_));
				double coorx = xsize_bound/2.0 + (2*times-1) * 
					(adj_sizex - ref_sizex)/((cur_pos.divs_[X_]-ref_pos.divs_[X_]))/2.0;
				if (cur_pos.lpos_[i][X_] < 0){
					coorx = -coorx;
				}
				cur_grid -> coors_assign(X_, coorx); 

			}else{
				int times_x = int((fabs(cur_grid ->coors(X_)) - xsize_bound/2.0)
					/cur_grid->sizes(X_))+1;
				int times_y = int((fabs(cur_grid ->coors(Y_)) - ysize_bound/2.0)
					/cur_grid->sizes(Y_))+1;
				cur_grid -> sizes_assign(X_, 
					(adj_sizey-ref_sizex)/(cur_pos.divs_[X_]-ref_pos.divs_[X_]));
				cur_grid -> sizes_assign(Y_, 
					(adj_sizey-ref_sizey)/(cur_pos.divs_[Y_]-ref_pos.divs_[Y_]));
				double coorx = xsize_bound/2.0 + (2*times_x-1) * 
					(adj_sizex - ref_sizex)/((cur_pos.divs_[X_]-ref_pos.divs_[X_]))/2.0;
				if (cur_pos.lpos_[i][X_] < 0){
					coorx = -coorx;
				}
				double coory = ysize_bound/2.0 + (2*times_y-1) * 
					(adj_sizey - ref_sizey)/((cur_pos.divs_[Y_]-ref_pos.divs_[Y_]))/2.0;
				if (cur_pos.lpos_[i][Y_] < 0){
					coory = -coory;
				}
				cur_grid -> coors_assign(X_, coorx); 
				cur_grid -> coors_assign(Y_, coory);
			}
		}
	}
}


/*===========================================================================*/
/* Sort all the components by their x/y sizes */
void sort_initpar_size(const vector<Component_type> & comps, 
    multiset<double> & xsize_set, multiset<double> & ysize_set){ 

	int comp_num = comps.size();

	multiset<double>::iterator x_iter = xsize_set.begin();
	multiset<double>::iterator y_iter = ysize_set.begin();
	for (int i = 0; i < comp_num; ++i){
		xsize_set.insert(comps[i].layers[0].size[X_]);	
		ysize_set.insert(comps[i].layers[0].size[Y_]);	
	}
}
  
/*===========================================================================*/
/* Sort all the components by their x/y sizes */
void sort_comp_sizes(const vector<Component_type> & comps, 
    vector<pair<double, double> > & sorted_comp_sizes, 
    vector<int> & comp_order){

	int comp_num = comps.size();
    vector<double> comps_xsize;
    for (int i = 0; i < comp_num; ++i){
        comps_xsize.push_back(comps[i].layers[0].size[X_]);
    }

    double ref_minsize = FMIN_;
    int loop = comp_num;
    while (loop){
        double min_xsize = FMAX_;
        int index = -1;
        //for (unsigned int i = 0; i < comps_xsize.size(); ++i){
        for (int i = 0; i < comp_num; ++i){
            //if ((comps_xsize[i]<min_xsize) & 
            //    (comps_xsize[i] > ref_minsize)){
            if (comps_xsize[i]<min_xsize) {
                //Make sure this component hasn't been included before by 
                //going through the indices already pushed in
                bool already_used = false;
                for (unsigned int j = 0; j < comp_order.size(); j++) {
                    if(comp_order[j] == i) {
                        already_used = true;
                        break;
                    }
                }
                if(already_used) continue;

                min_xsize = comps_xsize[i];
                index = i;
            }
        }
        sorted_comp_sizes.push_back(pair<double, double>
            (comps[index].layers[0].size[X_],
            comps[index].layers[0].size[Y_]));
        comp_order.push_back(index);
        ref_minsize = comps[index].layers[0].size[X_];
        loop--;
    }
}

/*===========================================================================*/
/* Find the smallest x/y sizes, as well as the number of grids (x_num, y_num) 
 * in x/y dimension of the basic partition among the components */
void basic_par_para(pair<double, double> & min_sizes, 
    pair<double, double> & min_psizes, pair<int,int> & basic_divs, 
    const vector<Component_type> & comps){

	int comp_num = comps.size();

	vector<double> max_sizes(2, FMIN_);
	min_sizes.first = FMAX_;
	min_sizes.second = FMAX_;
	for (int i = 0; i < comp_num; ++i){
		min_sizes.first = min(min_sizes.first, comps[i].layers[0].size[X_]);
		min_sizes.second = min(min_sizes.second, comps[i].layers[0].size[Y_]);
		max_sizes[0] = max(max_sizes[0], comps[i].layers[0].size[X_]);
		max_sizes[1] = max(max_sizes[1], comps[i].layers[0].size[Y_]);
	}

	min_psizes.first = FMAX_;
	min_psizes.second = FMAX_;
	for (int i = 0; i < comp_num; ++i){
		bool has_p = false;
		int index = -1;
		for (int j = 0; j < int(comps[i].layers.size()); ++j){	
			if (comps[i].layers[j].property == POWER_){
				has_p = true;
				index = j;
				continue;
			}
		}
		if (has_p){
			const vector<Element_type> & p_profile = comps[i].layers[index].
				power_profile;
			for (int k = 0; k < int(p_profile.size());++k){	
				min_psizes.first = min(min_psizes.first, 
					p_profile[k].size[X_]); 
				min_psizes.second = min(min_psizes.second, 
					p_profile[k].size[Y_]);
			} 
		}
	}

    basic_divs.first = int(ceil(min_sizes.first/min_psizes.first));
    basic_divs.second = int(ceil(min_sizes.second
        /min_psizes.second));

	int div_max = 9;
	basic_divs.first = min(basic_divs.first, int(min_sizes.first/
		min_psizes.first * 9));
	basic_divs.second = min(basic_divs.second, int(min_sizes.first/
		min_psizes.first * 9));

/* The equivalent min_psizes used for the paritition */	
	min_psizes.first = max(max_sizes[0]/div_max, min_psizes.first);
	min_psizes.second = max(max_sizes[1]/div_max, min_psizes.second);
}


/*===========================================================================*/
/* Generate the chip grid information(root grid) for the chip package */
void gene_chip_grid (par_grid * root, par_grid * isot, par_grid * cont){

	root->id_assign(0);
    root->act_level_assign(0);
    root->levelid_assign(0, 0);
    root->nei_assign(X_NEG_, isot);
    root->nei_assign(X_PLU_, isot);
    root->nei_assign(Y_NEG_, isot);
    root->nei_assign(Y_PLU_, isot);
    root->nei_assign(Z_NEG_, isot);
    root->nei_assign(Z_PLU_, cont);
}


/*============================================================================*/
/* Convert the unit from meter to millimeter for the "chip_3D" parameters */
void unit_conversion(ChipPkg_type & chip_3D){
	
	int comp_num = chip_3D.components.size(); 
	for (int i = 0; i < comp_num; ++i){
		int layer_num = chip_3D.components[i].layers.size();
		for (int j = 0; j < layer_num; ++j){
			Layer_type & cur_layer = chip_3D.components[i].layers[j];
			for (int k = 0; k < DIM_C_; ++k){
				cur_layer.size[k] *= 1E3;
				cur_layer.coor[k] *= 1E3;
			}
			cur_layer.cond *= 1E-3;
            int cond_map_num = cur_layer.cond_map.size();
            for(int k = 0; k < cond_map_num; ++k) {
                cur_layer.cond_map[k].value *= 1E-3;
                for(int l = 0; l < DIM_C_; ++l) {
                    cur_layer.cond_map[k].size[l] *= 1E3;
                    cur_layer.cond_map[k].center[l] *= 1E3;
                }
            }
			cur_layer.cap *= 1E-9;
            int cap_map_num = cur_layer.cap_map.size();
            for(int k = 0; k < cap_map_num; ++k) {
                cur_layer.cap_map[k].value *= 1E-9;
                for(int l = 0; l < DIM_C_; ++l) {
                    cur_layer.cap_map[k].size[l] *= 1E3;
                    cur_layer.cap_map[k].center[l] *= 1E3;
                }
            }
			if (cur_layer.property == POWER_){
				vector<Element_type> & cur_power = cur_layer.power_profile;
				int pelts = cur_power.size();
				for (int l = 0; l < pelts; ++l){
					for (int m = 0; m < DIM_C_; ++m){
						cur_power[l].size[m] *= 1E3;
						cur_power[l].center[m] *= 1E3;
					}
				} 
			}
		}
	}
}


/*============================================================================*/
/* Copy informatin from one "ref_grid" to "cur_grid", while "diff" means
 * the difference for the Z coordinate for two grids */
void copy_pgrid(par_grid * cur_grid, par_grid * ref_grid, int diff){

	double coor_diff = diff * 1.0;
	int elts_num = ref_grid->int_chi_array().size();
	for (int i = 0; i < elts_num; ++i){
		par_grid * src_grid = ref_grid->int_chi_array(i);
		par_grid * dst_grid = new par_grid;
		for (int l = 0; l < DIM_C_; ++l){
			dst_grid->sizes_assign(src_grid->sizes(X_), src_grid->sizes(Y_),
				src_grid->sizes(Z_));
			dst_grid->coors_assign(src_grid->coors(X_), src_grid->coors(Y_),
				(src_grid->coors(Z_) + coor_diff));
		}
		dst_grid->act_level_assign(src_grid->act_level());
		dst_grid->id_assign(src_grid->id());
		dst_grid->par_assign(cur_grid);
		dst_grid->power_assign(0.0);
		cur_grid->int_chi_array_assign(dst_grid);
	}
	for (int i = 0; i < elts_num; ++i){
		par_grid * src_grid = ref_grid->int_chi_array(i);
		par_grid * dst_grid = cur_grid->int_chi_array(i);
		for (int l = 0; l < NEI_C_; ++l){
			int nei_num = src_grid->nei(l).size();
			for (int m = 0; m < nei_num; ++m){
 				if (src_grid->nei(l,m)->id() < 0){
					dst_grid->nei_assign(l,src_grid->nei(l,m));
				}else {
					dst_grid->nei_assign(l, cur_grid->int_chi_array
						(src_grid->nei(l,m)->id()));

				}
			}
		}
	}
}


/*===========================================================================*/
/* For a vector of bool variables, return true if one of them has a value
 * 'true', return false only when all the variables have value 'false' */
bool flag_true(const vector<bool> & flag){

    for (unsigned int i = 0; i < flag.size(); ++i){
        if (flag[i] == true){
            return true;
        }
    }

    return false;
}


/*===========================================================================*/
/* For two vectors with the same size, find out the largest difference
 * among their elements with same position */
double max_diff(const vector<double> & vec_a, const vector<double> & vec_b){

    int size = vec_a.size();
    double max_diff = 0;
    for (int i = 0; i < size; ++i){
        max_diff = max(max_diff, fabs(vec_a[i] - vec_b[i]));
    }

    return max_diff;
}


/*===========================================================================*/
/* For two vectors with the same size, find out the largest difference
 * among their elements with position difference of "start_id" */
double max_diff(const vector<double> & vec_a, const vector<double> & vec_b, 
	int start_id){

    int size = vec_a.size();
    double max_diff = 0;
    for (int i = 0; i < size; ++i){
        max_diff = max(max_diff, fabs(vec_a[i] - vec_b[i+start_id]));
    }

    return max_diff;
}


/*===========================================================================*/
/* Calculate the area intersect in x/y dimension for a grid and an Element */
double area_intersect(const par_grid * grid, const Element_type & elt){

	double iarea = 1.0;

    for (int i = 0; i < (DIM_C_ - 1); ++i){
		const double this_min = grid->coors(i) - grid->sizes(i)/2.0;
		const double this_max = grid->coors(i) + grid->sizes(i)/2.0;
		const double b_this_min = elt.center[i] - elt.size[i]/2.0;
		const double b_this_max = elt.center[i] + elt.size[i]/2.0;
		
		const double min_p = max(this_min, b_this_min);	
		const double max_p = min(this_max, b_this_max);

		if (max_p < min_p){
			return 0.0;
		}	

		iarea = iarea * (max_p - min_p);
	}

    return iarea;	
}


/*===========================================================================*/
/* Calculate the area intersect in x/y dimension for two grids and an Element */
double area_intersect(const par_grid * grid_a, const par_grid * grid_b, 
    const Element_type & elt){
    
    double iarea = 1.0;
    for (int i = 0; i < (DIM_C_ - 1); ++i){
		const double a_this_min = grid_a->coors(i) - grid_a->sizes(i)/2.0;
		const double a_this_max = grid_a->coors(i) + grid_a->sizes(i)/2.0;
		const double b_this_min = grid_b->coors(i) - grid_b->sizes(i)/2.0;
		const double b_this_max = grid_b->coors(i) + grid_b->sizes(i)/2.0;
		const double c_this_min = elt.center[i] - elt.size[i]/2.0;
		const double c_this_max = elt.center[i] + elt.size[i]/2.0;
        double min_p = max(a_this_min, b_this_min);
        min_p = max(min_p, c_this_min); 
        double max_p = min(a_this_max, b_this_max);
        max_p = min(max_p, c_this_max); 
        
        if (max_p < min_p){
            return 0.0;
        }
        iarea = iarea *(max_p - min_p);
    } 
    return iarea;
}


/*===========================================================================*/
/* Calculate the area intersect in x/y dimension for two Elements */
double area_intersect(const Element_type & elt_a, const Element_type & elt_b){
    double iarea = 1.0;
    for (int i = 0; i < (DIM_C_ - 1); ++i){
		const double a_this_min = elt_a.center[i] - elt_a.size[i]/2.0;
		const double a_this_max = elt_a.center[i] + elt_a.size[i]/2.0;
		const double b_this_min = elt_b.center[i] - elt_b.size[i]/2.0;
		const double b_this_max = elt_b.center[i] + elt_b.size[i]/2.0;
        double min_p = max(a_this_min, b_this_min);
        double max_p = min(a_this_max, b_this_max);

        if (max_p < min_p){
            return 0.0;
        }
        iarea = iarea *(max_p - min_p);
    }
    return iarea;
}

/*===========================================================================*/
void print_vec_vec(const vector<vector<double> > & vec){

    int isize = vec.size();

    for (int i = 0;i < isize; ++i){
        for (int j = 0; j < int(vec[i].size()); ++j){
            cout << vec[i][j] << " ";
        } cout << endl;
    } cout << "finishes_print_vec_vec " << endl;
}
       
/*===========================================================================*/
void print_eltvec_top(vector<Element_type> & elt_vec){

    int isize = elt_vec.size();

    double max_t = 0;
    double min_t = FMAX_;

    int ind_min;
    double loc, min_loc;
    Element_type tmp;

    //sort elements for gnuplot
    for (int i = 0; i < isize-1; i++)
    {
        min_loc = elt_vec[i].center[X_] + elt_vec[i].center[Y_]*1000;
        ind_min = i;
        for (int j = i+1; j < isize; j++)
        {
            loc = elt_vec[j].center[X_] + elt_vec[j].center[Y_]*1000;
            if (loc < min_loc)
            {
                min_loc = loc;
                ind_min = j;
            }
        }
        tmp = elt_vec[i];
        elt_vec[i] = elt_vec[ind_min];
        elt_vec[ind_min] = tmp;
    }

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   

    double last_y = 0;

    //Only print top layer
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- top_z)<FMIN_ )
        {
            max_t = max(max_t, elt_vec[i].value);
            min_t = min(min_t, elt_vec[i].value);
            if(fabs(elt_vec[i].center[Y_]-last_y)>FMIN_){
                cout<<endl;
            }
            cout << elt_vec[i].center[X_]<< " ";
            cout << elt_vec[i].center[Y_]<< " ";
            cout << elt_vec[i].value;
            cout << endl;
            last_y = elt_vec[i].center[Y_];

        }
    } cout << endl; 

    cout <<"max_min_t " << max_t << " " << min_t << endl;
    cout <<"finishes_print_eltvec " << endl;
}

/*===========================================================================*/
//Prints out the temperatures of elements for gnuplot
void print_eltvec_gnuplot(vector<Element_type> & elt_vec){

    int isize = elt_vec.size();

    int ind_min;
    double loc, min_loc;
    Element_type tmp;

    //sort elements for gnuplot
    for (int i = 0; i < isize-1; i++)
    {
        min_loc = elt_vec[i].center[Y_] + elt_vec[i].center[X_]*1000;
        ind_min = i;
        for (int j = i+1; j < isize; j++)
        {
            loc = elt_vec[j].center[Y_] + elt_vec[j].center[X_]*1000;
            if (loc < min_loc)
            {
                min_loc = loc;
                ind_min = j;
            }
        }
        tmp = elt_vec[i];
        elt_vec[i] = elt_vec[ind_min];
        elt_vec[ind_min] = tmp;
    }

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   

    double last_x = 0;

    //Only print top layer
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- top_z)<FMIN_ )
        {
            if(fabs(elt_vec[i].center[X_]-last_x)>FMIN_){
                cout<<endl;
            }
            cout << elt_vec[i].center[X_]<< " ";
            cout << elt_vec[i].center[Y_]<< " ";
            cout << elt_vec[i].value;
            cout << endl;
            last_x = elt_vec[i].center[X_];

        }
    } cout << endl; 
}

/*===========================================================================*/
//Prints out the heterogeneous temperature elements for gnuplot.
//Converts to homogeneous first
void print_hetero_eltvec_gnuplot(vector<Element_type> & elt_vec, int size){

    int isize = elt_vec.size();

    int ind_min;
    double loc, min_loc;
    Element_type tmp;

    //sort elements for gnuplot
    for (int i = 0; i < isize-1; i++)
    {
        min_loc = elt_vec[i].center[Y_] + elt_vec[i].center[X_]*1000;
        ind_min = i;
        for (int j = i+1; j < isize; j++)
        {
            loc = elt_vec[j].center[Y_] + elt_vec[j].center[X_]*1000;
            if (loc < min_loc)
            {
                min_loc = loc;
                ind_min = j;
            }
        }
        tmp = elt_vec[i];
        elt_vec[i] = elt_vec[ind_min];
        elt_vec[ind_min] = tmp;
    }

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   


    vector<int> top_layer_elts;

    //Get the top layer elements
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- top_z)<FMIN_ )
        {
            top_layer_elts.push_back(i);
        }
    } 

    int top_size = top_layer_elts.size();

    //Find the minimum and maximum x and y from the elements in the top layer
    double min_x, min_y, max_x, max_y; 
    min_x = elt_vec[top_layer_elts[0]].center[X_] -
            elt_vec[top_layer_elts[0]].size[X_]/2.0;
    min_y = elt_vec[top_layer_elts[0]].center[Y_] -
            elt_vec[top_layer_elts[0]].size[Y_]/2.0;
    max_x = elt_vec[top_layer_elts[0]].center[X_] +
            elt_vec[top_layer_elts[0]].size[X_]/2.0;
    max_y = elt_vec[top_layer_elts[0]].center[Y_] +
            elt_vec[top_layer_elts[0]].size[Y_]/2.0;
    for(int i = 1; i < top_size; i++) {
        double x1 = elt_vec[top_layer_elts[i]].center[X_] -
            elt_vec[top_layer_elts[i]].size[X_]/2.0;
        double y1 = elt_vec[top_layer_elts[i]].center[Y_] -
            elt_vec[top_layer_elts[i]].size[Y_]/2.0;
        double x2 = elt_vec[top_layer_elts[i]].center[X_] +
            elt_vec[top_layer_elts[i]].size[X_]/2.0;
        double y2 = elt_vec[top_layer_elts[i]].center[Y_] +
            elt_vec[top_layer_elts[i]].size[Y_]/2.0;

        if(x1 < min_x) min_x = x1;
        if(y1 < min_y) min_y = y1;
        if(x2 > max_x) max_x = x2;
        if(y2 > max_y) max_y = y2;
       
    }

    //Find the temperatures in the homogeneous grid defined by min_x, max_x,
    //min_y, max_y, and size
    double x_init = min_x;
    double y_init = min_y;
    double stepX = (max_x - min_x) / ((double) size);
    double stepY = (max_y - min_y) / ((double) size);
    double area = stepX * stepY;

    double x,y;

    vector<double> mapped;
    mapped.resize(size * size);
    
    x = x_init;
    for(int i = 0; i < size; i++) {
        y = y_init;
        for(int j = 0; j < size; j++) {
            mapped[i * size + j] = 0;
            for(int k = 0; k < top_size; k++) {
                double x1 = elt_vec[top_layer_elts[k]].center[X_] -
                    elt_vec[top_layer_elts[k]].size[X_]/2.0;
                double y1 = elt_vec[top_layer_elts[k]].center[Y_] -
                    elt_vec[top_layer_elts[k]].size[Y_]/2.0;
                double x2 = elt_vec[top_layer_elts[k]].center[X_] +
                    elt_vec[top_layer_elts[k]].size[X_]/2.0;
                double y2 = elt_vec[top_layer_elts[k]].center[Y_] +
                    elt_vec[top_layer_elts[k]].size[Y_]/2.0;

                if(areasIntersect(x, x + stepX, y, y + stepY, x1, x2, y1, y2)) {

                    min_x = max(x, x1);
                    max_x = min(x + stepX, x2);
                    min_y = max(y, y1);
                    max_y = min(y + stepY, y2);
                    //scale the temperature value according to the area
                    //intersection
                    mapped[i * size + j] += elt_vec[top_layer_elts[k]].value *
                        ((max_x - min_x) * (max_y - min_y));
                }
            }
            mapped[i * size + j] /= area;
            cout << (x + x + stepX)/2.0 << " ";
            cout << (y + y + stepY)/2.0 << " ";
            cout << mapped[i * size + j] << endl;
            y += stepY;
        }
        cout << endl;
        x += stepX;
    }
}



/*===========================================================================*/
void print_eltvec_top(std::vector<isac::Element_type> & elt_vec, const char * filename){

    int isize = elt_vec.size();

    double max_t = 0;
    double min_t = FMAX_;

    int ind_min;
    double loc, min_loc;
    isac::Element_type tmp;

    //sort elements for gnuplot
    for (int i = 0; i < isize-1; i++)
    {
        min_loc = elt_vec[i].center[X_] + elt_vec[i].center[Y_]*1000;
        ind_min = i;
        for (int j = i+1; j < isize; j++)
        {
            loc = elt_vec[j].center[X_] + elt_vec[j].center[Y_]*1000;
            if (loc < min_loc)
            {
                min_loc = loc;
                ind_min = j;
            }
        }
        tmp = elt_vec[i];
        elt_vec[i] = elt_vec[ind_min];
        elt_vec[ind_min] = tmp;
    }

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   

    double last_y = 0;

    ofstream fout;
    fout.open(filename);

    //Only print top layer
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- top_z)<FMIN_ )
        {
            max_t = max(max_t, elt_vec[i].value);
            min_t = min(min_t, elt_vec[i].value);
            if(fabs(elt_vec[i].center[Y_]-last_y)>FMIN_){
                fout<<endl;
            }
            fout << elt_vec[i].center[Y_]<< " ";
            fout << elt_vec[i].center[X_]<< " ";
            fout << elt_vec[i].value;
            fout << endl;
            last_y = elt_vec[i].center[Y_];

        }
    } fout << endl; 

    fout.close();
}

/*===========================================================================*/
void print_eltvec_top2(std::vector<isac::Element_type> & elt_vec, const char * filename){

    int isize = elt_vec.size();

    int ind_min;
    double loc, min_loc;
    isac::Element_type tmp;

    //sort elements for gnuplot
    for (int i = 0; i < isize-1; i++)
    {
        min_loc = elt_vec[i].center[X_] + elt_vec[i].center[Y_]*1000;
        ind_min = i;
        for (int j = i+1; j < isize; j++)
        {
            loc = elt_vec[j].center[X_] + elt_vec[j].center[Y_]*1000;
            if (loc < min_loc)
            {
                min_loc = loc;
                ind_min = j;
            }
        }
        tmp = elt_vec[i];
        elt_vec[i] = elt_vec[ind_min];
        elt_vec[ind_min] = tmp;
    }

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   

    ofstream fout;
    fout.open(filename);

    fout.precision(7);
    //Only print top layer
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- top_z)<FMIN_ )
        {
            fout << elt_vec[i].value;
            fout << endl;

        }
    } 

    fout.close();
}

/*===========================================================================*/
//Print the temperature of the center element in the top layer
void print_eltvec_top_center_elt_static(std::vector<isac::Element_type> & elt_vec){

    cout.precision(10);

    static int center_index = -1;

#ifdef DEBUG
    cout << "Center Index : " << center_index << endl;
#endif
    
    if(center_index == -1) {

       int isize = elt_vec.size();

       //Find the nearest element to center on top layer
       double min_x = 100000;
       double min_y = 100000;
       double min_z = 100000;

       //Find x of the nearest element to center with in top right quad
       for(int i = 0; i < isize; ++i){
           if(elt_vec[i].center[X_]<min_x && elt_vec[i].center[X_]>0 ){
               min_x = elt_vec[i].center[X_];
           }
       }

       //Find y of the nearest element to center in top right quad
       for(int i = 0; i < isize; ++i){
           if(elt_vec[i].center[Y_]<min_y && elt_vec[i].center[Y_]>0 ){
               min_y = elt_vec[i].center[Y_];
           }
       }
    
       //Find z of top layer that has the x = min_x and y = min_y
       for (int i = 0; i < isize; ++i){
           if(elt_vec[i].center[Z_]<min_z && 
               fabs(elt_vec[i].center[X_]- min_x)<FMIN_ &&
               fabs(elt_vec[i].center[Y_]- min_y)<FMIN_ ){
               min_z = elt_vec[i].center[Z_];
           }
       }

       //Get the index of the center element

       for (int i = 0; i < isize; ++i){
           if( fabs(elt_vec[i].center[Z_]- min_z)<FMIN_ )
           {
               if( fabs(elt_vec[i].center[X_]- min_x)<FMIN_ )
               {
                   if( fabs(elt_vec[i].center[Y_]- min_y)<FMIN_ )
                   {
                        center_index = i; 
                        cout << "Got the center index : " << center_index << endl;
                        cout << elt_vec[i].value - 328.15;
                   }
               }
           }
       } cout << endl;
    }
    else {
        cout << elt_vec[center_index].value - 328.15 << endl;
    }
        
}



/*===========================================================================*/
//Print the temperature of the center element in the top layer
void print_eltvec_top_center_elt(std::vector<isac::Element_type> & elt_vec){

    int isize = elt_vec.size();

    //Find the nearest element to center on top layer
    double min_x = 100000;
    double min_y = 100000;
    double min_z = 100000;

    //Find x of the nearest element to center with in top right quad
    for(int i = 0; i < isize; ++i){
        if(elt_vec[i].center[X_]<min_x && elt_vec[i].center[X_]>0 ){
            min_x = elt_vec[i].center[X_];
        }
    }

    //Find y of the nearest element to center in top right quad
    for(int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Y_]<min_y && elt_vec[i].center[Y_]>0 ){
            min_y = elt_vec[i].center[Y_];
        }
    }
    
    //Find z of top layer that has the x = min_x and y = min_y
    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<min_z && 
            fabs(elt_vec[i].center[X_]- min_x)<FMIN_ &&
            fabs(elt_vec[i].center[Y_]- min_y)<FMIN_ ){
            min_z = elt_vec[i].center[Z_];
        }
    }


    //Only print the temperature of this element
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- min_z)<FMIN_ )
        {
            if( fabs(elt_vec[i].center[X_]- min_x)<FMIN_ )
            {
                if( fabs(elt_vec[i].center[Y_]- min_y)<FMIN_ )
                {
                    cout.precision(10);                    
                    cout << elt_vec[i].value - 328.15;
                }
            }
        }
    } cout << endl;
}

/*===========================================================================*/
//Print the temperature of the elements on the diagonal on the top layer
void get_indices_top_diagonal_elts(std::vector<isac::Element_type> & elt_vec, std::vector<int> & indices){

    cout.precision(10);

    int isize = elt_vec.size();

    //Find the nearest element to center on top layer
    double min_x = 100000;
    double min_y = 100000;
    double min_z = 100000;

    //Find x of the nearest element to center with +ve x,y quad
    for(int i = 0; i < isize; ++i){
        if(elt_vec[i].center[X_]<min_x && elt_vec[i].center[X_]>0 ){
            min_x = elt_vec[i].center[X_];
        }
    }

    //Find y of the nearest element to center in top right quad
    for(int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Y_]<min_y && elt_vec[i].center[Y_]>0 ){
            min_y = elt_vec[i].center[Y_];
        }
    }
    
    //Find z of top layer that has the x = min_x and y = min_y
    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_] < min_z && 
           fabs(elt_vec[i].center[X_]- min_x)<DMIN_ &&
           fabs(elt_vec[i].center[Y_]- min_y)<DMIN_ ){
            min_z = elt_vec[i].center[Z_];
        }
    }

    //Get the index of the center element

#ifdef DEBUG
    cout << "Diagonal Elements: " << endl;
#endif
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]- min_z)<DMIN_ )
        {
            if( fabs(elt_vec[i].center[X_]- min_x)<DMIN_ )
            {
                if( fabs(elt_vec[i].center[Y_]- min_y)<DMIN_ )
                {
#ifdef DEBUG
                    cout << elt_vec[i].center[X_]<< " ";
                    cout << elt_vec[i].center[Y_]<< " ";
                    cout << elt_vec[i].center[Z_]<< endl;
#endif
                        
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    vector<int> diagonal_index;
    double x_coor, y_coor, z_coor;

    //Find all elements on the diagonal in the +ve x,y quad
    for(int i = 0; i < isize; ++i) {
        x_coor = elt_vec[i].center[X_];
        y_coor = elt_vec[i].center[Y_];
        if( (fabs(x_coor - y_coor) < DMIN_) && x_coor > 0 && y_coor > 0 ) {
#ifdef DEBUG
            cout << x_coor << " " << y_coor << " " << endl;
#endif
            diagonal_index.push_back(i);
        }
    }

    //Sort the elements with increasing x,y
    int diag_size = diagonal_index.size();
    
    min_x = 1000000;
    double x_coor2, y_coor2, z_coor2;
    int temp;

    //use bubble sort
    for(int i = 0; i < diag_size; i++) {
        for(int  j = 0; j < diag_size - 1; j++) {
            x_coor  = elt_vec[diagonal_index[j]].center[X_];
            x_coor2 = elt_vec[diagonal_index[j+1]].center[X_];

            if(x_coor > x_coor2) {
                temp = diagonal_index[j+1];
                diagonal_index[j+1] = diagonal_index[j];
                diagonal_index[j] = temp;
            }
        }
    }

#ifdef DEBUG
    cout << "The number of diagonal elements = " << diag_size << endl;
    cout << "Sorted diagonal elements : " << endl;
    for(int i = 0; i < diag_size; i++) {
        cout << elt_vec[diagonal_index[i]].center[X_] << " ";
        cout << elt_vec[diagonal_index[i]].center[Y_] << " ";
        cout << elt_vec[diagonal_index[i]].center[Z_] << endl;
    }   
#endif    

    vector<int>::iterator the_iterator;
    the_iterator = diagonal_index.begin();

    //remove elements with non-minimum z
    for(unsigned int i = 0; i < diagonal_index.size(); i++) {
        x_coor = elt_vec[diagonal_index[i]].center[X_];
        y_coor = elt_vec[diagonal_index[i]].center[Y_];
        z_coor = elt_vec[diagonal_index[i]].center[Z_];
        min_z = z_coor;

        unsigned int j = i+1;
        while(j < diagonal_index.size()) {
        //for(unsigned int j = i+1; j < diagonal_index.size(); j++) {
            x_coor2 = elt_vec[diagonal_index[j]].center[X_];
            y_coor2 = elt_vec[diagonal_index[j]].center[Y_];
            z_coor2 = elt_vec[diagonal_index[j]].center[Z_];
                
            if( (fabs(x_coor - x_coor2) < DMIN_) &&
                (fabs(y_coor - y_coor2) < DMIN_)) {
                if(z_coor2 < min_z) {
                    min_z = z_coor2;
                    //overwrite and delete
                    diagonal_index[i] = diagonal_index[j];
                    diagonal_index.erase(the_iterator+j);
                }
                //otherwise delete the other one without overwriting
                else {
                    diagonal_index.erase(the_iterator+j);
                }
            }
            else{
                j++;
            }
        }
    }

    diag_size = diagonal_index.size();
#ifdef DEBUG
    cout << "The new number of diagonal elements = " << diag_size << endl;
    cout << "Sorted diagonal elements only with minimum Z: " << endl;
    for(int i = 0; i < diag_size; i++) {
        cout << elt_vec[diagonal_index[i]].center[X_] << " ";
        cout << elt_vec[diagonal_index[i]].center[Y_] << " ";
        cout << elt_vec[diagonal_index[i]].center[Z_] << endl;
    }   
#endif    

    for(int i = 0; i < diag_size; i++) {
        indices.push_back(diagonal_index[i]);
    }
   

}

/*===========================================================================*/
void print_eltvec_diagonal(vector<Element_type> & elt_vec,
    vector<int> & indices, ofstream files[]) {

    int size = indices.size();

    for(int i = 0; i < size; i++) {
        files[i].precision(10);
        files[i] << elt_vec[indices[i]].value - 328.15 << endl;
    }

}


/*===========================================================================*/
void print_eltvec_hetero(std::vector<isac::Element_type> & elt_vec){

    int isize = elt_vec.size();

    //Find top layer z_coor
    double top_z = 100000;

    for (int i = 0; i < isize; ++i){
        if(elt_vec[i].center[Z_]<top_z){
            top_z = elt_vec[i].center[Z_];
        }
    }   
    cout<<"Min z is "<<top_z<<endl;    


    //Only print top layer
    for (int i = 0; i < isize; ++i){
        if( fabs(elt_vec[i].center[Z_]-top_z)<FMIN_ )
        {
            cout << elt_vec[i].center[X_] - elt_vec[i].size[X_]/2.0<< " ";
            cout << elt_vec[i].center[Y_] - elt_vec[i].size[Y_]/2.0<< " ";
            cout << elt_vec[i].center[X_] + elt_vec[i].size[X_]/2.0<< " ";
            cout << elt_vec[i].center[Y_] + elt_vec[i].size[Y_]/2.0<< " ";
            cout << elt_vec[i].value;
            cout << endl;

        }
    } cout << endl; 

}


/*===========================================================================*/
void print_eltvec(std::vector<isac::Element_type> & elt_vec){

	int isize = elt_vec.size();
   
    double max_t = 0;
    double min_t = FMAX_; 

    for (int i = 0; i < isize; ++i){
        max_t = max(max_t, elt_vec[i].value);
        min_t = min(min_t, elt_vec[i].value);
		cout << elt_vec[i].value<< " ";
		cout << elt_vec[i].center[X_]<< " ";
		cout << elt_vec[i].center[Y_]<< " ";
		cout << elt_vec[i].center[Z_]<< " ";
        cout << endl;

	} cout << endl;

    cout <<"max_min_t " << max_t << " " << min_t << endl;
    cout <<"finishes_print_eltvec " << endl;
}


/*===========================================================================*/
/* Print AGC matrix format(only non-zero values are printed) */
void print_AGC(const RC_network_type & AGC){

	cout << "print conductivity matrix A==G:" << endl;

    for (int i = 0; i < AGC.size_; ++i){
		//cout << "i, AGC.A_[i].size() " << i << " " << AGC.A_[i].size() << endl;
		cout << i << ":  " << endl;
		for (int j = 0; j < int(AGC.A_[i].size()); ++j){
			cout<< "<" << AGC.A_[i][j] << "=="<<
				 AGC.G_[i][j] << "> " ;
		} cout << endl;
		cout << AGC.C_[i] << endl;
	}

    cout << "Finish printing conductivity matrix A==G" << endl;
}


/*===========================================================================*/
/* Print AGC matrix format(complete matrix) */
void print_wholeAGC(const RC_network_type & AGC){

	cout << "Printing conductivity matrix "  << endl;
	bool flag = false;
	int index = 0;

    for (int i = 0; i <=(AGC.size_-1); ++i){
		for (int j = 0; j <= (AGC.size_-1); ++j){
			flag = false;
			for (unsigned int k = 0; k < AGC.A_[i].size(); ++k){
				if (AGC.A_[i][k] == j){
					flag = true;
					printf("%-10.2e ", AGC.G_[i][k]);
					++index;
				}
			}
			if (!flag){
				cout << "0.00       " ;
				++index;
			}
		} cout << "\n\n";
	}

    cout << "Finish printing conductivity matrix "  << endl;
}

/*============================================================================*/
// Map the values of vector1 which is in the ISAC format, to the output format
// in which the locations start with minimum x and y, incrementing x first
void map_isac2xy(vector<int> & mapper, vector<Element_type> & vector1,
                vector<double> & vector2) {

    int size = mapper.size();

    vector2.resize(size);

    for(int i = 0; i < size; i++) {
        vector2[i] = vector1[mapper[i]].value;
    }
}

/*============================================================================*/
// Map the values of vector1 which is in the ISAC format, to the output format
// in which the locations start with minimum x and y, incrementing x first
void map_isac2xy(vector<int> & mapper, vector<double> & vector1,
                vector<double> & vector2) {

    int size = mapper.size();

    vector2.resize(size);

    for(int i = 0; i < size; i++) {
        vector2[i] = vector1[mapper[i]];
    }
}

/*============================================================================*/
// Map the values of vector1 which is in the XY format (locations start with
// minimum x and y, incrementing x first) to the isac format
void map_xy2isac(vector<int> mapper, vector<double> & vector1,
                vector<double> & vector2) {

    int size = mapper.size();

    vector2.resize(size);

    for(int i = 0; i < size; i++) {
        vector2[mapper[i]] = vector1[i];
    }
}

/*============================================================================*/
//Check if two areas are intersecting
bool areasIntersect(double xmin1, double xmax1, double ymin1, double ymax1,
                    double xmin2, double xmax2, double ymin2, double ymax2) {

    if( (!(xmin1 > xmax2)) && (!(xmax1 < xmin2)) ) {
        if( (!(ymin1 > ymax2)) && (!(ymax1 < ymin2)) ) {
            return 1;
        }
    }
    
    return 0;
}

/*===========================================================================*/
//print the temperatures of a level for gnuplot. Chip sizes are in mm.
void print_level(vector<vector<double> > & temp, int sizeX,
                int sizeY, double chipSizeX, double chipSizeY){

    //sizeX = 16, sizeY = 16

    double x, y; 
    double stepX = chipSizeX/(sizeX * sizeX);
    double stepY = chipSizeY/(sizeY * sizeY);
    
    y = -(chipSizeY/2.0);
    for(int i = 0; i < sizeY * sizeY; i++) {
        x = -(chipSizeX/2.0); 
        for(int j = 0; j < sizeX * sizeX; j++) {
            //cout << (j/sizeX) + sizeX * (i/sizeY) << ",";
            //cout << (i*sizeX)%(sizeX*sizeX) + j%sizeX << endl;
            cout << y << " " << x << " ";
            cout << temp[(j/sizeX) + sizeX * (i/sizeY)][
                                    (i*sizeX)%(sizeX*sizeX) + j%sizeX] << endl;
            x += stepX;
        }
        y += stepY;
        cout << endl;
    }
    
}

/*===========================================================================*/
//print the temperatures of a level for gnuplot. Chip sizes are in mm.
void print_level(vector<vector<double> > & temp,
                int numXcellsInBlock, int numYcellsInBlock,
                int numXcells, int numYcells,
                double chipSizeX, double chipSizeY, const char * filename){

    //sizeX = 16, sizeY = 16

    double x, y; 
    double stepX = chipSizeX/(numXcells);
    double stepY = chipSizeY/(numYcells);
    
    //Number of cells in a block
    int numCells = numXcellsInBlock * numYcellsInBlock;

    ofstream fout;
    fout.open(filename);
    
    y = -(chipSizeY/2.0);
    for(int i = 0; i < numYcells; i++) {
        x = -(chipSizeX/2.0); 
        for(int j = 0; j < numXcells; j++) {
            //fout << (j/sizeX) + sizeX * (i/sizeY) << ",";
            //fout << (i*sizeX)%(sizeX*sizeX) + j%sizeX << endl;
            fout << y << " " << x << " ";
            fout << temp[(j/numXcellsInBlock) + numXcellsInBlock * 
                        (i/numYcellsInBlock)][(i*numXcellsInBlock)%(numCells)
                        + j%numXcellsInBlock] << endl;
            x += stepX;
        }
        y += stepY;
        fout << endl;
    }

    fout.close();
    
}


/*===========================================================================*/
bool compute_error(vector<double> & previous_temp, 
                    vector<double> & current_temp, double error_thres) {

    int size1 = previous_temp.size();
    int size2 = current_temp.size();

    assert(size1 == size2);

    double error_num = 0;
    double error_denom = 0;

    for(int i = 0; i < size1; i++) {
        error_num += pow(current_temp[i] - previous_temp[i], 2);
        error_denom += pow(current_temp[i], 2);
    }

    double error = sqrt(error_num / error_denom);
#ifdef VERBOSE
    cout << "Error = " << error << endl;
#endif
   
    if(error > error_thres) 
        return 0;
    else
        return 1;
}


/*===========================================================================*/
// print out an error message and terminate the program
void err_sys(const char *const s) {
  cerr << s << endl;
  exit(1);
}


/*===========================================================================*/
// convert a string to an unit
void convert_unit(double &next_time, const char *const s) {
  if (next_time < 0) 
    err_sys("Negative timepoint supplied...");
  
  double unit;
  if (!strcmp(s, "s"))
    unit = 1.0;
  else if (!strcmp(s, "ms"))
    unit = 1e-3;
  else if (!strcmp(s, "us"))
    unit = 1e-6;
  else if (!strcmp(s, "ns"))
    unit = 1e-9;
  else if (!strcmp(s, "ps"))
    unit = 1e-12;
  else 
    err_sys("Incorrect unit supplied...");
    
  next_time *= unit;
}
