/****************************************************************************
**
** ISAC2: Integrated Space and Time Adaptive Chip-Package Thermal Analysis, 
**        version 2.2
**
** Copyright (C) 2018 University of Colorado at Boulder and University 
** of Michigan at Ann Arbor. All rights reserved.
**
** This file is part of ISAC2.
**
** ISAC2 is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, version 2 of the License.
**
** ISAC2 is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with ISAC2 in the file COPYING.  If not, see 
** <http://www.gnu.org/licenses/>.
**
** For feedback, comments, or questions, please contact Zyad Hassan 
** (zyad.hassan@colorado.edu) and Xi Chen (chexi@umich.edu)
**
** Please contact the authors if you are interested in making commercial
** use of ISAC2.  It may be possible to re-release under a different license.
****************************************************************************/

/****************************************************************************
** util.h
****************************************************************************/

#ifndef UTIL_H_
#define UTIL_H_

//#define VERBOSE
//#define DEBUG

#include<sys/time.h>
#include<vector>
#include<set>
#include<utility>
#include <string>
#include <cstring>
#include<cmath>
#include "isac.h"
#include "grids.h"


/* Specify the constants and parameters for the 
 * convection flow (using lump model in book
 * "heat transfer" by Anthony F. Mills 1992 edition
 * P35. Treating the temperature of the air blewed
 * from the fan as having constant temperature 
 * (45 degree) */

/* millimeter is used as unit for length */
const double Tabm_conv = 0.0;     //offset temperature of air flow
const double k_conv = 385E-3;     //fin thermal conductivity W/mm-k
const double kf_conv = 2.57E-5;   //fluid thermal conductivity W/mm-k
const double gamma_conv = 1.511E1;//kinematic viscosity mm^2/s 
const double alpha_conv = 3.807E1;// thermal diffusivity mm^2/s
const double Pr_conv = 0.713;     // gamma_conv/alpha_conv 
/* The following are user specified parameters, here
 * we define it as constant for testing */
const int N_conv = 10; // number of fins
const double U_conv = 4.0E3;    // average velocity, mm/s
//const double H_conv = 6.0E1;  // fin height mm
//const double t_conv = 2.0E0;  // fin thickness mm
//const double L_conv = 1.0E2;  // fin, baseplate length mm
//const double b_conv = 2.0E1;  //channel spacing mm

const double H_conv = 2.0E1;    // fin height mm
const double t_conv = 1.0E0;    // fin thickness mm
const double L_conv = 4.2E1;    // fin, baseplate length mm
const double b_conv = 1.05E1;   //channel spacing mm


/*===========================================================================*/
/* Some constants */
const double conductivity_1 = 157;
const double conductivity_2 = 400;
const double capacitivity_1 = 1.638E6;
const double capacitivity_2 = 3.55E6;

const double FMAX_ = 1.0e10;
const double FMIN_ = 1.0e-7;
const double DMAX_ = 1.0e15;
const double DMIN_ = 1.0e-15;
const double NEG_FMIN_ = -1.0e-7;
const int MAX_L_ = 100;
const int START_LEVEL_ = 4;

const double E_VALUE = 2.718;
const double ONEE3_ = 1E3;

class error_type {
public:
	void update_max(double a) {max_ = a > max_ ? a:max_;}
	void update_min(double a) {min_ = a < min_ ? a:min_;}
	void incre_avg(double a) {avg_ += a;}
	void update_avg(int n) {avg_ /= n;}
	void update_flag() {flag_ = max_ > thre_ ? 1:0;}
	void update_flag(double a) {flag_ = a >thre_ ? 1:flag_;}
	void update_thre(double a) {thre_ = a;}
	void reset_flag() {flag_ = 0;}
	
	error_type();
	error_type(double thre);

	double flag() {return flag_;}
	double flag() const {return flag_;}
	double thre() {return thre_;}
	double thre() const {return thre_;}

private:
	int flag_;
	double avg_;
	double max_;
	double min_;
	double thre_;
};

/*============================================================================*/
/* Obtain the CPU time for functions/operations */
void timer_set(timeval * t);
void timer_report(timeval * t);

/* Check whether two elements are neighbor or not in one dimension by given
 * the information of two elements' coordinates and sizes information */
int adj_check(double ac, double as, double bc, double bs);

/* Convert the coordinate, size information of the lowest level grids to
 * the vector of Element_type */
void pargrid_to_element(std::vector<isac::Element_type> & grid_elements,
    std::vector<grid_contour> & hoct_contour_level,
    const std::vector<double> & value);

/* Clear the contour levels that are greater than START_ELVEL_ 3 in
 * hoct_contour. Hoct_contour levels (0,1,2,3) are always kept the same,
 * they are never cleared */
void clear_contour(std::vector<std::vector<grid_contour> > & hoct_contour);

/* Based on the input chip packge info "chippkg" and the floorplan info
 * "flrplan", read in all the package and floorplan information 
 * to parameter "components", which will be used later to construct
 * the ISAC class; */
void parser(std::vector<isac::Component_type> & components, 
	const std::vector<isac::Layer_type> & chippkg,
    const std::vector<std::vector<isac::FlrplanElt_type> > & flrplan);

/* Based on the input chip packge file "chippkg_file" and floorplan file
 * "flrplan_file", read in all the package and floorplan information 
 * to parameter "components", which will be used later to construct
 * the ISAC class; 
 * parser() function does not require the user to specify the total 
 * number of layers in first line of the package file, the parser will
 * automatically calcualte the total number of layers based on the input
 * files */
void parser(std::vector<isac::Component_type> & components, 
	const char *chippkg_file, const char *flrplan_file);

/* Generate the intial partition for "chip_3D" based on the information from 
 * "input_components"(one partition for each component). Later, each layer
 * inside one component gets the same partition. */
void gene_partition(isac::ChipPkg_type & chip_3D, 
   const std::vector<isac::Component_type> & input_components);

/* Copy the layers from input_components to ChipPkg chip_3D */
void copy_comps(isac::ChipPkg_type & chip_3D, 
	const std::vector<isac::Component_type> & input_components);

/* Generate the components grids in init_partition in "chip_3D" based on 
 * information from "input_components" */
void gene_comps_par(isac::ChipPkg_type & chip_3D, 
	const std::vector<isac::Component_type> & input_components);

/* Generate the simple info. for power_grid (only sizes, coors, no neighbor 
 * info) in "chip_3D", neighbor information is generated later */
void gene_pgrid(isac::ChipPkg_type & chip_3D, 
	const std::vector<isac::Component_type> & input_components);

/* Generate the direct children of the init_partition in "chip_3D" */
void gene_initpar_children(isac::ChipPkg_type & chip_3D);

/* Form the neighbor information in Z direction for the init_parition 
 * children grids in "chip_3D" */
void form_Z_neighbor(isac::ChipPkg_type & chip_3D);

/* Generate the power grid based on the input power profile */
/* The sizes for the power_profile has already been converted to 
 * millimeter */
void gene_power_grid (par_grid * root, par_grid * isot, par_grid * cont,
    isac::Layer_type & power_layer, int order);

/* Generate the power grid simple info.(only sizes and coors) based on the 
 * input power profile */
void gene_simp_power_grid (par_grid * root, par_grid * isot, par_grid * cont,
    isac::Layer_type & power_layer, int order);

/* Generate the sizes and coors info. for the cur_grid based on 
 * "comp_xysizes" */
void gene_basicpar_grid(par_grid * cur_grid, 
    const std::pair<double, double> & comp_xysizes);

/* Generate homogeneous children partition in x/y dimension for "parent_grid"
 * "x_num" children in X direction and "y_num" children in Y direction*/
void gene_homo_child(par_grid * parent_grid, par_grid * isot, 
    par_grid * cont, std::vector<par_logic_pos> & par_pos, 
    int x_num, int y_num, int order);

/* Determine the partition */
std::vector<int> gene_divs(std::pair<double, double> min_pelts_xysizes, 
    std::pair<double, double> ref_layer_sizes, const isac::Component_type & comp, 
    int ref_xdiv, int ref_ydiv);

/* Adjust the size and coordinate information and size information of the 
 * children if necessary (when two layers have different sizes, the init_partition
 * of the layers with larger size needs to be adjusted based on the information
 * of the smallest layer) */
/* Only adjust the sizes in the x/y dimension since this function
 * will only be used by init_partition (Z size is 0)*/
void adj_init_par(par_grid * adj_grid, par_grid * ref_grid,
    int ref_xelt_num, int ref_yelt_num, int adj_xelt_num,
    int adj_yelt_num);

/* Adjust the size and coordinate information and size information of the 
 * children if necessary (when two layers has different sizes, the init_partition
 * of the layers with larger size needs to be adjusted based on the information
 * of the smallest layer) */
/* Only consider adjust the sizes in the x/y dimension since this function
 * will only be used by init_partition (Z size is 0) */
void adj_init_par(par_grid * adj_par, par_grid * ref_par,
    const std::vector<par_logic_pos> & par_pos, int order);

/* Sort all the components by their x/y sizes */
void sort_initpar_size(const std::vector<isac::Component_type> & comps, 
    std::multiset<double> & xsize_set, std::multiset<double> & ysize_set); 

/*===========================================================================*/
/* Sort all the components by their x/y sizes */
void sort_comp_sizes(const std::vector<isac::Component_type> & comps, 
    std::vector<std::pair<double, double> > & sorted_comp_sizes, 
    std::vector<int> & comp_order);

/* Find the smallest x/y sizes, as well as the number of grids (x_num, y_num) 
 * in x/y dimension of the basic partition among the components */
void basic_par_para(std::pair<double, double> & min_sizes, 
    std::pair<double, double> & min_psizes, std::pair<int, int> & basic_divs, 
    const std::vector<isac::Component_type> & comps);

/* Generate the chip grid information(root grid) for the chip package */
void gene_chip_grid (par_grid * root, par_grid * isot, par_grid * cont);

/* Convert the unit from meter to millimeter for the "chip_3D" parameters */
void unit_conversion(isac::ChipPkg_type & chip_3D);

/* Copy informatin from one "ref_grid" to "cur_grid", while "diff" means
 * the difference for the Z coordinate for two grids */
void copy_pgrid(par_grid * dst_grid, par_grid * src_grid, int diff);

/* For a vector of bool varialbes, return true if one of them has a value
 * 'true', return false only when all the variables have value 'false'*/
bool flag_true(const std::vector<bool> & flag);

/* For two vectors with the same size, find out the largest difference 
 * among their elements with same position */
double max_diff(const std::vector<double> & vec_a, 
    const std::vector<double> & vec_b);

/* For two vectors with the same size, find out the largest difference 
 * among their elements with position difference of "start_id" */
double max_diff(const std::vector<double> & vec_a, 
    const std::vector<double> & vec_b, int start_id);

/* Calculate the area intersect in x/y dimension for a grid and an Element */
double area_intersect(const par_grid * grid, const isac::Element_type & elt);

/* Calculate the area intersect in x/y dimension for two grids and an Element */
double area_intersect(const par_grid * grid_a, const par_grid * grid_b, 
    const isac::Element_type & elt);

/* Calculate the area intersect in x/y dimension for two Elements */
double area_intersect(const isac::Element_type & elt_a, 
    const isac::Element_type & elt_b);

//Print the temperature of the elements on the diagonal on the top layer
void get_indices_top_diagonal_elts(std::vector<isac::Element_type> & elt_vec,
    std::vector<int> & indices);

void print_vec_vec(const std::vector<std::vector<double> > & vec);
void print_eltvec_top(std::vector<isac::Element_type> & elt_vec);
void print_eltvec_gnuplot(std::vector<isac::Element_type> & elt_vec);
void print_hetero_eltvec_gnuplot(std::vector<isac::Element_type> & elt_vec,
        int size);
void print_eltvec_top(std::vector<isac::Element_type> & elt_vec, 
                        const char * filename);
void print_eltvec_top2(std::vector<isac::Element_type> & elt_vec, 
                        const char * filename);
void print_eltvec_top_oneblock(std::vector<isac::Element_type> & elt_vec, 
                        const char * filename);
void print_eltvec_top_center_elt_static(std::vector<isac::Element_type> & elt_vec);
void print_eltvec_top_center_elt(std::vector<isac::Element_type> & elt_vec);
void print_eltvec_hetero(std::vector<isac::Element_type> & elt_vec);
void print_eltvec(std::vector<isac::Element_type> & elt_vec);
void print_eltvec_diagonal(std::vector<isac::Element_type> & elt_vec,
    std::vector<int> & indices, std::ofstream files[]);
/* Print AGC matrix format(only non-zero values are printed) */
void print_AGC(const RC_network_type & AGC);
/* Print AGC matrix format(complete matrix) */
void print_wholeAGC(const RC_network_type & AGC);

/*===========================================================================*/

/*============================================================================*/
// Map the values of vector1 which is in the ISAC format, to the output format
// in which the locations start with minimum x and y, incrementing x first
void map_isac2xy(std::vector<int> & mapper,
                std::vector<isac::Element_type> & vector1,
                std::vector<double> & vector2);

// Map the values of vector1 which is in the ISAC format, to the output format
// in which the locations start with minimum x and y, incrementing x first
void map_isac2xy(std::vector<int> & mapper,
                std::vector<double> & vector1,
                std::vector<double> & vector2);

// Map the values of vector1 which is in the XY format (locations start with
// minimum x and y, incrementing x first) to the isac format
void map_xy2isac(std::vector<int> mapper, std::vector<double> & vector1,
                std::vector<double> & vector2);

//Check if two areas are intersecting
bool areasIntersect(double xmin1, double xmax1, double ymin1, double ymax1,
                    double xmin2, double xmax2, double ymin2, double ymax2);

//print the temperatures of a level for gnuplot
void print_level(std::vector<std::vector<double> > & temp, int sizeX, 
                    int sizeY, double chipSizeX, double chipSizeY);
void print_level(std::vector<std::vector<double> > & temp,
                    int numXcellsInBlock, int numYcellsInBlock, 
                    int numXcells, int numYcells,
                    double chipSizeX, double chipSizeY,
                    const char * filename);

bool compute_error(std::vector<double> & previous_temp,
                   std::vector<double> & current_temp, double error_thres);

// print out an error message and terminate the program
void err_sys(const char *const s);

// convert a string to an unit
void convert_unit(double &next_time, const char *const s);

#endif /* util.h */
    
